(* General Thread Machine *)
(* Copyright (c) 2005, FLINT Group *)

Require Import Classical.

Require Import Util.
(* Require Import HiMach. *)
Require Import SoundSafe.
Require Import Omega.
Require Import EqNat.

Set Implicit Arguments.

(* Thread IDs *)
Notation tid := nat (only parsing).


(* Label *)
Notation label := nat (only parsing).

(* Word Value *)
Notation word := nat (only parsing).


(* Registers *)
Inductive reg : Set := 
  | r0  | r1  | r2  | r3  | r4  | r5  | r6  | r7
  | r8  | r9  | r10 | r11 | r12 | r13 | r14 | r15
  | tn  | ta.

(* Instructions *)
Inductive insn : Set :=
  | add (rd rs rt : reg)
  | sub (rd rs rt : reg)
  | movi (rd : reg) (w : word)
  | beq (rs rt : reg) (f : label)
  | bgt (rs rt : reg) (f : label)
  | bnz (rs : reg) (f : label)
  | ld (rd rs : reg) (w : word)
  | st (rs : reg) (w : word) (rt : reg)
  | fork (f : label) (rs : reg)
  | yield.


(* Blocks *)
Inductive block : Set :=
  | seq (i : insn) (I : block)
  | jd (f : label)
  | exit.

(* Code Heap *)
Definition code := label ->? block.

(* Register File *)
Definition regs := reg -> word.

(* Memory Heap *)
Definition heap := label -> word.

(* State *)
Definition state := (regs * heap)%type.

(* Thread Queue *)
Definition queue := tid ->? (regs * block)%type.

(* Program *)
Definition prog := (code * state * queue * block)%type.

(* Register Equality *)
Definition beq_r (r r' : reg) :=
  match r,r' with
  | r0,r0 => true    | r1,r1 => true
  | r2,r2 => true    | r3,r3 => true
  | r4,r4 => true    | r5,r5 => true
  | r6,r6 => true    | r7,r7 => true
  | r8,r8 => true    | r9,r9 => true
  | r10,r10 => true  | r11,r11 => true
  | r12,r12 => true  | r13,r13 => true
  | r14,r14 => true  | r15,r15 => true
  | tn,tn => true    | ta,ta => true
  | _,_ => false
  end.

Definition ext_r (A : Type) (f : reg -> A) := ext beq_r f.
Definition ins_r (A : Type) (m : reg ->? A) := ins beq_r m.
Definition del_r (A : Type) (m : reg ->? A) := del beq_r m.

(* Vord Value Equality *)
Definition beq_w := beq_nat.
Definition ext_w (B : Type) (f : word -> B) := ext beq_w f.
Definition ins_w (B : Type) (m : word ->? B) := ins beq_w m.
Definition del_w (B : Type) (m : word ->? B) := del beq_w m.

(* Next State *)
Inductive Next : insn -> state -> state -> Prop :=
  | Next_add : forall rd rs rt R M,
               rd <> tn -> 
	       rd <> ta ->
               Next (add rd rs rt) (R,M) (ext_r R rd (R rs + R rt),M)
  | Next_sub : forall rd rs rt R M,
               rd <> tn ->
	       rd <> ta ->
               Next (sub rd rs rt) (R,M) (ext_r R rd (R rs - R rt),M)
  | Next_movi : forall rd w R M,
                rd <> tn ->
		rd <> ta ->
                Next (movi rd w) (R,M) (ext_r R rd w,M)
  | Next_ld : forall rd rs w R M,
              rd <> tn -> 
	      rd <> ta ->
              Next (ld rd rs w) (R,M) (ext_r R rd (M (R rs + w)),M)
  | Next_st : forall rs w rt R M,
              Next (st rs w rt) (R,M) (R,ext_w M (R rs + w) (R rt)).

(* Step Relation *)
Inductive Step : prog -> prog -> Prop :=
  | Step_c : forall C S Q c I' S',
             Next c S S' ->
             Step (C,S,Q,seq c I') (C,S',Q,I')

  | Step_bnz_1 : forall C R M Q rs f I',
                 R rs = 0 ->
                 Step (C,(R,M),Q,seq (bnz rs f) I') (C,(R,M),Q,I')
  | Step_bnz_2 : forall C R M Q rs f I' I'',
                 R rs <> 0 ->
                 get C f I'' ->
                 Step (C,(R,M),Q,seq (bnz rs f) I') (C,(R,M),Q,I'')

  | Step_beq_1 : forall C R M Q rs rt f I',
                 R rs <> R rt ->
                 Step (C, (R,M), Q, seq (beq rs rt f) I') (C, (R,M), Q, I')
  | Step_beq_2 : forall C R M Q rs rt f I' I'',
                 R rs = R rt ->
                 get C f I'' ->
                 Step (C, (R,M), Q, seq (beq rs rt f) I') (C, (R,M), Q, I'')
  | Step_bgt_1 : forall C R M Q rs rt f I',
                 R rs <= R rt ->
                 Step (C, (R,M), Q, seq (bgt rs rt f) I') (C, (R,M), Q, I')
  | Step_bgt_2 : forall C R M Q rs rt f I' I'',
                 R rs > R rt ->
                 get C f I'' ->
                 Step (C, (R,M), Q, seq (bgt rs rt f) I') (C, (R,M), Q, I'')

  | Step_fork : forall C R M Q f rs I' I'' t,
                t <> R tn ->
                ~has Q t ->
                get C f I'' ->
                Step (C, (R,M), Q, seq (fork f rs) I') (C, (R,M), ins_w Q t (ext_r (ext_r R tn t) ta (R rs), I''), I')

  | Step_yield : forall C R M Q I' t R' I'',
                 get (ins_w Q (R tn) (R,I')) t (R',I'') ->
                 Step (C,(R,M),Q,seq yield I') (C,(R',M),del_w (ins_w Q (R tn) (R,I')) t,I'')

  | Step_jd : forall C S Q f I',
              get C f I' ->
              Step (C,S,Q,jd f) (C,S,Q,I')

  | Step_exit_1 : forall C R M Q t R' I',
                  get Q t (R',I') ->
                  Step (C,(R,M),Q,exit) (C,(R',M),del_w Q t,I')
  | Step_exit_2 : forall C S Q,
                  ~(exists t, has Q t) ->
                  Step (C,S,Q,exit) (C,S,Q,exit).

(* Invariant *)
Definition INV := heap -> Prop.

(* Pre-Condition *)
Definition precond := state -> Prop.

(* Guarantee *)
Definition guarantee := state -> heap -> Prop.

(* Code Heap Specification *)
Definition cspec := label ->? (precond * guarantee * guarantee * guarantee)%type.

(* Thread Queue Specification *)
Definition qspec := tid ->? (precond * guarantee * guarantee)%type.

(*Definition p_and : precond -> precond -> precond :=
  fun (p1 p2 : precond) (s : state) => p1 s /\ p2 s. *)

(* Well-Formed Block *)
Inductive I_infer : INV -> cspec -> (precond * guarantee * guarantee * guarantee) 
                             -> block -> Prop :=
  | I_infer_add : forall Y rd rs rt I' (Inv : INV) (p p' : precond) (g g' A A' G G': guarantee),
                  rd <> tn ->
		  rd <> ta ->
                  (forall R M, Inv M -> p (R,M) 
		    -> p' (ext_r R rd (R rs + R rt),M)) ->
                  (forall R M M', Inv M -> p (R,M) 
		    -> g' (ext_r R rd (R rs + R rt),M) M' -> g (R,M) M') ->
                  (forall R M M' M'', Inv M -> p (R,M) 
		    -> A (R,M') M'' -> A' (ext_r R rd (R rs + R rt),M') M'') ->
                  (forall R M M' M'', Inv M -> p (R,M) 
		    -> G' (ext_r R rd (R rs + R rt),M') M'' -> G (R,M') M'') ->
                  I_infer Inv Y (p',g',A',G') I' ->
                  I_infer Inv Y (p,g,A,G) (seq (add rd rs rt) I')

  | I_infer_sub : forall Y rd rs rt I' (Inv : INV) (p p' : precond) (g g' A A' G G': guarantee),
                  rd <> tn ->
		  rd <> ta ->
                  (forall R M, Inv M -> p (R,M) -> p' (ext_r R rd (R rs - R rt),M)) ->
                  (forall R M M', Inv M -> p (R,M) 
		    -> g' (ext_r R rd (R rs - R rt),M) M' -> g (R,M) M') ->
                  (forall R M M' M'', Inv M -> p (R,M) 
		    -> A (R,M') M'' -> A' (ext_r R rd (R rs - R rt),M') M'') ->
                  (forall R M M' M'', Inv M -> p (R,M) 
		    -> G' (ext_r R rd (R rs - R rt),M') M'' -> G (R,M') M'') ->
                  I_infer Inv Y (p',g',A',G') I' ->
                  I_infer Inv Y (p,g,A,G) (seq (sub rd rs rt) I')

  | I_infer_movi : forall Y rd w I' (Inv: INV) (p p' : precond) (g g' A A' G G': guarantee),
                   rd <> tn ->
		   rd <> ta ->
                   (forall R M, Inv M -> p (R,M) -> p' (ext_r R rd w,M)) ->
                   (forall R M M', Inv M -> p (R,M) 
		     -> g' (ext_r R rd w,M) M' -> g (R,M) M') ->
                   (forall R M M' M'', Inv M -> p (R,M) 
		     -> A (R,M') M'' -> A' (ext_r R rd w,M') M'') ->
                   (forall R M M' M'', Inv M -> p (R,M) 
		     -> G' (ext_r R rd w,M') M'' -> G (R,M') M'') ->
                   I_infer Inv Y (p',g',A',G') I' ->
                   I_infer Inv Y (p,g,A,G) (seq (movi rd w) I')
                  
  | I_infer_bnz : forall Y rs f I' (Inv: INV) (p p' p'' : precond) 
                                   (g g' g'' A A' A'' G G' G'': guarantee),
                  (forall R M, Inv M -> p (R,M) -> R rs <> 0 -> p'' (R,M)) ->
                  (forall R M M', Inv M -> p (R,M) -> R rs <> 0 -> g'' (R,M) M' -> g (R,M) M') ->
                  (forall R M M' M'', Inv M -> p (R,M) -> R rs <> 0 
		    -> A (R,M') M'' -> A'' (R,M') M'') ->
                  (forall R M M' M'', Inv M -> p (R,M) -> R rs <> 0 
		    -> G'' (R,M') M'' -> G (R,M') M'') ->
                  get Y f (p'',g'',A'',G'') ->
                  (forall R M, Inv M -> p (R,M) -> R rs = 0 -> p' (R,M)) ->
                  (forall R M M', Inv M -> p (R,M) -> R rs = 0 -> g' (R,M) M' -> g (R,M) M') ->
                  (forall R M M' M'', Inv M -> p (R,M) -> R rs = 0 
		    -> A (R,M') M'' -> A' (R,M') M'') ->
                  (forall R M M' M'', Inv M -> p (R,M) -> R rs = 0 
		    -> G' (R,M') M'' -> G (R,M') M'') ->
                  I_infer Inv Y (p',g',A',G') I' ->
                  I_infer Inv Y (p,g,A,G) (seq (bnz rs f) I')

  | I_infer_beq : forall Y rs rt f I' (Inv: INV) (p p' p'' : precond) 
                                       (g g' g'' A A' A'' G G' G'': guarantee),
                  (forall R M, Inv M -> p (R,M) -> R rs = R rt -> p'' (R,M)) ->
                  (forall R M M', Inv M -> p (R,M) -> R rs = R rt -> g'' (R,M) M' -> g (R,M) M') ->
                  (forall R M M' M'', Inv M -> p (R,M) -> R rs = R rt 
		    -> A (R,M') M'' -> A'' (R,M') M'') ->
                  (forall R M M' M'', Inv M -> p (R,M) -> R rs = R rt 
		    -> G'' (R,M') M'' -> G (R,M') M'') ->
                  get Y f (p'',g'',A'',G'') ->
                  (forall R M, Inv M -> p (R,M) -> R rs <> R rt -> p' (R,M)) ->
                  (forall R M M', Inv M -> p (R,M) -> R rs <> R rt -> g' (R,M) M' -> g (R,M) M') ->
                  (forall R M M' M'', Inv M -> p (R,M) -> R rs <> R rt 
		    -> A (R,M') M'' -> A' (R,M') M'') ->
                  (forall R M M' M'', Inv M -> p (R,M) -> R rs <> R rt 
		    -> G' (R,M') M'' -> G (R,M') M'') ->
                  I_infer Inv Y (p',g',A',G') I' ->
                  I_infer Inv Y (p,g,A,G) (seq (beq rs rt f) I')

  | I_infer_bgt : forall Y rs rt f I' (Inv : INV) (p p' p'' : precond) 
                                     (g g' g'' A A' A'' G G' G'': guarantee),
                  (forall R M, Inv M -> p (R,M) -> R rs > R rt -> p'' (R,M)) ->
                  (forall R M M', Inv M -> p (R,M) -> R rs > R rt -> g'' (R,M) M' -> g (R,M) M') ->
                  (forall R M M' M'', Inv M -> p (R,M) -> R rs > R rt 
		    -> A (R,M') M'' -> A'' (R,M') M'') ->
                  (forall R M M' M'', Inv M -> p (R,M) -> R rs > R rt 
		    -> G'' (R,M') M'' -> G (R,M') M'') ->
                  get Y f (p'',g'',A'',G'') ->
                  (forall R M, Inv M -> p (R,M) -> R rs <= R rt -> p' (R,M)) ->
                  (forall R M M', Inv M -> p (R,M) -> R rs <= R rt -> g' (R,M) M' -> g (R,M) M') ->
                  (forall R M M' M'', Inv M -> p (R,M) -> R rs <= R rt 
		    -> A (R,M') M'' -> A' (R,M') M'') ->
                  (forall R M M' M'', Inv M -> p (R,M) -> R rs <= R rt 
		    -> G' (R,M') M'' -> G (R,M') M'') ->
                  I_infer Inv Y (p',g',A',G') I' ->
                  I_infer Inv Y (p,g,A,G) (seq (bgt rs rt f) I')

  | I_infer_ld : forall Y rd rs w I' (Inv : INV) (p p' : precond) (g g' A A' G G': guarantee),
                 rd <> tn ->
		 rd <> ta ->
                 (forall R M, Inv M -> p (R,M) -> p' (ext_r R rd (M (R rs + w)),M)) ->
                 (forall R M M', Inv M -> p (R,M) 
		   -> g' (ext_r R rd (M (R rs + w)),M) M' -> g (R,M) M') ->
                 (forall R M M' M'', Inv M -> p (R,M) 
		   -> A (R,M') M'' -> A' (ext_r R rd (M (R rs + w)),M') M'') ->
                 (forall R M M' M'', Inv M -> p (R,M) 
		   -> G' (ext_r R rd (M (R rs + w)),M') M'' -> G (R,M') M'') ->
                 I_infer Inv Y (p',g',A',G') I' ->
                 I_infer Inv Y (p,g,A,G) (seq (ld rd rs w) I')
                  
  | I_infer_st : forall Y rs w rt I' (Inv : INV) (p p' : precond) (g g' A A' G G': guarantee),
                 (forall R M, Inv M -> p (R,M) 
		   -> Inv (ext_w M (R rs + w) (R rt)) /\ p' (R,ext_w M (R rs + w) (R rt))) ->
                 (forall R M M', Inv M -> p (R,M) 
		   -> g' (R,ext_w M (R rs + w) (R rt)) M' -> g (R,M) M') ->
                 (forall R M M' M'', Inv M -> p (R,M) 
		   -> A (R,M') M'' -> A' (R,M') M'') ->
                 (forall R M M' M'', Inv M -> p (R,M) 
		   -> G' (R,M') M'' -> G (R,M') M'') ->
                 I_infer Inv Y (p',g',A',G') I' ->
                 I_infer Inv Y (p,g,A,G) (seq (st rs w rt) I')
                  
  | I_infer_fork : forall Y f rs I' (Inv: INV) (p p'': precond) (g A A' A'' G G' G'': guarantee),
                   (forall R M M' t, Inv M -> p (R,M) -> t <> R tn 
		     -> g (R,M) M' -> p'' (ext_r (ext_r R tn t) ta (R rs),M')) ->
                   (forall R M M' M'' t, Inv M -> p (R,M) -> t <> R tn 
		     -> G' (R,M') M'' -> A'' (ext_r (ext_r R tn t) ta (R rs),M') M'') ->
                   (forall R M M' M'' t, Inv M -> p (R,M) -> t <> R tn 
		     -> G'' (ext_r (ext_r R tn t) ta (R rs),M') M'' -> A' (R,M') M'') ->
                   (forall R M M' M'' t, Inv M -> p (R,M) -> t <> R tn 
		     -> A (R,M') M'' -> A'' (ext_r (ext_r R tn t) ta (R rs),M') M'') -> 
                   (forall R M M' M'' t, Inv M -> p (R,M) -> t <> R tn 
		     -> G'' (ext_r (ext_r R tn t) ta (R rs),M') M'' -> G (R,M') M'') ->
		   (forall R M M' M'', Inv M -> p (R,M) -> A (R, M') M'' -> A' (R,M') M'') ->
		   (forall R M M' M'', Inv M -> p (R,M) -> G' (R, M') M'' -> G (R,M') M'') ->
                   (forall R M M', Inv M -> p'' (R,M) -> A'' (R,M) M' -> p'' (R,M')) ->
                   get Y f (p'', G'', A'', G'') ->
                   I_infer Inv Y (p,g,A',G') I' ->
                   I_infer Inv Y (p,g,A,G) (seq (fork f rs) I')

  | I_infer_yield : forall Y I' (Inv:INV) (p : precond) (g A G : guarantee),
                    (forall R M, Inv M -> p (R,M) -> g (R,M) M) ->
                    (forall R M M', Inv M -> p (R,M) -> A (R,M) M'-> p (R,M')) ->
                    I_infer Inv Y (p,G,A,G) I' ->
                    I_infer Inv Y (p,g,A,G) (seq yield I')

  | I_infer_jd : forall Y f (Inv:INV) (p p' : precond) (g g' A A' G G': guarantee),
                 (forall R M, Inv M -> p (R,M) -> p' (R,M)) ->
                 (forall R M M', Inv M -> p (R,M) -> g' (R,M) M' -> g (R,M) M') ->
                 (forall R M M' M'', Inv M -> p (R,M) -> A (R,M') M'' -> A' (R,M') M'') ->
                 (forall R M M' M'', Inv M -> p (R,M) -> G' (R,M') M'' -> G (R,M') M'') ->
                 get Y f (p',g',A',G') ->
                 I_infer Inv Y (p,g,A,G) (jd f)

  | I_infer_exit : forall Y (Inv:INV) (p : precond) (g A G : guarantee),
                   (forall R M, Inv M -> p (R,M) -> g (R,M) M) ->
                   I_infer Inv Y (p,g,A,G) exit.

(* Well-Formed Code Heap *)
Inductive C_infer : INV -> cspec -> code -> Prop :=
  | C_infer_1 : forall Inv Y C,
                (forall f p g A G, get Y f (p,g,A,G) -> exists I, get C f I) ->
                (forall f I p g A G, get Y f (p,g,A,G) /\ get C f I -> I_infer Inv Y (p,g,A,G) I) ->
                C_infer Inv Y C.

(* Non-Interference *)
Inductive Q_nonin : qspec -> queue -> Prop :=
  | Q_nonin_1 : forall Z Q,
                (forall t t' p p' (A A' G G' : guarantee) R R' I I' M M', t' <> t /\ get Z t (p,A,G) /\ get Q t (R,I) /\ get Z t' (p',A',G') /\ get Q t' (R',I') -> G (R,M) M' -> A' (R',M) M') ->
                Q_nonin Z Q.

Inductive P_nonin : qspec -> queue -> guarantee -> guarantee -> regs -> Prop :=
  | P_nonin_1 : forall Z Q (A G : guarantee) R,
                (forall t' p' (A' G' : guarantee) R' I' M M', get Z t' (p',A',G') /\ get Q t' (R',I') -> G (R,M) M' -> A' (R',M) M') ->
                (forall t' p' (A' G' : guarantee) R' I' M M', get Z t' (p',A',G') /\ get Q t' (R',I') -> G' (R',M) M' -> A (R,M) M') ->
                P_nonin Z Q A G R.

(* Well-Formed Thread Queue *)
Inductive Q_infer : INV -> cspec -> qspec -> queue -> Prop :=
  | Q_infer_1 : forall Inv Y Z Q,
                (forall t R I, get Q t (R,I) -> R tn = t) ->
                (exists t, forall t', has Q t' -> t > t') ->
                (forall t R I, get Q t (R,I) -> exists p, exists A, exists G, get Z t (p,A,G)) ->
                (forall t R I p A G, get Z t (p,A,G) /\ get Q t (R,I) -> I_infer Inv Y (p,G,A,G) I) ->
                (forall t R I p A G M M', get Z t (p,A,G) /\ get Q t (R,I) 
		  -> Inv M -> p (R,M) -> A (R,M) M' -> p (R,M')) ->
                Q_nonin Z Q ->
                Q_infer Inv Y Z Q.

(* Well-Formed Program *)
Inductive P_infer : (INV * cspec * qspec * (precond * guarantee * guarantee * guarantee)) 
                               -> prog -> Prop :=
  | P_infer_1 : forall (Y: cspec) (Z : qspec) (Inv:INV) (p : precond) (g A G : guarantee) 
                           (C : code)  (R : regs) (M : heap) (Q : queue) (I : block),
                C_infer Inv Y C ->
                Q_infer Inv Y Z Q -> 
                I_infer Inv Y (p,g,A,G) I ->
		Inv M ->
                p (R,M) ->
                ~has Q (R tn) ->
                (forall (t : tid) (p' : precond) (A' G' : guarantee) 
		                     (R' : regs) (I' : block) (M' : heap), 
		  get Z t (p',A',G') /\ get Q t (R',I') -> g (R,M) M' -> p' (R',M')) ->
                P_nonin Z Q A G R ->
                P_infer (Inv,Y,Z,(p,g,A,G)) (C,(R,M),Q,I).

(* Hypothesis eta1 : forall p:precond, (fun x => p x) = p. *)

Lemma trv1 : forall t, ~t > t.
intro t.
omega.
Qed.

Lemma trv2 : forall t, ~t > S t.
intro t.
omega.
Qed.

Lemma trv3 : forall t, t <> S t.
intro t.
omega.
Qed.

Lemma trv4 : forall t, S t > t.
intro t.
omega.
Qed.

Lemma trv5 : forall t t' t'', t > t' -> t' > t'' -> t > t''.
intros t t' t''.
omega.
Qed.

Lemma trv6 : forall t t', ~ t > t' -> S t' > t.
intros t t'.
omega.
Qed.

Lemma trv7 : forall w w', beq_w w w' = true -> w' = w.
intros w w'.
unfold beq_w.
intro.
apply sym_eq.
apply EqNat.beq_nat_eq.
rewrite <- H.
auto.
Qed.


Hypothesis trv8 : forall r r', r <> r' -> beq_r r' r = false.

Lemma trv9 : forall w w', w = w' -> beq_w w w' = true.
intros.
rewrite H.
apply sym_eq.
unfold beq_w.
apply EqNat.beq_nat_refl.
Qed.


Lemma trv10 : forall w w', beq_w w w' = false -> w <> w'.
unfold beq_w.
double induction w w'.
intros.
unfold EqNat.beq_nat in H.
inversion H.
intros.
discriminate.
intros.
discriminate.
intros.
unfold  EqNat.beq_nat in H1.
fold  EqNat.beq_nat in H1.
assert (n0 <> n).
apply (H0 n H1).
omega.
Qed.


Lemma queue_nonin_ins : forall Z Q t (p : precond) (A G : guarantee) R I,
                        Q_nonin Z Q ->
                        P_nonin Z Q A G R ->
                        Q_nonin (ins_w Z t (p,A,G)) (ins_w Q t (R,I)).
Proof.
intros Z Q t p A G R I HQni HPni.
inversion_clear HQni.
rename H into HQni.
inversion_clear HPni.
rename H into HPnig.
rename H0 into HPnia.
apply Q_nonin_1.
intros t1 t2 p1 p2 A1 A2 G1 G2 R1 R2 I1 I2 M M' Ht Hgu.
elim Ht.
clear Ht.
intros Hne Ht.
elim Ht.
clear Ht.
intros Hz1 Ht.
elim Ht.
clear Ht.
intros Hq1 Ht.
elim Ht.
clear Ht.
intros Hz2 Hq2.
elim (get_ins_intro Hz1).
clear Hz1.
intro Ht.
elim Ht.
clear Ht.
intros Heq1 Hval1.
inversion Hval1.
clear Hval1.
rename H0 into Hvalp1.
rename H1 into HvalA1.
rename H2 into HvalG1.
elim (get_ins_intro Hq1).
clear Hq1.
intro Ht.
elim Ht.
clear Ht.
intros Ht Hval1.
clear Ht.
inversion Hval1.
clear Hval1.
rename H0 into HvalR1.
rename H1 into HvalI1.
elim (get_ins_intro Hz2).
clear Hz2.
intro Ht.
elim Ht.
clear Ht.
intros Heq2 Hval2.
inversion Hval2.
clear Hval2.
rename H0 into Hvalp2.
rename H1 into HvalA2.
rename H2 into HvalG2.
elim (get_ins_intro Hq2).
clear Hq2.
intro Ht.
elim Ht.
clear Ht.
intros Ht Hval2.
clear Ht.
inversion Hval2.
clear Hval2.
rename H0 into HvalR2.
rename H1 into HvalI2.

(* t1 = t /\ t2 = t *)
elim Hne.
rewrite <- (trv7 t1 t Heq1).
rewrite <- (trv7 t2 t Heq2).
trivial.

clear Hq2.
intro Ht.
elim Ht.
clear Ht.
intros Hne2 Hq2.
rewrite -> Hne2 in Heq2.
discriminate Heq2.
clear Hz2.
intro Ht.
elim Ht.
clear Ht.
intros Hne2 Hz2.
elim (get_ins_intro Hq2).
clear Hq2.
intro Ht.
elim Ht.
clear Ht.
intros Heq2 Hval2.
inversion Hval2.
clear Hval2.
rename H0 into HvalR2.
rename H1 into HvalI2.
rewrite -> Hne2 in Heq2.
discriminate Heq2.
clear Hq2.
intro Ht.
elim Ht.
clear Ht.
intros Ht Hq2.
clear Ht.

(* t1 = t /\ t2 <> t*)
apply HPnig with t2 p2 G2 I2.
split.
exact Hz2.
exact Hq2.
rewrite -> HvalG1.
rewrite -> HvalR1.
exact Hgu.

clear Hq1.
intro Ht.
elim Ht.
clear Ht.
intros Hne1 Hq1.
rewrite -> Hne1 in Heq1.
discriminate Heq1.
clear Hz1.
intro Ht.
elim Ht.
clear Ht.
intros Hne1 Hz1.
elim (get_ins_intro Hq1).
clear Hq1.
intro Ht.
elim Ht.
clear Ht.
intros Heq1 Hval1.
inversion Hval1.
clear Hval1.
rename H0 into HvalR1.
rename H1 into HvalI1.
rewrite -> Hne1 in Heq1.
discriminate Heq1.
clear Hq1.
intro Ht.
elim Ht.
clear Ht.
intros Ht Hq1.
clear Ht.
elim (get_ins_intro Hz2).
clear Hz2.
intro Ht.
elim Ht.
clear Ht.
intros Heq2 Hval2.
inversion Hval2.
clear Hval2.
rename H0 into Hvalp2.
rename H1 into HvalA2.
rename H2 into HvalG2.
elim (get_ins_intro Hq2).
clear Hq2.
intro Ht.
elim Ht.
clear Ht.
intros Ht Hval2.
clear Ht.
inversion Hval2.
clear Hval2.
rename H0 into HvalR2.
rename H1 into HvalI2.

(* t1 <> t /\ t2 = t *)
rewrite <- HvalA2.
rewrite <- HvalR2.
apply HPnia with t1 p1 A1 G1 R1 I1.
split.
exact Hz1.
exact Hq1.
exact Hgu.

clear Hq2.
intro Ht.
elim Ht.
clear Ht.
intros Hne2 Hval2.
rewrite -> Hne2 in Heq2.
discriminate Heq2.
clear Hz2.
intro Ht.
elim Ht.
clear Ht.
intros Hne2 Hz2.
elim (get_ins_intro Hq2).
clear Hq2.
intro Ht.
elim Ht.
clear Ht.
intros Heq2 Hval2.
inversion Hval2.
clear Hval2.
rename H0 into HvalR2.
rename H1 into HvalI2.
rewrite -> Hne2 in Heq2.
discriminate Heq2.
clear Hq2.
intro Ht.
elim Ht.
clear Ht.
intros Ht Hq2.
clear Ht.
apply HQni with t1 t2 p1 p2 A1 G1 G2 R1 I1 I2.
split.
exact Hne.
split.
exact Hz1.
split.
exact Hq1.
split.
exact Hz2.
exact Hq2.
exact Hgu.
Qed.

Lemma queue_nonin_del : forall Z Q t,
                        Q_nonin Z Q ->
                        Q_nonin (del_w Z t) (del_w Q t).
Proof.
intros Z Q t Hni.
apply Q_nonin_1.
intros t1 t2 p1 p2 A1 A2 G1 G2 R1 R2 I1 I2 M M' Ht Hgu.
elim Ht.
clear Ht.
intros Hne Ht.
elim Ht.
clear Ht.
intros Hz1 Ht.
elim Ht.
clear Ht.
intros Hq1 Ht.
elim Ht.
clear Ht.
intros Hz2 Hq2.
elim (get_del_intro Hz1).
clear Hz1.
intros Hne1 Hz1.
elim (get_del_intro Hq1).
clear Hq1.
intros Ht Hq1.
clear Ht.
elim (get_del_intro Hz2).
clear Hz2.
intros Hne2 Hz2.
elim (get_del_intro Hq2).
clear Hq2.
intros Ht Hq2.
clear Ht.
inversion_clear Hni.
rename H into Hni.
apply Hni with t1 t2 p1 p2 A1 G1 G2 R1 I1 I2.
split.
exact Hne.
split.
exact Hz1.
split.
exact Hq1.
split.
exact Hz2.
exact Hq2.
apply Hgu.
Qed.

Lemma prog_nonin_seq : forall Z Q (A A' G G' : guarantee) R R',
                       P_nonin Z Q A G R ->
                       (forall M M', A (R,M) M' -> A' (R',M) M') ->
                       (forall M M', G' (R',M) M' -> G (R,M) M') ->
                       P_nonin Z Q A' G' R'.
Proof.
intros Z Q A A' G G' R R' HPni Has Hgu.
inversion_clear HPni.
rename H into HPnig.
rename H0 into HPnia.
apply P_nonin_1.
intros tQ pQ AQ GQ RQ IQ M M' Ht Hg.
apply HPnig with tQ pQ GQ IQ.
exact Ht.
apply Hgu.
exact Hg.
intros tQ pQ AQ GQ RQ IQ M M' Ht Hg.
apply Has.
apply HPnia with tQ pQ AQ GQ RQ IQ.
exact Ht.
exact Hg.
Qed.

Lemma prog_nonin_ins : forall Z Q (A G : guarantee) R t p' (A' G' : guarantee) R' I',
                       P_nonin Z Q A G R ->
                       (forall M M', G (R,M) M' -> A' (R',M) M') ->
                       (forall M M', G' (R',M) M' -> A (R,M) M') ->                       
                       P_nonin (ins_w Z t (p',A',G')) (ins_w Q t (R',I')) A G R.
Proof.
intros Z Q A G R t p' A' G' R' I' HPni Hgu Has.
inversion_clear HPni.
rename H into HPnig.
rename H0 into HPnia.
apply P_nonin_1.
intros tQ pQ AQ GQ RQ IQ M M' Ht Hg.
elim Ht.
clear Ht.
intros Hz Hq.
elim (get_ins_intro Hz).
clear Hz.
intro Ht.
elim Ht.
clear Ht.
intros Heq Hval.
inversion Hval.
clear Hval.
rename H0 into Hvalp.
rename H1 into HvalA.
rename H2 into HvalG.
elim (get_ins_intro Hq).
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Ht Hval.
clear Ht.
inversion Hval.
clear Hval.
rename H0 into HvalR.
rename H1 into HvalI.

(* tQ = t *)
rewrite <- HvalA.
rewrite <- HvalR.
apply Hgu.
exact Hg.

clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Hne Hq.
rewrite -> Hne in Heq.
discriminate Heq.
clear Hz.
intro Ht.
elim Ht.
clear Ht.
intros Hne Hz.
elim (get_ins_intro Hq).
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Heq Hval.
inversion Hval.
clear Hval.
rename H0 into HvalR.
rename H1 into HvalI.
rewrite -> Hne in Heq.
discriminate Heq.
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Ht Hq.
clear Ht.

(* tQ <> t *)
apply HPnig with tQ pQ GQ IQ.
split.
exact Hz.
exact Hq.
exact Hg.

intros tQ pQ AQ GQ RQ IQ M M' Ht Hg.
elim Ht.
clear Ht.
intros Hz Hq.
elim (get_ins_intro Hz).
clear Hz.
intro Ht.
elim Ht.
clear Ht.
intros Heq Hval.
inversion Hval.
clear Hval.
rename H0 into Hvalp.
rename H1 into HvalA.
rename H2 into HvalG.
elim (get_ins_intro Hq).
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Ht Hval.
clear Ht.
inversion Hval.
clear Hval.
rename H0 into HvalR.
rename H1 into HvalI.

(* tQ = t *)
apply Has.
rewrite -> HvalG.
rewrite -> HvalR.
exact Hg.

clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Hne Hq.
rewrite -> Hne in Heq.
discriminate Heq.
clear Hz.
intro Ht.
elim Ht.
clear Ht.
intros Hne Hz.
elim (get_ins_intro Hq).
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Heq Hval.
inversion Hval.
clear Hval.
rename H0 into HvalR.
rename H1 into HvalI.
rewrite -> Hne in Heq.
discriminate Heq.
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Ht Hq.
clear Ht.

(* tQ <> t *)
apply HPnia with tQ pQ AQ GQ RQ IQ.
split.
exact Hz.
exact Hq.
exact Hg.
Qed.

Lemma prog_nonin_del : forall Z Q t p A G R I,
                       Q_nonin Z Q ->
                       get Z t (p,A,G) /\ get Q t (R,I) ->
                       P_nonin (del_w Z t) (del_w Q t) A G R.
Proof.
intros Z Q t p A G R I HQni Ht.
inversion_clear HQni.
rename H into HQni.
elim Ht.
clear Ht.
intros Hz' Hq'.
apply P_nonin_1.
intros tQ pQ AQ GQ RQ IQ M M' Ht Hg.
elim Ht.
clear Ht.
intros Hz Hq.
elim (get_del_intro Hz).
clear Hz.
intros Hne Hz.
elim (get_del_intro Hq).
clear Hq.
intros Ht Hq.
clear Ht.
apply HQni with t tQ p pQ A G GQ R I IQ.
split.
apply trv10.
exact Hne.
split.
exact Hz'.
split.
exact Hq'.
split.
exact Hz.
exact Hq.
exact Hg.
intros tQ pQ AQ GQ RQ IQ M M' Ht Hg.
elim Ht.
clear Ht.
intros Hz Hq.
elim (get_del_intro Hz).
clear Hz.
intros Hne Hz.
elim (get_del_intro Hq).
clear Hq.
intros Ht Hq.
clear Ht.
apply HQni with tQ t pQ p AQ GQ G RQ IQ I.
split.
intro Heq.
apply (trv10 Hne).
symmetry.
exact Heq.
split.
exact Hz.
split.
exact Hq.
split.
exact Hz'.
exact Hq'.
exact Hg.
Qed.

(* Progress Lemma *)
Lemma Progress : forall S P, P_infer S P -> exists P', Step P P'.
Proof.
intros S P HP.
elim HP.
clear S P HP.
intros Y Z Inv p g A G C R M Q I HC HQ HI Hinv Hp HPtn HPlo HPni.
inversion_clear HC.
rename H into HCex.
rename H0 into HCir.
inversion_clear HQ.
rename H into HQtn.
rename H0 into HQht.
rename H1 into HQex.
rename H2 into HQir.
rename H3 into HQas.
rename H4 into HQni.
cut (exists t, ~has Q t /\ t <> R tn).
intro Ht.
elim Ht.
clear Ht.
intros t Htp.
elim Htp.
clear Htp.
intros Htph Htpd.
inversion_clear HI.

(* add *)
rename H into HIro.
rename H0 into HIra.
rename H1 into HIpc.
rename H2 into HIlo.
rename H3 into HIas.
rename H4 into HIgu.
rename H5 into HIir.
exists (C,(ext_r R rd (R rs + R rt),M),Q,I').
apply Step_c.
apply Next_add; auto.

(* sub *)
rename H into HIro.
rename H0 into HIra.
rename H1 into HIpc.
rename H2 into HIlo.
rename H3 into HIas.
rename H4 into HIgu.
rename H5 into HIir.
exists (C,(ext_r R rd (R rs - R rt),M),Q,I').
apply Step_c.
apply Next_sub; auto.

(* movi *)
rename H into HIro.
rename H0 into HIra.
rename H1 into HIpc.
rename H2 into HIlo.
rename H3 into HIas.
rename H4 into HIgu.
rename H5 into HIir.
exists (C,(ext_r R rd w,M),Q,I').
apply Step_c.
apply Next_movi; auto.


(* bnz *)
rename H into HIpc'.
rename H0 into HIlo'.
rename H1 into HIas'.
rename H2 into HIgu'.
rename H3 into HIir'.
rename H4 into HIpc.
rename H5 into HIlo.
rename H6 into HIas.
rename H7 into HIgu.
rename H8 into HIir.
cut (R rs = 0 \/ R rs <> 0).
intro Ht.
elim Ht.
clear Ht.
intro Ht.
exists (C,(R,M),Q,I').
apply Step_bnz_1; auto.
clear Ht.
intro Ht.
elim (HCex f p'' g'' A'' G'').
intros I'' HCexp.
exists (C,(R,M),Q,I'').
apply Step_bnz_2; auto.
exact HIir'.
apply classic.

(* beq *)
rename H into HIpc'.
rename H0 into HIlo'.
rename H1 into HIas'.
rename H2 into HIgu'.
rename H3 into HIir'.
rename H4 into HIpc.
rename H5 into HIlo.
rename H6 into HIas.
rename H7 into HIgu.
rename H8 into HIir.
cut (R rs = R rt \/ R rs <> R rt).
intro Ht.
elim Ht.
clear Ht.
intro Ht.
elim (HCex f p'' g'' A'' G'').
intros I'' HCexp.
exists (C,(R,M),Q,I'').
apply Step_beq_2; auto.
exact HIir'.
clear Ht.
intro Ht.
exists (C,(R,M),Q,I').
apply Step_beq_1; auto.

apply classic.

(* bgt *)
rename H into HIpc'.
rename H0 into HIlo'.
rename H1 into HIas'.
rename H2 into HIgu'.
rename H3 into HIir'.
rename H4 into HIpc.
rename H5 into HIlo.
rename H6 into HIas.
rename H7 into HIgu.
rename H8 into HIir.
cut (R rs > R rt \/ R rs <= R rt).
intro Ht.
elim Ht.
clear Ht.
intro Ht.
elim (HCex f p'' g'' A'' G'').
intros I'' HCexp.
exists (C,(R,M),Q,I'').
apply Step_bgt_2; auto.
exact HIir'.
clear Ht.
intro Ht.
exists (C,(R,M),Q,I').
apply Step_bgt_1; auto.
omega.

(* ld *)
rename H into HIro.
rename H0 into HIra.
rename H1 into HIpc.
rename H2 into HIlo.
rename H3 into HIas.
rename H4 into HIgu.
rename H5 into HIir.
exists (C,(ext_r R rd (M (R rs + w)),M),Q,I').
apply Step_c.
apply Next_ld; auto.

(* st *)
rename H into HIpc.
rename H0 into HIlo.
rename H1 into HIas.
rename H2 into HIgu.
rename H3 into HIir.
exists (C,(R,ext_w M (R rs + w) (R rt)),Q,I').
apply Step_c.
apply Next_st.

(* fork *)
rename H into HIpc'.
rename H0 into HIgu1'.
rename H1 into HIgu2'.
rename H2 into HIas1'.
rename H3 into HIgu3'.
rename H4 into HIas2'.
rename H5 into HIgu4'.
rename H6 into HIas'.
rename H7 into HIir'.
rename H8 into HIir.

elim (HCex f p'' G'' A'' G'').
intros I'' HCexp.
exists (C,(R,M),ins_w Q t (ext_r (ext_r R tn t) ta (R rs),I''),I').
apply Step_fork; auto.
exact HIir'.


(* yield *)
rename H into HIlo.
rename H0 into HIas.
rename H1 into HIir.
clear t Htph Htpd.
cut (exists t, exists R', exists I'', get (ins_w Q (R tn) (R,I')) t (R',I'')).
intro Ht.
elim Ht.
clear Ht.
intros t Ht.
elim Ht.
clear Ht.
intros R' Ht.
elim Ht.
clear Ht.
intros I'' Ht.
exists (C,(R',M),del_w (ins_w Q (R tn) (R,I')) t,I'').
apply Step_yield.
exact Ht.
exists (R tn).
exists R.
exists I'.
unfold ins_w.
apply get_ins_elimi.
left.
split.
apply trv9.
trivial.
trivial.

(* jd *)
rename H into HIpc.
rename H0 into HIlo.
rename H1 into HIas.
rename H2 into HIgu.
rename H3 into HIir.
elim (HCex f p' g' A' G').
intros I' Ht.
exists (C,(R,M),Q,I').
apply Step_jd.
exact Ht.
exact HIir.

(* exit *)
rename H into HIlo.
clear t Htph Htpd.
cut ((exists t, has Q t) \/ ~(exists t, has Q t)).
intro Ht.
elim Ht.
clear Ht.
intro Ht.
elim Ht.
clear Ht.
intros t Ht.
elim Ht.
clear Ht.
intros (R',I') Ht.
exists (C,(R',M),del_w Q t,I').
apply Step_exit_1.
unfold get.
exact Ht.
clear Ht.
intro Ht.
exists (C,(R,M),Q,exit).
apply Step_exit_2.
exact Ht.
apply classic.
(* end *)

Lemma prog_tmp : forall (Q: queue) (t'' : tid), (exists t : tid, (forall t' : tid, has Q t' -> t > t')) 
                                                      -> exists t : tid, ~ has Q t /\ t <> t''.
intros Q t'' HQht.
elim HQht.
clear HQht.
intros t HQhtp.
cut (t = t'' \/ t <> t'').
intro Ht.
elim Ht.
clear Ht.
intro Ht.
exists (S t).
split.
cut (~t > S t -> ~has Q (S t)).
intros Ht'.
apply Ht'.
apply trv2.
intros Ht' Ht''.
apply Ht'.
apply HQhtp.
exact Ht''.
intro Ht'.
rewrite <- Ht' in Ht.
apply (trv3 Ht).
clear Ht.
intro Ht.
exists t.
split.
cut (~t > t -> ~has Q t).
intro Ht'.
apply Ht'.
apply trv1.
intros Ht' Ht''.
apply Ht'.
apply HQhtp.
exact Ht''.
exact Ht.
apply classic.
Qed.
apply prog_tmp; auto.
Qed.


(* Preservation Lemma *)
Lemma Preservation : forall S P P', P_infer S P -> Step P P' -> exists S', P_infer S' P'.
Proof.
intros S P P' HP Hstep.
inversion HP.
rename H into HC.
rename H0 into HQ.
rename H1 into HI.
rename H2 into Hinv.
rename H3 into Hp.
rename H4 into HPtn.
rename H5 into HPlo.
rename H6 into HPni.
rewrite <- H7 in HP.
rewrite <- H8 in HP.
rewrite <- H8 in Hstep.
clear S P H7 H8.
inversion HC.
rename H into HCex.
rename H0 into HCir.
clear Y0 C0 H1 H2 H3.
inversion HQ.
rename H into HQtn.
rename H0 into HQht.
rename H1 into HQex.
rename H2 into HQir.
rename H3 into HQas.
rename H4 into HQni.
clear Y0 Z0 Q0 H5 H6 H7 H8.
inversion HI.

(* add *)
rename H3 into HIro.
rename H4 into HIra.
rename H5 into HIpc.
rename H8 into HIlo.
rename H10 into HIas.
rename H11 into HIgu.
rename H12 into HIir.
rewrite <- H9 in Hstep.
rewrite <- H9 in HP.
rewrite <- H9 in HI.
clear Y0 p0 g0 A0 G0 I H6 H7 H H0 H1 H2 H9.
inversion Hstep.
rename H5 into Hnext.
rewrite <- H0 in Hstep.
clear C0 S Q0 c I'0 P' H H1 H2 H3 H4 H0.
inversion Hnext.
rewrite <- H1 in Hstep.
rewrite <- H1 in Hnext.
rename H5 into HSro.
rename H6 into HSra.
clear rd0 rs0 rt0 R0 M0 S' H H1 H2 H3 H4 H0.
exists (Inv,Y,Z,(p',g',A',G')).
apply P_infer_1.
exact HC.
exact HQ.
exact HIir.
exact Hinv.
apply HIpc; auto.

unfold ext_r.
unfold ext.
rewrite -> (trv8 HSro).
exact HPtn.

intros tQ pQ AQ GQ RQ IQ M' Ht Hg.
apply (HPlo tQ pQ AQ GQ RQ IQ M' Ht (HIlo R M M' Hinv Hp Hg)).



apply prog_nonin_seq with A G R.
exact HPni.
intros M' M'' Ha.
apply HIas with M; auto.
intros M' M'' Hg.
apply HIgu with M; auto.

(* sub *)
rename H3 into HIro.
rename H4 into HIra.
rename H5 into HIpc.
rename H8 into HIlo.
rename H10 into HIas.
rename H11 into HIgu.
rename H12 into HIir.
rewrite <- H9 in Hstep.
rewrite <- H9 in HP.
rewrite <- H9 in HI.
clear Y0 p0 g0 A0 G0 I H6 H7 H H0 H1 H2 H9.
inversion Hstep.
rename H5 into Hnext.
rewrite <- H0 in Hstep.
clear C0 S Q0 c I'0 P' H H1 H2 H3 H4 H0.
inversion Hnext.
rewrite <- H1 in Hstep.
rewrite <- H1 in Hnext.
rename H5 into HSro.
rename H6 into HSra.
clear rd0 rs0 rt0 R0 M0 S' H H1 H2 H3 H4 H0.
exists (Inv,Y,Z,(p',g',A',G')).
apply P_infer_1.
exact HC.
exact HQ.
exact HIir.
exact Hinv.
apply HIpc; auto.

unfold ext_r.
unfold ext.
rewrite -> (trv8 HSro).
exact HPtn.

intros tQ pQ AQ GQ RQ IQ M' Ht Hg.
apply (HPlo tQ pQ AQ GQ RQ IQ M' Ht (HIlo R M M' Hinv Hp Hg)).



apply prog_nonin_seq with A G R.
exact HPni.
intros M' M'' Ha.
apply HIas with M; auto.
intros M' M'' Hg.
apply HIgu with M; auto.

(* movi *)
rename H3 into HIro.
rename H4 into HIra.
rename H5 into HIpc.
rename H8 into HIlo.
rename H10 into HIas.
rename H11 into HIgu.
rename H12 into HIir.
rewrite <- H9 in Hstep.
rewrite <- H9 in HP.
rewrite <- H9 in HI.
clear Y0 p0 g0 A0 G0 I H6 H7 H H0 H1 H2 H9.
inversion Hstep.
rename H5 into Hnext.
rewrite <- H0 in Hstep.
clear C0 S Q0 c I'0 P' H H1 H2 H3 H4 H0.
inversion Hnext.
rewrite <- H2 in Hstep.
rewrite <- H2 in Hnext.
rename H4 into HSro.
rename H5 into HSra.
clear rd0 w0 R0 M0 S' H H1 H2 H3 H0.
exists (Inv,Y,Z,(p',g',A',G')).
apply P_infer_1.
exact HC.
exact HQ.
exact HIir.
exact Hinv.
apply HIpc; auto.

unfold ext_r.
unfold ext.
rewrite -> (trv8 HSro).
exact HPtn.

intros tQ pQ AQ GQ RQ IQ M' Ht Hg.
apply (HPlo tQ pQ AQ GQ RQ IQ M' Ht (HIlo R M M' Hinv Hp Hg)).


apply prog_nonin_seq with A G R.
exact HPni.
intros M' M'' Ha.
apply HIas with M; auto.
intros M' M'' Hg.
apply HIgu with M; auto.


(* bnz *)
rename H3 into HIpc'.
rename H4 into HIlo'.
rename H5 into HIas'.
rename H6 into HIgu'.
rename H7 into HIir'.
rename H8 into HIpc.
rename H11 into HIlo.
rename H13 into HIas.
rename H14 into HIgu.
rename H15 into HIir.
rewrite <- H12 in Hstep.
rewrite <- H12 in HP.
rewrite <- H12 in HI.
clear Y0 p0 g0 A0 G0 I H9 H10 H H0 H1 H2 H12.
inversion Hstep.
inversion H5.

rewrite <- H0 in Hstep.
rename H7 into HScd.
clear C0 R0 M0 Q0 rs0 f0 I'0 P' H H1 H2 H3 H4 H5 H6 H0.
exists (Inv, Y,Z,(p',g',A',G')).
apply P_infer_1.
exact HC.
exact HQ.
exact HIir.
exact Hinv.
apply HIpc; auto.

exact HPtn.

intros tQ pQ AQ GQ RQ IQ M' Ht Hg.
apply (HPlo tQ pQ AQ GQ RQ IQ M' Ht (HIlo R M M' Hinv Hp HScd Hg)).

apply prog_nonin_seq with A G R.
exact HPni.
intros M' M'' Ha.
apply HIas with M; auto.
intros M' M'' Hg.
apply HIgu with M; auto.


rewrite <- H1 in Hstep.
rename H7 into HScd.
rename H8 into HSge.
clear C0 R0 M0 Q0 rs0 f0 I'0 P' H H0 H2 H3 H4 H5 H6 H1.
exists (Inv, Y,Z,(p'',g'',A'',G'')).
apply P_infer_1.
exact HC.
exact HQ.
apply HCir with f.
split.
exact HIir'.
exact HSge.
exact Hinv.
apply HIpc'; auto.

exact HPtn.

intros tQ pQ AQ GQ RQ IQ M' Ht Hg.
apply (HPlo tQ pQ AQ GQ RQ IQ M' Ht (HIlo' R M M' Hinv Hp HScd Hg)).

apply prog_nonin_seq with A G R.
exact HPni.
intros M' M'' Ha.
apply HIas' with M; auto.
intros M' M'' Hg.
apply HIgu' with M; auto.

(* beq *)
rename H3 into HIpc'.
rename H4 into HIlo'.
rename H5 into HIas'.
rename H6 into HIgu'.
rename H7 into HIir'.
rename H8 into HIpc.
rename H11 into HIlo.
rename H13 into HIas.
rename H14 into HIgu.
rename H15 into HIir.
rewrite <- H12 in Hstep.
rewrite <- H12 in HP.
rewrite <- H12 in HI.
clear Y0 p0 g0 A0 G0 I H9 H10 H H0 H1 H2 H12.
inversion Hstep.
inversion H5.

rewrite <- H0 in Hstep.
rename H8 into HScd.
clear C0 R0 M0 Q0 rs0 f0 I'0 P' H H1 H2 H3 H4 H5 H6 H7 H0.
exists (Inv, Y,Z,(p',g',A',G')).
apply P_infer_1.
exact HC.
exact HQ.
exact HIir.
exact Hinv.
apply HIpc; auto.

exact HPtn.

intros tQ pQ AQ GQ RQ IQ M' Ht Hg.
apply (HPlo tQ pQ AQ GQ RQ IQ M' Ht (HIlo R M M' Hinv Hp HScd Hg)).

apply prog_nonin_seq with A G R.
exact HPni.
intros M' M'' Ha.
apply HIas with M; auto.
intros M' M'' Hg.
apply HIgu with M; auto.


rewrite <- H1 in Hstep.
rename H8 into HScd.
rename H9 into HSge.
clear C0 R0 M0 Q0 rs0 f0 I'0 P' H H0 H2 H3 H4 H5 H6 H7 H1.
exists (Inv, Y,Z,(p'',g'',A'',G'')).
apply P_infer_1.
exact HC.
exact HQ.
apply HCir with f.
split.
exact HIir'.
exact HSge.
exact Hinv.
apply HIpc'; auto.

exact HPtn.

intros tQ pQ AQ GQ RQ IQ M' Ht Hg.
apply (HPlo tQ pQ AQ GQ RQ IQ M' Ht (HIlo' R M M' Hinv Hp HScd Hg)).

apply prog_nonin_seq with A G R.
exact HPni.
intros M' M'' Ha.
apply HIas' with M; auto.
intros M' M'' Hg.
apply HIgu' with M; auto.


(* bgt *)
rename H3 into HIpc'.
rename H4 into HIlo'.
rename H5 into HIas'.
rename H6 into HIgu'.
rename H7 into HIir'.
rename H8 into HIpc.
rename H11 into HIlo.
rename H13 into HIas.
rename H14 into HIgu.
rename H15 into HIir.
rewrite <- H12 in Hstep.
rewrite <- H12 in HP.
rewrite <- H12 in HI.
clear Y0 p0 g0 A0 G0 I H9 H10 H H0 H1 H2 H12.
inversion Hstep.
inversion H5.

rewrite <- H0 in Hstep.
rename H8 into HScd.
clear C0 R0 M0 Q0 rs0 f0 I'0 P' H H1 H2 H3 H4 H5 H6 H7 H0.
exists (Inv, Y,Z,(p',g',A',G')).
apply P_infer_1.
exact HC.
exact HQ.
exact HIir.
exact Hinv.
apply HIpc; auto.

exact HPtn.

intros tQ pQ AQ GQ RQ IQ M' Ht Hg.
apply (HPlo tQ pQ AQ GQ RQ IQ M' Ht (HIlo R M M' Hinv Hp HScd Hg)).

apply prog_nonin_seq with A G R.
exact HPni.
intros M' M'' Ha.
apply HIas with M; auto.
intros M' M'' Hg.
apply HIgu with M; auto.


rewrite <- H1 in Hstep.
rename H8 into HScd.
rename H9 into HSge.
clear C0 R0 M0 Q0 rs0 f0 I'0 P' H H0 H2 H3 H4 H5 H6 H7 H1.
exists (Inv, Y,Z,(p'',g'',A'',G'')).
apply P_infer_1.
exact HC.
exact HQ.
apply HCir with f.
split.
exact HIir'.
exact HSge.
exact Hinv.
apply HIpc'; auto.

exact HPtn.

intros tQ pQ AQ GQ RQ IQ M' Ht Hg.
apply (HPlo tQ pQ AQ GQ RQ IQ M' Ht (HIlo' R M M' Hinv Hp HScd Hg)).

apply prog_nonin_seq with A G R.
exact HPni.
intros M' M'' Ha.
apply HIas' with M; auto.
intros M' M'' Hg.
apply HIgu' with M; auto.

(* ld *)
rename H3 into HIro.
rename H4 into HIra.
rename H5 into HIpc.
rename H8 into HIlo.
rename H10 into HIas.
rename H11 into HIgu.
rename H12 into HIir.
rewrite <- H9 in Hstep.
rewrite <- H9 in HP.
rewrite <- H9 in HI.
clear Y0 p0 g0 A0 G0 I H6 H7 H H0 H1 H2 H9.
inversion Hstep.
rename H5 into Hnext.
rewrite <- H0 in Hstep.
clear C0 S Q0 c I'0 P' H H1 H2 H3 H4 H0.
inversion Hnext.
rewrite <- H1 in Hstep.
rewrite <- H1 in Hnext.
rename H5 into HSro.
rename H6 into HSra.
clear rd0 rs0 R0 M0 S' H H1 H2 H3 H4 H0.
exists (Inv,Y,Z,(p',g',A',G')).
apply P_infer_1.
exact HC.
exact HQ.
exact HIir.
exact Hinv.
apply HIpc; auto.

unfold ext_r.
unfold ext.
rewrite -> (trv8 HSro).
exact HPtn.

intros tQ pQ AQ GQ RQ IQ M' Ht Hg.
apply (HPlo tQ pQ AQ GQ RQ IQ M' Ht (HIlo R M M' Hinv Hp Hg)).



apply prog_nonin_seq with A G R.
exact HPni.
intros M' M'' Ha.
apply HIas; auto.
intros M' M'' Hg.
apply HIgu with M; auto.


(* st *)
rename H3 into HIpc.
rename H6 into HIlo.
rename H8 into HIas.
rename H9 into HIgu.
rename H10 into HIir.
rewrite <- H7 in Hstep.
rewrite <- H7 in HP.
rewrite <- H7 in HI.
clear Y0 p0 g0 A0 G0 I H4 H5 H H0 H1 H2 H7.
inversion Hstep.
rename H5 into Hnext.
rewrite <- H0 in Hstep.
clear C0 S Q0 c I'0 P' H H1 H2 H3 H4 H0.
inversion Hnext.
rewrite <- H3 in Hstep.
rewrite <- H3 in Hnext.
clear rs0 w0 rt0 R0 M0 S' H0 H1 H2 H H4 H3.
exists (Inv,Y,Z,(p',g',A',G')).
apply P_infer_1.
exact HC.
exact HQ.
exact HIir.
assert (Inv (ext_w M (R rs + w) (R rt)) /\
         p' (R, ext_w M (R rs + w) (R rt))).
apply HIpc; auto.
inversion_clear H; auto.
assert (Inv (ext_w M (R rs + w) (R rt)) /\
         p' (R, ext_w M (R rs + w) (R rt))).
apply HIpc; auto.
inversion_clear H; auto.

exact HPtn.

intros tQ pQ AQ GQ RQ IQ M' Ht Hg.
apply (HPlo tQ pQ AQ GQ RQ IQ M' Ht (HIlo R M M' Hinv Hp Hg)).


apply prog_nonin_seq with A G R.
exact HPni.
intros M' M'' Ha.
apply HIas with M; auto.
intros M' M'' Hg.
apply HIgu with M; auto.

(* fork *)
rename H3 into HIpc'.
rename H4 into HIas1'.
rename H5 into HIas2'.
rename H6 into HIgu1'.
rename H7 into HIgu2'.
rename H8 into HIas.
rename H11 into HIgu.
rename H13 into HIas'.
rename H14 into HIir'.
rename H15 into HIir.
rewrite <- H12 in Hstep.
rewrite <- H12 in HP.
rewrite <- H12 in HI.
clear Y0 p0 g0 A0 G0 I H9 H10 H H0 H1 H2 H12.
inversion Hstep.
inversion H5.

rewrite <- H2 in Hstep.
rename H7 into HStn.
rename H8 into HSht.
rename H9 into HSge.
clear C0 R0 M0 Q0 rs0 f0 I'0 P' H H0 H1 H2 H4 H5 H6 H3.
exists (Inv,Y,ins_w Z t (p'',A'',G''),(p,g,A',G')).
apply P_infer_1.
exact HC.
apply Q_infer_1.

intros tQ RQ IQ Ht.
elim (get_ins_intro Ht).
clear Ht.
intro Ht.
elim Ht.
clear Ht.
intros Ht Ht'.
inversion Ht'.
clear Ht' H0 H1.
rewrite <- (trv7 tQ t Ht).
unfold ext_r.
unfold ext.
unfold beq_r.
trivial.
clear Ht.
intro Ht.
elim Ht.
clear Ht.
intros Ht Ht'.
apply HQtn with IQ.
exact Ht'.

elim HQht.
intros t' HQhtp.
cut (t > t' \/ ~t > t').
intro Ht.
elim Ht.
clear Ht.
intro Ht.
exists (S t).
intros tQ Hhas.
elim (has_ins_intro Hhas).
clear Hhas.
intro Hhas.
rewrite -> (trv7 tQ t Hhas).
apply trv4.
clear Hhas.
intro Ht'.
elim Ht'.
clear Ht'.
intros Hne Hhas.
apply (trv5 (trv5 (trv4 t) Ht) (HQhtp tQ Hhas)).
clear Ht.
intro Ht.
exists (S t').
intros tQ Hhas.
elim (has_ins_intro Hhas).
clear Hhas.
intro Heq.
rewrite <- (trv7 tQ t Heq).
apply trv6.
exact Ht.
clear Hhas.
intro Ht'.
elim Ht'.
clear Ht'.
intros Hne Hhas.
apply (trv5 (trv4 t') (HQhtp tQ Hhas)).
apply classic.

intros tQ RQ IQ Ht.
elim (get_ins_intro Ht).
clear Ht.
intros Ht.
elim Ht.
clear Ht.
intros Heq Ht.
exists p''.
exists A''.
exists G''.
unfold ins_w.
apply get_ins_elimi.
left.
split.
exact Heq.
trivial.
clear Ht.
intro Ht.
elim Ht.
clear Ht.
intros Hne Ht.
elim (HQex tQ RQ IQ Ht).
clear Ht.
intros pQ Ht.
elim Ht.
clear Ht.
intros AQ Ht.
elim Ht.
clear Ht.
intros GQ Ht.
exists pQ.
exists AQ.
exists GQ.
unfold ins_w.
apply get_ins_elimi.
right.
split.
exact Hne.
exact Ht.

intros tQ RQ IQ pQ AQ GQ Ht.
elim Ht.
clear Ht.
intros Hz Hq.
elim (get_ins_intro Hz).
clear Hz.
intro Ht.
elim Ht.
clear Ht.
intros Heq Hval.
inversion Hval.
clear Hval.
rename H0 into Hvalp.
rename H1 into HvalA.
rename H2 into HvalG.
elim (get_ins_intro Hq).
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Heq' Hval.
clear Heq'.
inversion Hval.
clear Hval.
rename H0 into HvalR.
rename H1 into HvalI.
apply HCir with f.
split.
rewrite <- Hvalp.
rewrite <- HvalG.
rewrite <- HvalA.
exact HIir'.
rewrite <- HvalI.
exact HSge.
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Hne Hq.
rewrite -> Hne in Heq.
discriminate.
clear Hz.
intro Ht.
elim Ht.
clear Ht.
intros Hne Hz.
elim (get_ins_intro Hq).
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Heq Hval.
rewrite -> Hne in Heq.
discriminate.
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Hne' Hq.
clear Hne'.
apply HQir with tQ RQ.
split.
exact Hz.
exact Hq.


intros tQ RQ IQ pQ AQ GQ M' M'' Ht Hinva Hpa Has.
elim Ht.
clear Ht.
intros Hz Hq.
elim (get_ins_intro Hz).
clear Hz.
intro Ht.
elim Ht.
clear Ht.
intros Heq Hval.
inversion Hval.
clear Hval.
rename H0 into Hvalp.
rename H1 into HvalA.
rename H2 into HvalG.
elim (get_ins_intro Hq).
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Heq' Hval.
clear Heq'.
inversion Hval.
clear Hval.
rename H0 into HvalR.
rename H1 into HvalI.
rewrite <- Hvalp.
apply HIas' with M'.
exact Hinva.
rewrite  Hvalp.
rewrite  HvalR.
exact Hpa.
rewrite  HvalA.
rewrite  HvalR.
exact Has.
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Hne Hq.
rewrite -> Hne in Heq.
discriminate.
clear Hz.
intro Ht.
elim Ht.
clear Ht.
intros Hne Hz.
elim (get_ins_intro Hq).
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Heq Hval.
rewrite -> Hne in Heq.
discriminate.
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Hne' Hq.
clear Hne'.
apply (HQas tQ RQ IQ pQ AQ GQ M' M'' (conj Hz Hq) Hinva Hpa Has).


apply queue_nonin_ins.
exact HQni.
apply prog_nonin_seq with A G R.
exact HPni.
intros M' M'' Ha.
apply HIgu1' with M; auto.
intros M' M'' Ha.
apply HIgu2' with M t; auto.
auto.
auto.
auto.

unfold ext_r.
unfold ext.
intro Ht.
elim (has_ins_intro Ht).
clear Ht.
intro Ht.
elim HStn.
apply (trv7 (R tn) t Ht).
clear Ht.
intro Ht.
elim Ht.
clear Ht.
intros Ht Ht'.
apply HPtn.
exact Ht'.

intros tQ pQ AQ GQ RQ IQ M' Ht Hg.
elim Ht.
clear Ht.
intros Hz Hq.
elim (get_ins_intro Hz).
clear Hz.
intro Ht.
elim Ht.
clear Ht.
intros Heq Hval.
inversion Hval.
clear Hval.
rename H0 into Hvalp.
rename H1 into HvalA.
rename H2 into HvalG.
elim (get_ins_intro Hq).
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Heq' Hval.
clear Heq'.
inversion Hval.
clear Hval.
rename H0 into HvalR.
rename H1 into HvalI.
rewrite <- Hvalp.
apply HIpc' with M; auto.

clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Hne Hq.
rewrite -> Hne in Heq.
discriminate.

clear Hz.
intro Ht.
elim Ht.
clear Ht.
intros Hne Hz.
elim (get_ins_intro Hq).
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Heq Hval.
rewrite -> Hne in Heq.
discriminate.
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Hne' Hq.
clear Hne'.
apply (HPlo tQ pQ AQ GQ RQ IQ M' (conj Hz Hq) Hg).

apply prog_nonin_ins.
apply prog_nonin_seq with A G R.
exact HPni.
intros M' M'' Ha.
apply HIas with M; auto.
intros M' M'' Hg.
apply HIgu with M; auto.
intros M' M'' Ha.
apply HIas1' with M; auto.
intros M' M'' Ha.
apply HIas2' with M t; auto.

(* yield *)
rename H6 into HIlo.
rename H7 into HIas.
rename H8 into HIir.
rewrite <- H4 in HP.
rewrite <- H4 in Hstep.
rewrite <- H4 in HI.
clear Y0 p0 g0 A0 G0 H2 H3 H H0 H1 H5 I H4.
inversion Hstep.
rename H5 into Hnext.
rewrite <- H0 in Hstep.
clear S H1 C0 H Q0 H2 c H3 I'0 H4 P' H0.
inversion Hnext.
rewrite <- H0 in Hstep.
rename H5 into HIge.
clear P' H0 C0 H R0 H1 M0 H2 Q0 H3 I'0 H4.
elim (get_ins_intro HIge).
clear HIge.
intro Ht.
elim Ht.
clear Ht.
intros Heq Hval.
inversion Hval.
rename H0 into HvalR.
rename H1 into HvalI.
clear Hval.
rewrite <- HvalR in Hstep.
rewrite <- HvalI in Hstep.
rewrite <- HvalR.
rewrite <- HvalI.
clear R' I'' HvalR HvalI.
rewrite <- (trv7 t (R tn) Heq).

exists (Inv,Y,del_w (ins_w Z (R tn) (p,A,G)) (R tn),(p,G,A,G)).
apply P_infer_1.
exact HC.
apply Q_infer_1.

intros tQ RQ IQ Ht.
elim (get_del_intro Ht).
clear Ht.
intros Hne Ht.
elim (get_ins_intro Ht).
clear Ht.
intro Ht.
elim Ht.
clear Ht.
intros Heq' Hval.
rewrite -> Heq' in Hne.
discriminate Hne.
clear Ht.
intro Ht.
elim Ht.
clear Ht.
intros Hne' Hq.
apply HQtn with IQ.
exact Hq.

elim HQht.
intros t' Ht.
exists t'.
intros t'' Hhas.
elim (has_del_intro Hhas).
clear Hhas.
intros Hne Hhas.
elim (has_ins_intro Hhas).
clear Hhas.
intro Heq'.
rewrite -> Heq' in Hne.
discriminate Hne.
clear Hhas.
intro Ht'.
elim Ht'.
clear Ht'.
intros Ht' Hhas.
clear Ht'.
apply Ht.
exact Hhas.

intros tQ RQ IQ Ht.
elim (get_del_intro Ht).
clear Ht.
intros Hne Ht.
elim (get_ins_intro Ht).
clear Ht.
intro Ht.
elim Ht.
clear Ht.
intros Heq' Hval.
rewrite -> Heq' in Hne.
discriminate Hne.
clear Ht.
intro Ht.
elim Ht.
clear Ht.
intros Hne' Hq.
elim (HQex tQ RQ IQ).
intros pQ Ht.
elim Ht.
clear Ht.
intros AQ Ht.
elim Ht.
clear Ht.
intros GQ Hz.
exists pQ.
exists AQ.
exists GQ.
unfold del_w.
apply get_del_elimi.
split.
exact Hne'.
unfold ins_w.
apply get_ins_elimi.
right.
split.
exact Hne'.
exact Hz.
exact Hq.

intros tQ RQ IQ pQ AQ GQ Ht.
elim Ht.
clear Ht.
intros Hz Hq.
elim (get_del_intro Hz).
clear Hz.
intros Hne Hz.
elim (get_ins_intro Hz).
clear Hz.
intro Ht.
elim Ht.
clear Ht.
intros Heq' Hval.
rewrite -> Heq' in Hne.
discriminate Hne.
clear Hz.
intro Ht.
elim Ht.
clear Ht.
intros Hne' Hz.
apply HQir with tQ RQ.
split.
exact Hz.
elim (get_del_intro Hq).
clear Hq.
intros Hne'' Hq.
elim (get_ins_intro Hq).
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Heq' Hval.
rewrite -> Heq' in Hne.
discriminate Hne.
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Hne''' Hq.
exact Hq.

intros tQ RQ IQ pQ AQ GQ M' M'' Ht Hinv' Ht' HIas'.
elim Ht.
clear Ht.
intros Hz Hq.
elim (get_del_intro Hz).
clear Hz.
intros Hne Hz.
elim (get_ins_intro Hz).
clear Hz.
intro Ht.
elim Ht.
clear Ht.
intros Heq' Hval.
rewrite -> Heq' in Hne.
discriminate Hne.
clear Hz.
intro Ht.
elim Ht.
clear Ht.
intros Hne' Hz.


assert (get Q tQ (RQ, IQ)).
elim (get_del_intro Hq).
clear Hq.
intros Hne'' Hq.
elim (get_ins_intro Hq).
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Heq' Hval.
rewrite -> Heq' in Hne.
discriminate Hne.
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Hne''' Hq.
exact Hq.

apply (HQas tQ RQ IQ pQ AQ GQ M' M'' (conj Hz H) Hinv' Ht' HIas').


apply queue_nonin_del.
apply queue_nonin_ins.
exact HQni.
exact HPni.
exact HIir.
exact Hinv.
exact Hp.

intro Hhas.
elim (has_del_intro Hhas).
clear Hhas.
intros Hne Hhas.
elim (trv10 Hne).
trivial.

intros tQ pQ AQ GQ RQ IQ M' Ht Hg.
elim Ht.
clear Ht.
intros Hz Hq.
elim (get_del_intro Hz).
clear Hz.
intros Hne Hz.
elim (get_ins_intro Hz).
clear Hz.
intro Ht.
elim Ht.
clear Ht.
intros Heq' Hval.
rewrite -> Heq' in Hne.
discriminate Hne.
clear Hz.
intro Ht.
elim Ht.
clear Ht.
intros Ht Hz.
clear Ht.
assert (get Q tQ (RQ, IQ)).
elim (get_del_intro Hq).
clear Hq.
intros Hne'' Hq.
elim (get_ins_intro Hq).
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Heq' Hval.
rewrite -> Heq' in Hne.
discriminate Hne.
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Hne''' Hq.
exact Hq.

assert (g (R, M) M).
apply HIlo.
exact Hinv.
exact Hp.

assert (pQ (RQ, M)).
apply (HPlo tQ pQ AQ GQ RQ IQ M (conj Hz H) H0).


assert (AQ (RQ, M) M').
inversion_clear HPni.
rename H2 into HPnig.
rename H3 into HPnia.
apply HPnig with tQ pQ GQ IQ.
split.
exact Hz.
exact H.
exact Hg.


apply (HQas tQ RQ IQ pQ AQ GQ M M' (conj Hz H) Hinv H1 H2).

apply prog_nonin_del with p I'.
apply queue_nonin_ins.
exact HQni.
exact HPni.
split.
unfold ins_w.
apply get_ins_elimi.
left.
split.
apply trv9.
trivial.
trivial.
unfold ins_w.
apply get_ins_elimi.
left.
split.
apply trv9.
trivial.
trivial.

clear HIge.
intro Ht.
elim Ht.
clear Ht.
intros Hne Hget.
elim (HQex t R' I'').
intros p' Ht.
elim Ht.
clear Ht.
intros A' Ht.
elim Ht.
clear Ht.
intros G' HIgez. 
exists (Inv,Y,del_w (ins_w Z (R tn) (p,A,G)) t,(p',G',A',G')).
apply P_infer_1.
exact HC.
apply Q_infer_1.

intros tQ RQ IQ Ht.
elim (get_del_intro Ht).
clear Ht.
intros Hne' Ht.
elim (get_ins_intro Ht).
clear Ht.
intro Ht.
elim Ht.
clear Ht.
intros Heq Hval.
inversion Hval.
clear Hval.
rename H0 into HvalR.
rename H1 into HvalI.
rewrite <- HvalR.
apply (trv7 tQ (R tn) Heq).
clear Ht.
intros Ht.
elim Ht.
clear Ht.
intros Hne'' Hget'.
apply HQtn with IQ.
exact Hget'.

elim HQht.
intros t' HQhtp.
cut (R tn > t' \/ ~R tn > t').
intro Ht.
elim Ht.
clear Ht.
intro Ht.
exists (S (R tn)).
intros tQ Hhas.
elim (has_del_intro Hhas).
clear Hhas.
intros Hne' Hhas.
elim (has_ins_intro Hhas).
clear Hhas.
intro Heq.
rewrite -> (trv7 tQ (R tn) Heq).
apply trv4.
clear Hhas.
intro Ht'.
elim Ht'.
clear Ht'.
intros Hne'' Hhas.
apply (trv5 (trv5 (trv4 (R tn)) Ht) (HQhtp tQ Hhas)).
clear Ht.
intro Ht.
exists (S t').
intros tQ Hhas.
elim (has_del_intro Hhas).
clear Hhas.
intros Hne' Hhas.
elim (has_ins_intro Hhas).
clear Hhas.
intro Heq.
rewrite <- (trv7 tQ (R tn) Heq).
apply trv6.
exact Ht.
clear Hhas.
intro Ht'.
elim Ht'.
clear Ht'.
intros Hne'' Hhas.
apply (trv5 (trv4 t') (HQhtp tQ Hhas)).
apply classic.

intros tQ RQ IQ Hget'.
elim (get_del_intro Hget').
clear Hget'.
intros Hne' Hget'.
elim (get_ins_intro Hget').
clear Hget'.
intro Ht.
elim Ht.
clear Ht.
intros Heq Hval.
inversion Hval.
clear Hval.
rename H0 into HvalR.
rename H1 into HvalI.
exists p.
exists A.
exists G.
unfold del_w.
apply get_del_elimi.
split.
exact Hne'.
unfold ins_w.
apply get_ins_elimi.
left.
split.
rewrite <- HvalR.
exact Heq.
trivial.
clear Hget'.
intro Ht.
elim Ht.
clear Ht.
intros Hne'' Hget'.
elim (HQex tQ RQ IQ Hget').
intros pQ Ht.
elim Ht.
clear Ht.
intros AQ Ht.
elim Ht.
clear Ht.
intros GQ Hget''.
exists pQ.
exists AQ.
exists GQ.
unfold del_w.
apply get_del_elimi.
split.
exact Hne'.
unfold ins_w.
apply get_ins_elimi.
right.
split.
exact Hne''.
exact Hget''.

intros tQ RQ IQ pQ AQ GQ Ht.
elim Ht.
clear Ht.
intros Hz Hq.
elim (get_del_intro Hz).
clear Hz.
intros Hne' Hz.
elim (get_del_intro Hq).
clear Hq.
intros Ht Hq.
clear Ht.
elim (get_ins_intro Hz).
clear Hz.
intro Ht.
elim Ht.
clear Ht.
intros Heq Hval.
inversion Hval.
clear Hval.
rename H0 into Hvalp.
rename H1 into HvalA.
rename H2 into HvalG.
elim (get_ins_intro Hq).
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Ht Hval.
clear Ht.
inversion Hval.
clear Hval.
rename H0 into HvalR.
rename H1 into HvalI.
rewrite <- Hvalp.
rewrite <- HvalA.
rewrite <- HvalG.
rewrite <- HvalI.
exact HIir.
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Hne'' Hget'.
rewrite -> Heq in Hne''.
discriminate.
clear Hz.
intro Ht.
elim Ht.
clear Ht.
intros Hne'' Hget'.
elim (get_ins_intro Hq).
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Heq Hget''.
rewrite -> Heq in Hne''.
discriminate.
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Ht Hget''.
clear Ht.
apply HQir with tQ RQ.
split.
exact Hget'.
exact Hget''.

intros tQ RQ IQ pQ AQ GQ M' M'' Ht Hinv' Hp' Ha'.
elim Ht.
clear Ht.
intros Hz Hq.
elim (get_del_intro Hz).
clear Hz.
intros Hne' Hz.
elim (get_del_intro Hq).
clear Hq.
intros Ht Hq.
clear Ht.
elim (get_ins_intro Hz).
clear Hz.
intro Ht.
elim Ht.
clear Ht.
intros Heq Hval.
inversion Hval.
clear Hval.
rename H0 into Hvalp.
rename H1 into HvalA.
rename H2 into HvalG.
elim (get_ins_intro Hq).
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Ht Hval.
clear Ht.
inversion Hval.
clear Hval.
rename H0 into HvalR.
rename H1 into HvalI.
rewrite <- Hvalp.
rewrite <- HvalR.
apply HIas with M'.
exact Hinv'.
rewrite -> Hvalp.
rewrite -> HvalR.
exact Hp'.
rewrite -> HvalA.
rewrite -> HvalR.
exact Ha'.
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Hne'' Hget'.
rewrite -> Heq in Hne''.
discriminate.
clear Hz.
intro Ht.
elim Ht.
clear Ht.
intros Hne'' Hget'.
elim (get_ins_intro Hq).
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Heq Hget''.
rewrite -> Heq in Hne''.
discriminate.
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Ht Hget''.
clear Ht.
apply (HQas tQ RQ IQ pQ AQ GQ M' M'' (conj Hget' Hget'') Hinv' Hp' Ha').


apply queue_nonin_del.
apply queue_nonin_ins.
exact HQni.
exact HPni.

apply HQir with t R'.
split.
exact HIgez.
exact Hget.
exact Hinv.
apply (HPlo t p' A' G' R' I'' M (conj HIgez Hget) (HIlo R M Hinv Hp)).


rewrite -> (HQtn t R' I'' Hget).
intro Hhas.
elim (has_del_intro Hhas).
clear Hhas.
intros Hne' Hhas.
elim (trv10 Hne').
trivial.

intros tQ pQ AQ GQ RQ IQ M' Ht Hg.
elim Ht.
clear Ht.
intros Hz Hq.
elim (get_del_intro Hz).
clear Hz.
intros Hne' Hz.
elim (get_del_intro Hq).
clear Hq.
intros Ht Hq.
clear Ht.
elim (get_ins_intro Hz).
clear Hz.
intro Ht.
elim Ht.
clear Ht.
intros Heq Hval.
inversion Hval.
clear Hval.
rename H0 into Hvalp.
rename H1 into HvalA.
rename H2 into HvalG.
elim (get_ins_intro Hq).
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Ht Hval.
clear Ht.
inversion Hval.
clear Hval.
rename H0 into HvalR.
rename H1 into HvalI.
rewrite <- Hvalp.
rewrite <- HvalR.
apply HIas with M.
exact Hinv.
exact Hp.
inversion_clear HPni.
rename H into HPnig.
rename H0 into HPnia.
apply HPnia with t p' A' G' R' I''.
split.
exact HIgez.
exact Hget.
exact Hg.
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Hne'' Hget'.
rewrite -> Heq in Hne''.
discriminate.
clear Hz.
intro Ht.
elim Ht.
clear Ht.
intros Hne'' Hget'.
elim (get_ins_intro Hq).
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Heq Hget''.
rewrite -> Heq in Hne''.
discriminate.
clear Hq.
intro Ht.
elim Ht.
clear Ht.
intros Ht Hget''.
clear Ht.

assert (pQ (RQ, M)).
apply (HPlo tQ pQ AQ GQ RQ IQ M (conj Hget' Hget'') (HIlo R M Hinv Hp)).


assert (AQ (RQ, M) M').
inversion_clear HQni.
rename H0 into HQni.
apply HQni with t tQ p' pQ A' G' GQ R' I'' IQ.
split.
apply trv10.
exact Hne'.
split.
exact HIgez.
split.
exact Hget.
split.
exact Hget'.
exact Hget''.
exact Hg.

apply (HQas tQ RQ IQ pQ AQ GQ M M' (conj Hget' Hget'') Hinv H H0).


apply prog_nonin_del with p' I''.
apply queue_nonin_ins.
exact HQni.
exact HPni.
split.
unfold ins_w.
apply get_ins_elimi.
right.
split.
exact Hne.
exact HIgez.
unfold ins_w.
apply get_ins_elimi.
right.
split.
exact Hne.
exact Hget.

exact Hget.

(* jd *)
rename H3 into HIpc.
rename H6 into HIlo.
rename H8 into HIas.
rename H9 into HIgu.
rename H10 into HIir.
rewrite <- H7 in Hstep.
rewrite <- H7 in HP.
rewrite <- H7 in HI.
clear Y0 p0 g0 A0 G0 I H4 H5 H H0 H1 H2 H7.
inversion Hstep.
rewrite <- H0 in Hstep.
rename H4 into HSge.
clear C0 S Q0 f0 P' H H1 H2 H3 H0.
exists (Inv,Y,Z,(p',g',A',G')).
apply P_infer_1.
exact HC.
exact HQ.
apply HCir with f.
split.
exact HIir.
exact HSge.
exact Hinv.
apply HIpc.
exact Hinv.
exact Hp.
exact HPtn.

intros tQ pQ AQ GQ RQ IQ M' Ht Hg.
apply (HPlo tQ pQ AQ GQ RQ IQ M' Ht (HIlo R M M' Hinv Hp Hg)).


apply prog_nonin_seq with A G R.
exact HPni.
intros M' M'' Ha.
apply HIas with M; auto.
intros M' M'' Hg.
apply HIgu with M; auto.


(* exit *)
rename H6 into HIlo.
rewrite <- H2 in HP.
rewrite <- H2 in Hstep.
rewrite <- H2 in HI.
clear Y0 p0 g0 A0 G0 H0 H1 H H3 H4 H5 I H2.
inversion Hstep.
rewrite <- H0 in Hstep.
rename H4 into HIge.
clear C0 H R0 H1 M0 H2 Q0 H3 P' H0.
elim (HQex t R' I' HIge).
intros p' Ht.
elim Ht.
clear Ht.
intros A' Ht.
elim Ht.
clear Ht.
intros G' HQqu.
exists (Inv,Y,del_w Z t,(p',G',A',G')).
apply P_infer_1.
exact HC.
apply Q_infer_1.

intros tQ RQ IQ Ht.
apply HQtn with IQ.
elim (get_del_intro Ht).
clear Ht.
intros  Ht Ht'.
exact Ht'.

elim HQht.
intros t' Ht.
exists t'.
intros t'' Ht'.
apply Ht.
clear Ht.
elim (has_del_intro Ht').
clear Ht'.
intros Ht Ht'.
exact Ht'.

intros t' RQ IQ Ht.
elim (get_del_intro Ht).
clear Ht.
intros Ht Ht'.
elim (HQex t' RQ IQ Ht').
clear Ht'.
intros pQ Ht'.
elim Ht'.
clear Ht'.
intros AQ Ht'.
elim Ht'.
clear Ht'.
intros GQ Ht'.
exists pQ.
exists AQ.
exists GQ.
unfold del_w.
apply get_del_elimi.
split.
exact Ht.
exact Ht'.

intros t' RQ IQ pQ AQ GQ Ht.
apply HQir with t' RQ.
elim Ht.
clear Ht.
intros Hz Hq.
elim (get_del_intro Hz).
clear Hz.
intros Ht Ht'.
split.
exact Ht'.
clear Ht Ht'.
elim (get_del_intro Hq).
clear Hq.
intros Ht Ht'.
exact Ht'.

intros tQ RQ IQ pQ AQ GQ M' M'' Ht Hinv' Hp' Ha'.
assert (get Z tQ (pQ, AQ, GQ) /\ get Q tQ (RQ, IQ)).
elim Ht.
clear Ht.
intros Ht Ht'.
elim (get_del_intro Ht).
elim (get_del_intro Ht').
clear Ht Ht'.
intros Ht Hq Ht' Hz.
split.
exact Hz.
exact Hq.
apply (HQas tQ RQ IQ pQ AQ GQ M' M'' H Hinv' Hp' Ha').


apply queue_nonin_del.
exact HQni.

apply HQir with t R'.
split.
exact HQqu.
exact HIge.
exact Hinv.
apply (HPlo t p' A' G' R' I' M (conj HQqu HIge) (HIlo R M Hinv Hp)).


rewrite -> (HQtn t R' I' HIge).
intro Hhas.
elim (has_del_intro Hhas).
clear Hhas.
intros Hne Hhas.
elim (trv10 Hne).
trivial.

intros t' pQ AQ GQ RQ IQ M' Ht Hg.

assert (get Z t' (pQ, AQ, GQ) /\ get Q t' (RQ, IQ)).
elim Ht.
clear Ht.
intros Hz Hq.
elim (get_del_intro Hz).
clear Hz.
intros Ht Ht'.
split.
exact Ht'.
clear Ht Ht'.
elim (get_del_intro Hq).
clear Hq.
intros Ht Ht'.
exact Ht'.

assert (pQ (RQ, M)).
apply (HPlo t' pQ AQ GQ RQ IQ M H (HIlo R M Hinv Hp)).

assert (AQ (RQ, M) M').
elim Ht.
clear Ht.
intros Hz Hq.
elim (get_del_intro Hz).
clear Hz.
intros Hne Hz.
elim (get_del_intro Hq).
clear Hq.
intros Hne' Hq.
clear Hne'.
inversion HQni.
rename H1 into HQnip.
clear Z0 Q0 H2 H3.
apply HQnip with t t' p' pQ A' G' GQ R' I' IQ.
clear HQnip.
split.
apply trv10.
exact Hne.
split.
exact HQqu.
split.
exact HIge.
split.
exact Hz.
exact Hq.
exact Hg.
apply (HQas t' RQ IQ pQ AQ GQ M M' H Hinv H0 H1).


apply prog_nonin_del with p' I'.
exact HQni.
split.
exact HQqu.
exact HIge.

exists (Inv,Y,Z,(p,g,A,G)).
apply P_infer_1.
exact HC.
exact HQ.
exact HI.
exact Hinv.
exact Hp.
exact HPtn.
exact HPlo.
exact HPni.
Qed.

(* Soundness Theorem *)
Theorem Soundness : forall n S P, P_infer S P -> exists P', Step_n Step n P P'.
Proof.
apply (Soundness Step P_infer Progress Preservation).
Qed.


