(***************************************************************)
(* Coq code for PCAP                                           *)
(*                                                             *)
(* Map Implementation                                          *)
(*    nat -> B : Set                                           *)
(*    Function                                                 *)
(***************************************************************)

(* $Id: FuncB_MAP.v,v 1.1.1.1 2006/12/07 16:43:24 ssg05 Exp $ *)

Require Import bnat.

Require Export MAP.

Require Import extac.

Set Implicit Arguments.
Unset Strict Implicit.

Module FuncMap (BDef : ONE_SET) <: MAPDEF.

Definition A := nat.

Definition beq : A -> A -> bool := beq_nat.

Lemma beq_true_eq : forall a a' : A, 
  beq a a' = true -> a = a'.
Proof.
  apply beq_true_eq.
Qed.

Lemma beq_false_neq : forall a a' : A,
    beq a a' = false -> a <> a'.
Proof.
  apply beq_false_neq.
Qed.

Lemma eq_beq_true : forall a a' : A, 
    a = a' -> beq a a' = true.
Proof.
  apply eq_beq_true.
Qed.

Lemma neq_beq_false : forall a a' : A, 
    a <> a' -> beq a a' = false.
Proof.
  apply neq_beq_false.
Qed.

Lemma eq_dec : forall a a' : A, {a=a'}+{a<>a'}.
Proof.
  apply nat_eq_dec.
Qed.

Definition B := BDef.B.

(***********************)
(** map                *)
(***********************)

Definition map := A -> option B.

Definition emp_map : map := fun _ => None.

Definition sig_map (a' : A) (b : B) : map := 
  fun a : A =>
    match (beq a a') with
      | true => Some b
      | false => None
    end.

(***********************)
(** set                *)
(** get                *)
(***********************)

Definition get (m : map) (a : A) : option B := m a.

Lemma get_dec : forall (m : map) (a : A),
  {exists b, get m a = Some b} 
  + {get m a = None}.
Proof.
  intros m a; unfold get.
  destruct (m a).
  left; exists b; trivial.
  right; trivial.
Qed.

Definition set (m : map) (a : A) (b : B) :=
  fun a' =>
    match (beq a a') with
      | true => Some b
      | false => m a'
    end.

Definition upd (m : map) (a : A) (b : B) : option map :=
  match (m a) with
    | Some b' => Some (fun a' => match (beq a a') with
                                     | true => Some b
                                     | false => m a'
                                   end)
    | None => None
  end.

Lemma set_sem : forall m a a' b, 
  match (beq a a') with
    | true => get (set m a b) a = Some b
    | false => get (set m a b) a' = get m a'
  end.
Proof.
  intros m a a' b.
  unfold set, get, beq.
  simplbnat.
  decbeqnat a a'; trivial.
Qed.

Lemma upd_sem : forall m a b, 
  match get_dec m a with 
    | left _ => upd m a b = Some (set m a b)
    | right _ => upd m a b = None
  end.
Proof.
  intros.
  destruct (get_dec m a) as [[b' Hin] | Hnin].
  unfold upd, set, get in * .
  rewrite Hin; simpl.
  trivial.
  unfold upd, set, get in * .
  rewrite Hnin; simpl.
  trivial.
Qed.

Lemma emp_sem : forall a : A, get emp_map a = None.
Proof.
  intros a; unfold get, emp_map; trivial.
Qed.

Lemma sig_sem : forall (a a' : A) (b : B),
  get (sig_map a' b) a = 
  match (beq a a') with
    | true => Some b
    | false => None
  end.
Proof.
  intros; unfold get, sig_map; trivial.
Qed.

(************************)
(** merge               *)
(************************)

Definition merge (m1 m2 : map) a :=
  match m1 a with
  | None => m2 a
  | Some b => Some b
  end.

Lemma merge_sem : forall (m1 m2 : map) (a : A),
  get (merge m1 m2) a 
  = match get m1 a, get m2 a with
      | Some b1, Some b2 => Some b1
      | Some b1, None => Some b1
      | None, Some b2 => Some b2
      | None, None => None
    end.
Proof.
  unfold get, merge; intros m1 m2 a.
  destruct (m1 a); destruct (m2 a); trivial.
Qed.

Ltac psi_get label :=
  match goal with
    | H : get ?psi ?f = Some ?a
      |- _ =>
      decbeqnat f label;
      [ unfold psi, get in H; simplbnat; simpl in H;
        substbnat f;
          (match type of H with 
             | None = Some a  => discriminate H
             | Some ?b = Some a => inj_hyp H; subst a
           end)
        | idtac]
    | _ => idtac
  end.

Ltac psi_get_not :=
  match goal with
    | H : get ?psi ?f = Some ?a |- _ =>
      unfold psi, get in H; simplbnat; simpl in H;
        (match type of H with
           | None = Some a => discriminate H
           | _ => fail
         end)
    | _ => fail
  end.

End FuncMap.

Unset Implicit Arguments.
