(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* Tiny Machine                                                *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

(* $Id: LIN_ORDER.v,v 1.1.1.1 2006/12/07 16:43:24 ssg05 Exp $ *)

Require Import extac.

Set Implicit Arguments.
Unset Strict Implicit.

Module Type LIN_ORDER.

  Parameter A : Set.
  Parameter lt : A -> A -> Prop.
  
  Definition gt a b := lt b a.

  Parameter compare_dec : 
    forall a b, {lt a b} + {a = b} + {lt b a}.

  Parameter eq_not_lt : forall a b, a = b -> lt a b -> False.
  Parameter lt_not_eq : forall a b, lt a b -> a = b -> False.
  Parameter lt_not_gt : forall a b, lt a b -> lt b a -> False.
  Parameter lt_trans : forall a b c, lt a b -> lt b c -> lt a c.

  Parameter eq_dec : forall (a b : A), {a = b} + {a <> b}.

  Parameter beq : A -> A -> bool.

  Axiom beq_true_eq : forall a a' : A, 
    beq a a' = true -> a = a'.

  Axiom beq_false_neq : forall a a' : A,
    beq a a' = false -> a <> a'.

  Axiom eq_beq_true : forall a a' : A, 
    a = a' -> beq a a' = true.

  Axiom neq_beq_false : forall a a' : A, 
    a <> a' -> beq a a' = false.

  (* Fixme: other linear-order-type lemmas required *)

End LIN_ORDER.

Module LinLib (Lin : LIN_ORDER).
Import Lin.

(* Lemma eq_dec_lt : forall (a b : A) (ltprf : lt a b),  *)
(*   eq_dec a b = right (a=b) (lt_not_eq a b ltprf). *)
(* Proof. *)
(*   intros a b ltprf. *)
(*   destruct (eq_dec a b) as [Heq | Hneq]. *)
(*   destruct (lt_not_eq a b ltprf). *)
(*   apply Heq. *)
(* (* need proof_irrelevant *)   *)
(* Qed. *)

Ltac eqdec_eq :=  
  match goal with
    | H: context [eq_dec ?a ?a] |- _ 
      =>  let He := fresh "He" in
          let Hne := fresh "Hne" in
            (destruct (eq_dec a a) as [He | Hne ]; 
              [ clear He | discri ])
    | |- context [eq_dec ?a ?a]
      =>  let He := fresh "He" in
          let Hne := fresh "Hne" in
            (destruct (eq_dec a a) as [He | Hne ]; 
              [ clear He | discri ])
    | _ => fail 1 "no eq_dec term found "
  end.

Ltac eqdec_neq :=  
  match goal with
    | H: context [eq_dec ?a ?b] |- _ 
      =>  let He := fresh "He" in
          let Hne := fresh "Hne" in
            (destruct (eq_dec a b) as [He | Hne];
              [
                match goal with 
                  | Hlt : lt a b |- _ =>
                    destruct (eq_not_lt a b He Hlt)
                  | Hlt : lt b a |- _ => 
                    destruct (eq_not_lt b a He Hlt)
                  | Hneq : not (eq a b) |- _ =>
                    destruct (Hneq He)
                  | Hneq : not (eq b a) |- _ =>
                    destruct (Hneq (sym_eq He))
                  | _ => fail 1 "xx"
                end
              | idtac ])
    | |- context [eq_dec ?a ?b]
      =>  let He := fresh "He" in
          let Hne := fresh "Hne" in
            (destruct (eq_dec a b) as [He | Hne];
              [
                match goal with 
                  | Hlt : lt a b |- _ =>
                    destruct (eq_not_lt a b He Hlt)
                  | Hlt : lt b a |- _ => 
                    destruct (eq_not_lt b a He Hlt)
                  | Hneq : not (eq a b) |- _ =>
                    destruct (Hneq He)
                  | Hneq : not (eq b a) |- _ =>
                    destruct (Hneq (sym_eq He))
                  | _ => fail
                end
              | idtac])
    | _ => fail 1 "no eq_dec term found "
  end.


Lemma beq_refl : forall a, beq a a = true.
Proof.
  intros a; apply eq_beq_true; trivial.
Qed.

Lemma beq_sym : forall a a', beq a a' = beq a' a.
Proof.
  intros a a'; destruct (eq_dec a a').
  subst a; trivial.
  rewrite (neq_beq_false n).
  rewrite (neq_beq_false (sym_not_eq n)).
  trivial.
Qed.


Ltac simplbeq :=
  match goal with 
    
    | H : beq ?r ?r = false |- _ =>
        rewrite (beq_refl r) in H;
          discriminate H; 
            simplbeq
        
    | H : context [ beq ?r ?r ] |- _ =>
        rewrite (beq_refl r) in H;
          simplbeq
   
    | |- context [ beq ?r ?r] =>
        rewrite (beq_refl r);
          simplbeq

    | H : beq ?rx ?ry = ?f
      |- context [beq ?rx ?ry] => 
        rewrite H; simplbeq
    | H : beq ?rx ?ry = ?f
      |- context [beq ?ry ?rx] => 
        rewrite (beq_sym rx ry) in H; 
          rewrite H; 
            simplbeq

    | H : beq ?rx ?ry = ?f,
      H1 : context [beq ?rx ?ry] |- _ =>
        rewrite H in H1; simplbeq
    | H : beq ?rx ?ry = ?f,
      H1 : context [beq ?ry ?rx] |- _ =>
        rewrite (beq_sym rx ry) in H; 
          rewrite H in H1;
            simplbeq

    | H : ?x = ?y 
      |- context [(beq ?x ?y)] => 
        rewrite (eq_beq_true H); simplbeq
    | H : ?y = ?x 
      |- context [(beq ?x ?y)] => 
        rewrite (eq_beq_true (sym_eq H));
          simplbeq
    | H : ?x = ?y,
      H1 : context [(beq ?x ?y)] |- _ => 
        rewrite (eq_beq_true H) in H1; simplbeq
    | H : ?y = ?x,
      H1 : context [(beq ?x ?y)] |- _ => 
        rewrite (eq_beq_true (sym_eq H)) in H1;
          simplbeq

    | H : ?x <> ?y 
      |- context [(beq ?x ?y)] => 
        rewrite (neq_beq_false H); simplbeq
    | H : ?y <> ?x 
      |- context [(beq ?x ?y)] => 
        rewrite (neq_beq_false (sym_not_eq H));
          simplbeq
    | H : ?x <> ?y,
      H1 : context [(beq ?x ?y)] |- _ => 
        rewrite (neq_beq_false H) in H1; simplbeq
    | H : ?y <> ?x,
      H1 : context [(beq ?x ?y)] |- _ => 
        rewrite (neq_beq_false (sym_not_eq H)) in H1;
          simplbeq

    | H : beq ?rx ?ry = true
      |- _ =>
        let Heq := fresh "Heq" in
        (generalize (beq_true_eq H); 
          intro Heq; subst rx); simplbeq

    | _ => idtac

  end.

(* TODO : more lemma and more tactics *)
End LinLib.

Unset Implicit Arguments.

