(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* Map                                                         *)
(*    using coq list                                           *)
(*                                                             *)
(***************************************************************)

(* $Id: ListB_MAPT.v,v 1.1.1.1 2006/12/07 16:43:24 ssg05 Exp $ *)

Require Import Coq.Bool.Bool.

Require Import bnat.
Require Export optionT.
Require Export prodT.
Require Export ListT.
Require Export MAPT.

Require Import extac.

Open Local Scope listt_scope.


Set Implicit Arguments.
Unset Strict Implicit.

Module ListMapT (BDef : ONE_TYPE) <: MAPDEF.

Definition A := nat.

Definition beq : A -> A -> bool := beq_nat.

Lemma beq_true_eq : forall a a' : A, 
  beq a a' = true -> a = a'.
Proof.
  apply beq_true_eq.
Qed.

Lemma beq_false_neq : forall a a' : A,
    beq a a' = false -> a <> a'.
Proof.
  apply beq_false_neq.
Qed.

Lemma eq_beq_true : forall a a' : A, 
    a = a' -> beq a a' = true.
Proof.
  apply eq_beq_true.
Qed.

Lemma neq_beq_false : forall a a' : A, 
    a <> a' -> beq a a' = false.
Proof.
  apply neq_beq_false.
Qed.

Lemma eq_dec : forall a a' : A, {a=a'}+{a<>a'}.
Proof.
  apply nat_eq_dec.
Qed.

Definition B := BDef.B.

(***********************)
(** map                *)
(***********************)

Definition map := listT (prodT A B).

Definition emp_map : map := nilT.

Definition sig_map (a : A) (b : B) : map := 
  (pairT a b) :: nilT.

(***********************)
(** set                *)
(** get                *)
(***********************)

Fixpoint get (m : map) (a : A) {struct m} : optionT B :=
  match m with 
    | nilT => NoneT
    | (pairT a' b') :: m' =>
      match beq a a' with
        | true => SomeT b'
        | false => get m' a
      end
  end.

Lemma get_dec :
  forall (m : map) (a : A),
    {exists b, get m a = SomeT b} + {get m a = NoneT}.
Proof.
  intros m a.
  destruct (get m a).
  left.
  exists b.
  trivial.
  right; trivial.
Qed.

Fixpoint set (m : map) (a : A) (b : B) {struct m} : map :=
  match m return map with
    | nilT => (pairT a b) :: nilT
    | (pairT a' b') :: m' =>
      match beq a a' with
        | true => (pairT a b) :: m'
        | false => (pairT a' b') :: (set m' a b)
      end
  end.

Fixpoint upd (m : map) (a : A) (b : B) {struct m} : optionT map :=
  match m return optionT map with
      | nilT => NoneT
      | (pairT a' b') :: m' =>
        match beq a a' with
          | true => SomeT ((pairT a b) :: m')
          | false => 
              let m'' := (upd m' a b) in
                match m'' with 
                  | NoneT => NoneT
                  | SomeT m' => SomeT ((pairT a' b') :: m')
                end
        end
  end.

(*
Lemma eq_not_lt : forall a b, a = b -> lt a b -> False.
Proof.
  intros a b.
  intros H1 H2.
  subst a.
  apply (lt_irrefl b H2).
Qed.

Ltac eqdec_eq :=  
  match goal with
    | H: context [eq_dec ?a ?a] |- _ 
      =>  let He := fresh "He" in
          let Hne := fresh "Hne" in
            (destruct (eq_dec a a) as [He | Hne ]; 
              [ clear He | discri ])
    | |- context [eq_dec ?a ?a]
      =>  let He := fresh "He" in
          let Hne := fresh "Hne" in
            (destruct (eq_dec a a) as [He | Hne ]; 
              [ clear He | discri ])
    | _ => fail 1 "no eq_dec term found "
  end.

Ltac eqdec_neq :=  
  match goal with
    | H: context [eq_dec ?a ?b] |- _ 
      =>  let He := fresh "He" in
          let Hne := fresh "Hne" in
            (destruct (eq_dec a b) as [He | Hne];
              [
                match goal with 
                  | Hlt : lt a b |- _ =>
                    destruct (eq_not_lt a b He Hlt)
                  | Hlt : lt b a |- _ => 
                    destruct (eq_not_lt b a He Hlt)
                  | Hneq : not (eq a b) |- _ =>
                    destruct (Hneq He)
                  | Hneq : not (eq b a) |- _ =>
                    destruct (Hneq (sym_eq He))
                  | _ => fail 1 "xx"
                end
              | idtac ])
    | |- context [eq_dec ?a ?b]
      =>  let He := fresh "He" in
          let Hne := fresh "Hne" in
            (destruct (eq_dec a b) as [He | Hne];
              [
                match goal with 
                  | Hlt : lt a b |- _ =>
                    destruct (eq_not_lt a b He Hlt)
                  | Hlt : lt b a |- _ => 
                    destruct (eq_not_lt b a He Hlt)
                  | Hneq : not (eq a b) |- _ =>
                    destruct (Hneq He)
                  | Hneq : not (eq b a) |- _ =>
                    destruct (Hneq (sym_eq He))
                  | _ => fail
                end
              | idtac])
    | _ => fail 1 "no eq_dec term found "
  end.
*)

Lemma set_sem_old : forall m a a' b, 
  match (beq a a') with
    | true => get (set m a b) a = SomeT b
    | false => get (set m a b) a' = get m a'
  end.
Proof.
  intros m a a0 b.
  induction m as [|(a',b') m' IHm']; 
    unfold get, set, beq in * ; simplbnat.
  decbeqnat a0 a; trivial.
  fold get in * .
  fold set in * .
  decbeqnat a a0; 
    decbeqnat a a'; 
    decbeqnat a0 a'; 
    simpl; unfold beq; simplbnat; trivial.
  discribnat.
Qed.

Lemma set_sem: forall m a a' b,
  get (set m a b) a' =
  match (beq a a') with
    | true => SomeT b
    | false => get m a'
  end.
Proof.
  intros; assert (H := set_sem_old).
  substH H with (H m a a' b).
  assert (H1:=beq_true_eq); substH H1 with (H1 a a').
  assert (H2:=beq_false_neq); substH H2 with (H2 a a').
  destruct beq.
  assert (H3 := refl_equal).
  substH H3 with (H3 bool true).
  substH H1 with (H1 H3).
  rewrite H1; rewrite H1 in H; trivial.
  trivial.
Qed.


Lemma upd_sem_0 : forall m a b b0, 
  (get m a = SomeT b -> upd m a b0 = SomeT (set m a b0))
  /\ (get m a = NoneT -> upd m a b0 = NoneT).
Proof.
  intros.
  split.
  intros.
  induction m as [|(a',b') m' IHm'].
  discriminate.
  simpl in * ; unfold beq in * .
  decbeqnat a a'; trivial.
  rewrite (IHm' H).
  trivial.
  induction m as [|(a',b') m' IHm'].
  trivial.
  simpl in * ; unfold beq in * .
  decbeqnat a a'; trivial.
  intro Hf; discriminate Hf.
  intros H.
  rewrite (IHm' H).
  trivial.
Qed.

Lemma upd_sem_old : forall m a b,
    match get_dec m a with 
      | left _ => upd m a b = SomeT (set m a b)
      | right _ => upd m a b = NoneT
    end.
Proof.
  intros.
  destruct (get_dec m a) as [[b' Hi] | Hni].
  assert (H := proj1 (upd_sem_0 m a b' b)).
  tauto.
  assert (H := proj2 (upd_sem_0 m a b b)).
  tauto.
Qed.

Lemma upd_sem : forall m a b,
  upd m a b =
  match get_dec m a with
    | left _ => SomeT (set m a b)
    | right _ => NoneT
  end.
Proof.
  intros.
  assert (H := upd_sem_old).
  substH H with (H m a b).
  destruct get_dec; trivial.
Qed.

Lemma emp_sem : 
  forall a : A, get emp_map a = NoneT.
Proof.
  intros; unfold get, emp_map; trivial.
Qed.

Lemma sig_sem : forall (a a' : A) (b : B),
  get (sig_map a b) a' = 
  match (beq a a') with
    | true => SomeT b
    | false => NoneT
  end.
Proof.
  unfold get, sig_map. intros a a' b.
  destruct (eq_dec a a').
  rewrite e; trivial.
  rewrite (neq_beq_false n).
  rewrite (neq_beq_false (sym_not_eq n)).
  trivial.
Qed.

(************************)
(** merge               *)
(************************)

Fixpoint merge (m1 m2 : map) {struct m1} : map := 
  match m1 with 
    | nilT => m2
    | (pairT a b) :: m1' => (pairT a b) :: (merge m1' m2)
  end.

Lemma merge_sem : forall (m1 m2 : map) (a : A),
  get (merge m1 m2) a 
  = match get m1 a, get m2 a with
      | SomeT b1, SomeT b2 => SomeT b1
      | SomeT b1, NoneT => SomeT b1
      | NoneT, SomeT b2 => SomeT b2
      | NoneT, NoneT => NoneT
    end.
Proof.
  intros m1 m2 a. unfold merge.
  induction m1 as [| (a1, b1) m1' IHm1'].
  unfold get at 2.
  destruct (get m2 a); trivial.
  unfold merge in *; fold merge in *.
  unfold get; fold get.
  destruct (eq_dec a a1) as [Heq | Hneq].
  subst a1.
  unfold beq; simplbnat.
  destruct (get m2 a).
  trivial.
  trivial.
  unfold beq; simplbnat.
  assumption.
Qed.

End ListMapT.

Unset Implicit Arguments.
