(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* SETMap Implementation (List)                                *)
(*    with extensionality                                      *)
(*    nat -> B : Set                                           *)
(***************************************************************)

(* $Id: ListB_SETMAP.v,v 1.1.1.1 2006/12/07 16:43:24 ssg05 Exp $ *)


Require Import Coq.Lists.List.
Require Import Coq.Logic.Classical_Prop.
Require Import bnat.
Require Import SETMAP.
Require Import extac.

Module ListMap (BDef : ONE_SET) <: MAPDEF.

Import BDef.

Definition A := nat.

Definition beq : A -> A -> bool := beq_nat.

Lemma beq_true_eq : forall a a' : A, 
  beq a a' = true -> a = a'.
Proof.
  apply beq_true_eq.
Qed.

Lemma beq_false_neq : forall a a' : A,
    beq a a' = false -> a <> a'.
Proof.
  apply beq_false_neq.
Qed.

Lemma eq_beq_true : forall a a' : A, 
    a = a' -> beq a a' = true.
Proof.
  apply eq_beq_true.
Qed.

Lemma neq_beq_false : forall a a' : A, 
    a <> a' -> beq a a' = false.
Proof.
  apply neq_beq_false.
Qed.

Lemma eq_dec : forall a a' : A, {a=a'}+{a<>a'}.
Proof.
  apply nat_eq_dec.
Qed.

Definition B := BDef.B.

(*
   .....
*)

(*
 * definition of ordered sets
*)

(* strict lower bound of a list *)

(* raw_list *)
Definition rlist : Set := list (prod A B).

Fixpoint lb (a : A) (rl : rlist) {struct rl} : Prop :=
  match rl with 
    | nil => True
    | (a', b') :: rl' => (lt a a') /\ (lb a rl')
  end.

Lemma lb_trans : 
  forall rl a1 a2, lb a2 rl -> lt a1 a2 -> lb a1 rl.
Proof.
  unfold lb.
  induction rl as [|(a, b) rl' IHrl']; fold lb in *; trivial.
  intros a1 a2 [H1 H2] H.
  split.
  apply lt_trans with a2; trivial.
  apply (IHrl' a1 a2); trivial.
Qed.

Fixpoint is_orderset (rl : rlist) : Prop :=
  match rl with 
    | nil => True
    | (a', b') :: rl' => (lb a' rl') /\ (is_orderset rl')
  end.

(***********************)
(** map                *)
(***********************)

Inductive listset : Set := 
  | listset_con : forall lst : rlist, 
                 is_orderset lst -> listset.

Definition map : Set := listset.

Definition lst (m : listset) := let (lst, _) := m in lst.

Definition prf (m : listset) := 
  let (lst, prf) as m return (is_orderset (lst m)) 
    := m in prf.

Definition emp' : rlist := nil.

Lemma is_orderset_emp' : is_orderset emp'.
Proof.
  red; apply I.
Qed.

Definition emp_map : map := listset_con emp' is_orderset_emp'.

Definition sig' (a : A) (b : B) := (a, b) :: nil.

Lemma is_orderset_sig' : forall a b, is_orderset (sig' a b).
Proof.
  unfold is_orderset, lb; intros; split; apply I.
Qed.

Definition sig_map (a : A) (b : B) := 
  listset_con (sig' a b) (is_orderset_sig' a b).

(***********************)
(** get                *)
(***********************)

Fixpoint get' (rl : rlist) (a : A) {struct rl} : option B :=
  match rl with 
    | nil => None
    | (a', b') :: rl' =>
        match (beq a a') with
          | true => Some b'
          | false => get' rl' a
        end
  end.

Lemma lb_notin : 
  forall rl a, lb a rl -> get' rl a = None.
Proof.
  unfold get', lb.
  induction rl as [|(a', b') rl' IHrl']; trivial.
  intros a [H1 H2].
  destruct (eq_dec a a') as [Heq | Hneq].
  subst a'.
  destruct (lt_irrefl a H1).
  unfold beq.
  simplbnat.
  apply IHrl'.
  apply H2.
Qed.

Lemma in_notlb :
  forall rl a b, get' rl a = Some b -> ~ lb a rl.
Proof.
  unfold not;intros.
  rewrite (lb_notin rl a H0) in H.
  discriminate.
Qed.

Definition get (m : map) (a : A) := get' (lst m) a.

Lemma get_dec :
  forall (m : map) (a : A),
    {exists b, get m a = Some b} + {get m a = None}.
Proof.
  intros m a.
  destruct (get m a).
  left.
  exists b.
  trivial.
  right; trivial.
Qed.

(***********************)
(** set                *)
(***********************)

Fixpoint set' (a : A) (b : B) (rl : rlist) 
  {struct rl}: rlist :=
  match rl with
    | nil => (a, b) :: nil
    | (a', b') :: rl' =>
      match blt_nat a a' with
        | true  => (a, b) :: rl
        | false =>           
          match beq_nat a a' with 
            | true => (a', b) :: rl'
            | false => (a', b') :: (set' a b rl')
          end

      end
  end.

Lemma get'_set' : forall rl a b,
  get' (set' a b rl) a = Some b.
Proof.
  intros rl a b.
  induction rl as [|(a', b') rl' IHrl'].
  simpl; unfold beq. 
  simplbnat.
  trivial.
  simpl.
  decbltnat a a'; simpl; unfold beq; simplbnat;
    trivial.
  decbeqnat a a'; simpl; unfold beq; simplbnat;
    trivial.
Qed.

Lemma get'_set'_neq : forall rl a a0 b,
  a <> a0 -> get' (set' a b rl) a0 = get' rl a0.
Proof.
  intros rl a a0 b Hneq.
  induction rl as [|(a', b') rl' IHrl'].
  simpl; unfold beq.
  simplbnat. trivial.
  simpl; unfold beq.
  decbltnat a a'; simpl; unfold beq;
    simplbnat; trivial.
  decbeqnat a a'; simpl; unfold beq;
    simplbnat; trivial.
  decbeqnat a0 a'; simpl; unfold beq; 
    simplbnat; trivial.
  discribnat.
  decbeqnat a0 a';simpl; unfold beq; 
    simplbnat; trivial.
Qed.

Lemma lb_set'_lt : forall rl a a0 b,
  lb a rl -> lt a0 a -> lb a0 (set' a b rl).
Proof.
  intros rl a a0 b Ha Hlta0.
  induction rl as [|(a', b') rl' IHrl'].
  simpl in *.
  split; trivial.
  simpl in *.
  destruct Ha as [Hlta Hlb].
  decbltnat a a'; simpl; unfold beq;
    simplbnat; trivial.
  split_l.
  split_l by (eapply lt_trans; eauto).
  eapply lb_trans; eauto.
  decbeqnat a a'; simpl; unfold beq;
    simplbnat; trivial.
  split_l.
  eapply lt_trans; eauto.
  eapply lb_trans; eauto.
  split_l by (eapply lt_trans; eauto).
  apply (IHrl' Hlb).
Qed.

Lemma lb_set'_gt : forall rl a a0 b,
  lb a0 rl -> lt a0 a -> lb a0 (set' a b rl).
Proof.
  intros rl a a0 b Ha0 Hlta.
  induction rl as [|(a', b') rl' IHrl'].
  simpl in *.
  split; trivial.
  simpl in *.
  destruct Ha0 as [Hlta0 Hlb].
  decbltnat a a'; simpl; unfold beq;
    simplbnat; trivial.
  split_l.
  split_lr.
  decbeqnat a a'; simpl; unfold beq;
    simplbnat; trivial.
  split_lr.
  split_l.
  apply (IHrl' Hlb).
Qed.

Lemma is_orderset_set' : forall rl : rlist, 
  is_orderset rl -> forall a b, is_orderset (set' a b rl).
Proof.
  intros rl Hrl a b.
  induction rl as [|(a', b') rl' IHrl'].
  split; trivial.
  simpl in *.
  destruct Hrl as [Hlb Hrl'].
  decbltnat a a'; simpl; unfold beq;
    simplbnat; trivial.
  split; split; trivial.
  apply blt_true_lt; trivial.
  eapply lb_trans; eauto.
  apply blt_true_lt; trivial.
  decbeqnat a a'; simpl; unfold beq;
    simplbnat; trivial.
  split_lr.
  split.
  eapply lb_set'_gt; eauto.
  apply blt_true_lt; trivial.
  solvebnat.
  apply (IHrl' Hrl').
Qed.

Definition set (m : map) (a : A) (b : B) : map :=
  listset_con (set' a b (lst m)) 
         (is_orderset_set' (lst m) (prf m) a b).

Lemma set_sem_old : forall m a a' b, 
  match (beq a a') with
    | true => get (set m a b) a = Some b
    | false => get (set m a b) a' = get m a'
  end.
Proof.
  intros m a a' b.
  unfold get, set.
  destruct m as [m prf0].
  destruct (eq_dec a a') as [Heq | Hneq]; 
    unfold beq; simpl; subst; simplbnat.
  eapply get'_set'; eauto.
  eapply get'_set'_neq; eauto.
Qed.

Lemma set_sem : forall m a a' b,
  get (set m a b) a' =
  match (beq a a') with
    | true => Some b
    | false => get m a'
  end.
  intros.
  assert (H := set_sem_old). substH H with (H m a a' b).
  assert (H1 := beq_true_eq). substH H1 with (H1 a a').
  assert (H2 := beq_false_neq). substH H2 with (H2 a a').
  destruct (beq a a').
  assert (H3 := refl_equal). substH H3 with (H3 bool true).
  substH H1 with (H1 H3).
  rewrite H1 in H; rewrite H1; trivial.
  trivial.
Qed.
  

(***********************)
(** upd                *)
(***********************)

(* Fixpoint upd' (a: A) (b : B) (rl : rlist)  *)
(*   {struct rl} : option rlist := *)
(*   match rl with *)
(*     | nil => None *)
(*     | (a', b') :: rl' => *)
(*       match (eq_dec a a') with *)
(* 	| left _ => Some ((a', b) :: rl') *)
(* 	| right _ =>  *)
(*             let rl'' := (upd' a b rl') in *)
(*               match rl'' with  *)
(*                 | None => None *)
(*                 | Some rl' => Some ((a', b') :: rl') *)
(*               end *)
(*       end *)
(*   end. *)

(* Lemma lb_upd' : forall rl rlx a0 a b, *)
(*   upd' a b rl = Some rlx -> lb a0 rl  *)
(*   -> lb a0 rlx. *)
(* Proof. *)
(*   induction rl as [|(a',b') rl' IHrl']. *)
(*   simpl; trivial. *)
(*     intros; discriminate. *)
(*   simpl in * . *)
(*     intros rlx a0 a b H [H1 H2]. *)
(*     destruct rlx as [|(ax, bx) rlx]. *)
(*     simpl;trivial. *)
(*     destruct (eq_dec a a') as [Heq | Hneq]. *)
(*     subst a'. *)
(*     injection H. *)
(*     intros; subst rlx; subst bx; subst ax. *)
(*     apply (conj H1 H2). *)
(*     assert (H3 := IHrl' rlx a0 a b). *)
(*     destruct (upd' a b rl'). *)
(*     injection H; intros; subst rlx; subst bx; subst ax. *)
(*     simpl. *)
(*     split_l. *)
(*     auto. *)
(*     discriminate. *)
(* Qed. *)

(* Lemma is_orderset_upd' : forall rl rlx a b, *)
(*   upd' a b rl = Some rlx -> is_orderset rl  *)
(*   -> is_orderset rlx. *)
(* Proof. *)
(*   induction rl as [|(a',b') rl' IHrl']. *)
(*   simpl; trivial. *)
(*     intros; discriminate. *)
(*   intros rlx a b H [H1 H2]. *)
(*   destruct rlx as [|(ax, bx) rlx]. *)
(*     simpl;trivial. *)
(*   simpl in H. *)
(*   destruct (eq_dec a a') as [Heq | Hneq]. *)
(*   subst a'. *)
(*   injection H; intros; subst rlx; subst bx; subst ax. *)
(*   simpl; split_lr. *)
(*   assert (H3 := IHrl' rlx a b). *)
(*   assert (H0 := lb_upd' rl' rlx a' a b). *)

(*   destruct (upd' a b rl'). *)
(*   injection H; intros; subst rlx; subst bx; subst ax. *)
(*   simpl. *)
(*   split_r. *)
(*   auto. *)
(*   auto. *)
(*   discriminate. *)
(* Qed. *)

(* Lemma upd'_dec : forall m a b, *)
(*   {m' : rlist |  upd' a b m = Some m'} + {upd' a b m = None}. *)
(* Proof. *)
(*   induction m. *)
(*   intros. *)
(*   right. *)
(*   simpl. *)
(*   trivial. *)
(*   simpl. *)
(*   destruct a. *)
(*   intros a0 b0. *)
(*   destruct (eq_dec a0 a) as [Heq | Hneq]. *)
(*   left. *)
(*   exists ((a, b0)::m). *)
(*   trivial. *)
(*   destruct (upd' a0 b0 m). *)
(*   left. *)
(*   exists ((a,b)::r). *)
(*   trivial. *)
(*   right. *)
(*   trivial. *)
(* Qed. *)

(* Definition upd (m : map) (a : A) (b : B) : option map.  *)
(*   intros m a b. *)
(*   destruct m as [rl pt]. *)
(*   destruct (upd'_dec rl a b) as [[m' Hs] | Hn]. *)
(*   apply (Some (listset_con m'  *)
(*             (is_orderset_upd' rl m' a b  *)
(*               Hs pt))). *)
(*   apply (@None map). *)
(* Defined. *)

(* Lemma upd'_get' : *)
(*   forall rl a b, *)
(*     get' rl a = Some b  *)
(*     -> exists rl', upd' a b rl = Some rl'. *)
(* Proof. *)
(*   induction rl as [|(a', b') rl' IHrl']. *)
(*   simpl; intros;discriminate. *)
(*   simpl; intros a b He. *)
(*   destruct (eq_dec a a') as [Heq | Hneq]. *)
(*   exists ((a', b) :: rl'); trivial. *)
(*   assert (H := IHrl' a b He). *)
(*   destruct (upd' a b rl'). *)
(*   exists ((a', b') :: r); trivial. *)
(*   destruct H; discriminate. *)
(* Qed. *)

(* Lemma upd_sem : forall m a b, *)
(*   match get_dec m a with  *)
(*     | left _ => upd m a b = Some (set m a b) *)
(*     | right _ => upd m a b = None *)
(*   end. *)
(* Proof. *)

(* Admitted. *)

Lemma emp_sem : 
  forall a : A, get emp_map a = None.
Proof.
  intros; unfold get, emp_map; trivial.
Qed.

Lemma sig_sem : forall (a a' : A) (b : B),
  get (sig_map a b) a' = 
  match (beq a a') with
    | true => Some b
    | false => None
  end.
Proof.
  unfold get, sig_map, get', sig', lst. intros a a' b.
  destruct (eq_dec a a').
  rewrite e; trivial.
  rewrite (neq_beq_false a a').
  rewrite (neq_beq_false a' a).
  trivial.
  apply sym_not_eq; trivial.
  trivial.
Qed.

Lemma equal_ext: forall m1 m2, (lst m1) = (lst m2) -> m1 = m2.
Proof.
  destruct m1 as [m1  prf1].
  destruct m2 as [m2  prf2].
  simpl.
  intros.
  subst m1.
  rewrite (proof_irrelevance _ prf1 prf2).
  auto.
Qed.


Lemma extensionality: forall (m1 m2 : map), 
  (forall a, get m1 a = get m2 a) -> m1 = m2.
Proof.
  intros m1 m2 H.
  apply equal_ext.
  destruct m1 as [m1 prf1].
  destruct m2 as [m2 prf2].
  gen_clear m1 m2 prf1 prf2 H.
  induction m1 as [|(a1, b1) m1' IHm1'];
    unfold get, get', lst, beq; 
      cbv beta in *; fold get' in *;
        intros.

  induction m2 as [|(a2, b2) m2' IHm2'];
    unfold get, get', lst, beq; 
      cbv beta in *; fold get' in *.
  reflexivity.
  simpl in H.
  unfold beq in H.
  assert (H2 := H a2); simplbnat; discri.
  destruct m2 as [|(a2,b2) m2']; simpl in *; 
    unfold beq in *.
  assert (H1 := H a1); simplbnat; discri.
  assert (He : 
    forall a : A, get' m1' a = get' m2' a).
(* subgoals := a1 < a2 , a1 = a2, a1 > a2 *)
  decbltnat a1 a2; decbeqnat a1 a2; 
    simpl; unfold beq; simplbnat; trivial.
  substH H with (H a1).
  simplbnat.
  assert (Hf := in_notlb m2' a1 b1).
  rewrite <- H in Hf; clear H.
  assert (Hlt := blt_true_lt a1 a2 Hb).
  assert (Hlb := lb_trans m2' a1 a2 (proj1 prf2) Hlt).
  destruct (Hf (refl_equal _) Hlb).
  intro a; substH H with (H a).
  assert (H0 := beq_true_eq a1 a2 Hb0).
  subst a1.
  destruct (eq_dec a a2) as [Heq | Hneq].
  subst a2.
  simplbnat.
  rewrite (lb_notin m1' a (proj1 prf1)).
  rewrite (lb_notin m2' a (proj1 prf2)).
  trivial.
  simplbnat.
  trivial.
  substH H with (H a2).
  simplbnat.
  assert (Hf := in_notlb m1' a2 b2).
  rewrite <- H in Hf; clear H.
  assert (Hbgt : blt_nat a2 a1 = true).
    solvebnat.
  assert (Hgt := blt_true_lt a2 a1 Hbgt).
  assert (Hlb := lb_trans m1' a2 a1 (proj1 prf1) Hgt).
  destruct (Hf (refl_equal _) Hlb).
(* solved *)
  substH IHm1' with
    (IHm1' m2' (proj2 prf1) (proj2 prf2) He).
  subst m2'.
  assert (H1 := H a1).
  assert (H2 := H a2).
  simplbnat.
  destruct (eq_dec a1 a2); simpl; unfold beq; 
    simplbnat.
  subst a1.
  simplbnat.
  inversion H2.
  auto.
  rewrite (lb_notin m1' a1 (proj1 prf1)) in H1.
  discri.
Qed.

(************************)
(** merge               *)
(************************)

Fixpoint app' (rl1 rl2 : rlist) {struct rl1} : rlist :=
  match rl1 with 
    | nil => rl2
    | (a,b) :: rl1' =>
      (set' a b (app' rl1' rl2))
  end.

Lemma is_set_app' : forall (rl1 rl2 : rlist), 
  is_orderset rl1 -> is_orderset rl2 -> is_orderset (app' rl1 rl2).
Proof.
  intros rl1 rl2 Hrl1 Hrl2.
  induction rl1 as [|(a', b') rl1' IHrl1'].
  simpl; trivial.
  simpl in *.
  apply is_orderset_set'.
  apply (IHrl1' (proj2 Hrl1)).
Qed.

Definition merge (m1 : map) (m2 : map) : map :=
  listset_con (app' (lst m1) (lst m2)) 
    (is_set_app' _ _ (prf m1) (prf m2)).

Lemma merge_sem : forall (m1 m2 : map) (a : A),
    get (merge m1 m2) a 
    = match (get m1 a, get m2 a) with
        | (Some b1, Some b2) => Some b1
        | (Some b1, None) => Some b1
        | (None, Some b2) => Some b2
        | (None, None) => None
      end.
Proof.
  intros m1 m2 a.
  unfold merge, get.
  destruct m1 as [rl1 prf1].
  destruct m2 as [rl2 prf2].
  simpl.
  induction rl1 as [|(a', b') rl1' IHrl1'].
  simpl.
  destruct (get' rl2 a); trivial.
  simpl in *.
  destruct (eq_dec a a') as [Heq | Hneq].
  subst a'; simpl.
  rewrite (get'_set' (app' rl1' rl2) a b').
  unfold beq; simplbnat.
  destruct (get' rl2 a); trivial.
  assert (Hneq' : a' <> a).
  intro H; apply Hneq; auto.
  unfold beq; simplbnat.
  rewrite (get'_set'_neq (app' rl1' rl2) 
    a' a b' Hneq').
  apply IHrl1'.
  apply (proj2 prf1).
Qed.

End ListMap.
