(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* Tiny Machine                                                *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

(* $Id: codeheap.v,v 1.1.1.1 2006/12/07 16:43:23 ssg05 Exp $ *)

Require Import Word.
Require Import Reg.
Require Import Instr.

Require Export Coq.Lists.List.

Require ListB_MAP.

Module CHeapDef := ListB_MAP.ListMap Instr.
Module CHeap := MAP.MapEx CHeapDef.

Notation codeheap := CHeap.map.

Definition get : codeheap -> label -> option instr 
  := CHeap.get.

Fixpoint getub (ch : codeheap) {struct ch} : nat :=
  match ch with
    | nil => O
    | cons (f, i) ch' => max_nat f (getub ch')
  end.

Lemma getub_ok : forall (ch : codeheap) (f : word) i,
  get ch f = Some i
  -> f > getub ch
  -> False.
Proof.
  induction ch as [ | (f', i') ch' ]; simpl.
  intros; discriminate.
  intros f i H H0.
  unfold CHeapDef.beq in H.
  decbeqnat f f'.
  substbnat f'.
  assert (Hx := max_nat_elim_l f (getub ch')).
  omega.
  apply IHch' with f i; trivial.
  unfold max_nat in H0.
  decbltnat f' (getub ch'); desbnat;
    omega.
Qed.

Fixpoint getis_aux (ch : codeheap) (f : label) (b : nat) 
  {struct b} : option instrseq :=
  match b with
    | O => None
    | S b' => 
      match get ch f with
        | None => None
        | Some (j f') => Some (ione (j f'))
        | Some (jal f') => Some (ione (jal f'))
        | Some (jr r') => Some (ione (jr r'))
        | Some i => match getis_aux ch (f+1) b' with
                      | Some is' => Some (iseq i is')
                      | None => None
                    end
      end
  end.

Definition getis (ch : codeheap) (f : label) :=
  getis_aux ch f (S (getub ch)).

Inductive getis2 : codeheap -> label -> instrseq -> Prop:=
  | getis2_j : 
    forall (ch : codeheap) (f f': label),
      (get ch f = Some (j f')) 
      -> getis2 ch f (ione (j f'))
  | getis2_jal : 
    forall (ch : codeheap) (f f': label),
      (get ch f = Some (jal f')) 
      -> getis2 ch f (ione (jal f'))
  | getis2_jr : 
    forall (ch : codeheap) (f : label) (r : reg),
      (get ch f = Some (jr r)) 
      -> getis2 ch f (ione (jr r))
  | getis2_addu : 
    forall (ch : codeheap) (f : label) (is : instrseq)
      (rd rs rt : reg),
      get ch f = Some (addu rd rs rt)
      -> getis2 ch (f+1) is 
      -> getis2 ch f (iseq (addu rd rs rt) is)
  | getis2_addiu : 
    forall (ch : codeheap) (f : label) (is : instrseq)
      (rd rs : reg) (w : word),
      get ch f = Some (addiu rd rs w)
      -> getis2 ch (f+1) is 
      -> getis2 ch f (iseq (addiu rd rs w) is)
  | getis2_subu : 
    forall (ch : codeheap) (f : label) (is : instrseq)
      (rd rs rt : reg),
      get ch f = Some (subu rd rs rt)
      -> getis2 ch (f+1) is 
      -> getis2 ch f (iseq (subu rd rs rt) is)
  | getis2_subiu : 
    forall (ch : codeheap) (f : label) (is : instrseq)
      (rd rs : reg) (w : word),
      get ch f = Some (subiu rd rs w)
      -> getis2 ch (f+1) is 
      -> getis2 ch f (iseq (subiu rd rs w) is)
  | getis2_move : 
    forall (ch : codeheap) (f : label) (is : instrseq)
      (rd rs : reg),
      get ch f = Some (move rd rs)
      -> getis2 ch (f+1) is 
      -> getis2 ch f (iseq (move rd rs) is)
  | getis2_li : 
    forall (ch : codeheap) (f : label) (is : instrseq)
      (rd : reg) (w : word),
      get ch f = Some (li rd w)
      -> getis2 ch (f+1) is 
      -> getis2 ch f (iseq (li rd w) is)
  | getis2_lw : 
    forall (ch : codeheap) (f : label) (is : instrseq)
      (rt rs : reg) (w : word),
      get ch f = Some (lw rt w rs)
      -> getis2 ch (f+1) is 
      -> getis2 ch f (iseq (lw rt w rs) is)
  | getis2_sw : 
    forall (ch : codeheap) (f : label) (is : instrseq)
      (rt rs : reg) (w : word),
      get ch f = Some (sw rt w rs)
      -> getis2 ch (f+1) is 
      -> getis2 ch f (iseq (sw rt w rs) is)
  | getis2_blt : 
    forall (ch : codeheap) (f : label) (is : instrseq)
      (rs rt : reg) (f' : label),
      get ch f = Some (blt rs rt f')
      -> getis2 ch (f+1) is 
      -> getis2 ch f (iseq (blt rs rt f') is)
  | getis2_beq : 
    forall (ch : codeheap) (f : label) (is : instrseq)
      (rs rt : reg) (f' : label),
      get ch f = Some (beq rs rt f')
      -> getis2 ch (f+1) is 
      -> getis2 ch f (iseq (beq rs rt f') is)
  | getis2_bgtz : 
    forall (ch : codeheap) (f : label) (is : instrseq)
      (rs : reg) (f' : label),
      get ch f = Some (bgtz rs f')
      -> getis2 ch (f+1) is 
      -> getis2 ch f (iseq (bgtz rs f') is).

Lemma getis_aux_ok : forall b ch f is,
  b + f > getub ch
  -> getis_aux ch f b = Some is
  -> getis2 ch f is.
Proof.
  induction b;
    intros.
  simpl in H0.
  discriminate.
  simpl in H0.
  unfold get in H0.
  destruct (CHeap.get_dec ch f) as [[i Hg] | Hg];
    rewrite Hg in H0;
    try discriminate.
  destruct i;
    try (
      generalize (refl_equal (getis_aux ch (f+1) b));
      pattern ch at 1;
      destruct (getis_aux ch (f+1) b); 
      try discriminate;
      intro His';
      inversion_clear H0;
      constructor; trivial;
      apply IHb; trivial;
      omega).
Qed.

Lemma getis_aux_ok_2 : forall b ch f is,
  b + f > getub ch
  -> getis2 ch f is
  -> getis_aux ch f b = Some is.
Proof.
  induction b;
    intros.
  simpl.
  simpl in H.
  inversion_clear H0; 
    destruct (getub_ok _ _ _ H1 H).
  inversion_clear H0;
    simpl; rewrite H1; trivial;
    rewrite (IHb ch (f+1) is0); trivial;
    omega.
Qed.

Lemma getis2_getis : forall ch f is,
  getis2 ch f is -> getis ch f = Some is.
Proof.
  intros ch f is Hgetis2.
  unfold getis.
  apply getis_aux_ok_2; trivial.
  omega.
Qed.

Lemma getis_getis2 : forall ch f is,
  getis ch f = Some is -> getis2 ch f is.
Proof.
  unfold getis.
  intros ch f is Hgetis.
  apply getis_aux_ok with (S (getub ch)); trivial.
  omega.
Qed.

Lemma getis_ione_get : 
  forall ch f i,
    getis ch f = Some (ione i)
    -> get ch f = Some i.
Proof.
  intros ch f i H.
  assert (Hx := getis_getis2 _ _ _ H).
  inversion_clear Hx; trivial.
Qed.

Lemma getis_iseq_get : 
  forall ch f i is,
    getis ch f = Some (iseq i is)
    -> get ch f = Some i /\ getis ch (f + 1) = Some is.
Proof.
  intros ch f i is H.
  assert (Hx := getis_getis2 _ _ _ H).
  inversion_clear Hx; split; trivial; 
    apply getis2_getis; trivial.
Qed.

Lemma getis2_merge_intro :
  forall (m1 m2 : codeheap) (f : label) (is : instrseq), 
    getis2 m1 f is \/ getis2 m2 f is
    -> CHeap.disj m1 m2
    -> getis2 (CHeap.merge m1 m2) f is.
Proof.
  intros m1 m2 f is His Hdisj.
  destruct His as [His | His].
  induction His as 
    [ m1 f f' H
      | m1 f f' H
      | m1 f r H
      | m1 f is rd rs rt H His IHis
      | m1 f is rd rs w H His IHis
      | m1 f is rd rs rt H His IHis
      | m1 f is rd rs w H His IHis
      | m1 f is rd rs H His IHis
      | m1 f is rd w H His IHis
      | m1 f is rt w rs H His IHis
      | m1 f is rt w rs H His IHis
      | m1 f is rs rt f' H His IHis
      | m1 f is rs rt f' H His IHis
      | m1 f is rs f' H His IHis
    ]; 
    constructor; trivial; unfold get in * ;
    try (rewrite (CHeap.merge_sem m1 m2 f);
    rewrite H; destruct (CHeap.get m2 f); 
      trivial);
    auto.

  induction His as 
    [ m2 f f' H
      | m2 f f' H
      | m2 f r H
      | m2 f is rd rs rt H His IHis
      | m2 f is rd rs w H His IHis
      | m2 f is rd rs rt H His IHis
      | m2 f is rd rs w H His IHis
      | m2 f is rd rs H His IHis
      | m2 f is rd w H His IHis
      | m2 f is rt w rs H His IHis
      | m2 f is rt w rs H His IHis
      | m2 f is rs rt f' H His IHis
      | m2 f is rs rt f' H His IHis
      | m2 f is rs f' H His IHis
    ];
    constructor; auto; unfold get in * ;
    assert (Hdisj' := Hdisj f);
    rewrite (CHeap.merge_sem m1 m2 f);
    destruct (CHeap.get m2 f);
    destruct (CHeap.get m1 f); 
    try contradiction; trivial;
    try discriminate.
Qed.

Lemma getis2_compat_merge_intro :
  forall (m1 m2 : codeheap) (f : label) (is : instrseq), 
    getis2 m1 f is \/ getis2 m2 f is
    -> CHeap.compat m1 m2
    -> getis2 (CHeap.merge m1 m2) f is.
Proof.
  intros m1 m2 f is His Hcomp.
  destruct His as [His | His].
  induction His as 
    [ m1 f f' H
      | m1 f f' H
      | m1 f r H
      | m1 f is rd rs rt H His IHis
      | m1 f is rd rs w H His IHis
      | m1 f is rd rs rt H His IHis
      | m1 f is rd rs w H His IHis
      | m1 f is rd rs H His IHis
      | m1 f is rd w H His IHis
      | m1 f is rt w rs H His IHis
      | m1 f is rt w rs H His IHis
      | m1 f is rs rt f' H His IHis
      | m1 f is rs rt f' H His IHis
      | m1 f is rs f' H His IHis
    ]; 
    constructor; trivial; unfold get in * ;
    try (rewrite (CHeap.merge_sem m1 m2 f);
    rewrite H; destruct (CHeap.get m2 f); 
      trivial);
    auto.

  induction His as 
    [ m2 f f' H
      | m2 f f' H
      | m2 f r H
      | m2 f is rd rs rt H His IHis
      | m2 f is rd rs w H His IHis
      | m2 f is rd rs rt H His IHis
      | m2 f is rd rs w H His IHis
      | m2 f is rd rs H His IHis
      | m2 f is rd w H His IHis
      | m2 f is rt w rs H His IHis
      | m2 f is rt w rs H His IHis
      | m2 f is rs rt f' H His IHis
      | m2 f is rs rt f' H His IHis
      | m2 f is rs f' H His IHis
    ];
    constructor; auto; unfold get in * ;
    assert (Hcomp' := Hcomp f);
    rewrite (CHeap.merge_sem m1 m2 f);
    destruct (CHeap.get m2 f);
    destruct (CHeap.get m1 f);
    try subst; trivial;
    try contradiction; trivial;
    try discriminate.
Qed.

Lemma getis_disj_merge_intro :
  forall (m1 m2 : codeheap) (f : label) (is : instrseq), 
    getis m1 f = Some is \/ getis m2 f = Some is
    -> CHeap.disj m1 m2
    -> getis (CHeap.merge m1 m2) f = Some is.
Proof.
  intros m1 m2 f is [H1 | H2] Hdisj.
  apply getis2_getis.
  apply getis2_merge_intro; trivial.
  left; apply getis_getis2; trivial.
  apply getis2_getis.
  apply getis2_merge_intro; trivial.
  right; apply getis_getis2; trivial.
Qed.

Lemma getis_compat_merge_intro :
  forall (m1 m2 : codeheap) (f : label) (is : instrseq), 
    getis m1 f = Some is \/ getis m2 f = Some is
    -> CHeap.compat m1 m2
    -> getis (CHeap.merge m1 m2) f = Some is.
Proof.
  intros m1 m2 f is [H1 | H2] Hcomp.
  apply getis2_getis.
  apply getis2_compat_merge_intro; trivial.
  left; apply getis_getis2; trivial.
  apply getis2_getis.
  apply getis2_compat_merge_intro; trivial.
  right; apply getis_getis2; trivial.
Qed.
