(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* Map Implementation (List)                                   *)
(*                                                             *)
(*    A -> B : Set                                             *)
(***************************************************************)

(* $Id: ListAB_MAP.v,v 1.1.1.1 2006/12/07 16:43:17 ssg05 Exp $ *)

Require Import Coq.Bool.Bool.
Require Import Coq.Lists.List.

Require Import LIN_ORDER.
Require Import MAP.

Require Import extac.

Set Implicit Arguments.

Module ListMap (ADef : LIN_ORDER) (BDef : ONE_SET) <: MAPDEF.

  Module ALib := LinLib ADef.

Definition A := ADef.A.

Definition beq := ADef.beq.

Lemma beq_true_eq : forall a a' : A, 
  beq a a' = true -> a = a'.
Proof.
  apply ADef.beq_true_eq.
Qed.

Lemma beq_false_neq : forall a a' : A,
    beq a a' = false -> a <> a'.
Proof.
  apply ADef.beq_false_neq.
Qed.

Lemma eq_beq_true : forall a a' : A, 
    a = a' -> beq a a' = true.
Proof.
  apply ADef.eq_beq_true.
Qed.

Lemma neq_beq_false : forall a a' : A, 
    a <> a' -> beq a a' = false.
Proof.
  apply ADef.neq_beq_false.
Qed.

Definition eq_dec := ADef.eq_dec.
(* Lemma A_eq_dec : forall a a' : A, {a=a'}+{a<>a'}. *)
(* Proof. *)
(*   apply eq_dec. *)
(* Qed. *)

Definition B := BDef.B.

(***********************)
(** map                *)
(***********************)

Definition map := list (prod A B).

Definition emp_map : map := nil.

Definition sig_map (a : A) (b : B) := (a, b) :: nil.

(***********************)
(** set                *)
(** get                *)
(***********************)

Fixpoint get (m : map) (a : A) {struct m} : option B :=
  match m with 
    | nil => None
    | (a', b') :: m' =>
      match beq a a' with
        | true => Some b'
        | false => get m' a
      end
  end.

Lemma get_dec :
  forall (m : map) (a : A),
    {exists b, get m a = Some b} + {get m a = None}.
Proof.
  intros m a.
  destruct (get m a).
  left.
  exists b.
  trivial.
  right; trivial.
Qed.

Fixpoint set (m : map) (a : A) (b : B) {struct m} : map :=
  match m return map with
    | nil => (a, b) :: nil
    | (a', b') :: m' =>
      match beq a a' with
        | true => (a, b) :: m'
        | false => (a', b') :: (set m' a b)
      end
  end.

Fixpoint upd (m : map) (a : A) (b : B) {struct m} : option map :=
  match m return option map with
      | nil => None
      | (a', b') :: m' =>
        match beq a a' with
          | true => Some ((a, b) :: m')
          | false => 
              let m'' := (upd m' a b) in
                match m'' with 
                  | None => None
                  | Some m' => Some ((a', b') :: m')
                end
        end
  end.

Lemma set_sem : forall m a a' b, 
  match (beq a a') with
    | true => get (set m a b) a = Some b
    | false => get (set m a b) a' = get m a'
  end.
Proof.
  intros m a a0 b.
  unfold set, get, beq. 
  induction m as [|(a',b') m' IHm'].
  unfold beq.
  ALib.simplbeq.
  destruct (ADef.eq_dec a a0) as [Heq | Hneq].
  subst a0; ALib.simplbeq;  trivial.
  ALib.simplbeq.
  trivial.
  
  unfold beq in * .
  destruct (ADef.eq_dec a a0) as [Heq | Hneq].
  subst a0.
  ALib.simplbeq.
  destruct (ADef.eq_dec a a') as [Heq' | Hneq'].
  subst a'.
  ALib.simplbeq.
  unfold beq.
  ALib.simplbeq.
  trivial.
  ALib.simplbeq.
  unfold beq.
  ALib.simplbeq.
  trivial.
  ALib.simplbeq.
  destruct (ADef.eq_dec a a') as [Heq'1 | Hneq'1].
  subst a'.
  ALib.simplbeq.
  unfold beq.
  ALib.simplbeq.
  trivial.
  ALib.simplbeq.
  unfold beq.
  ALib.simplbeq.
  destruct (ADef.eq_dec a0 a') as [Heq' | Hneq'].
  subst a'.
  ALib.simplbeq.
  trivial.
  ALib.simplbeq.
  trivial.
Qed.

Lemma upd_sem_0 : forall m a b b0, 
  (get m a = Some b -> upd m a b0 = Some (set m a b0))
  /\ (get m a = None -> upd m a b0 = None).
Proof.
  intros.
  split.
  intros.
  unfold upd, set, get in * .
  induction m as [|(a',b') m' IHm'].
  discriminate.
  destruct (eq_dec a a') as [Heq | Hneq].
  subst a'.
  unfold beq in * .
  ALib.simplbeq.
  trivial.
  unfold beq in * .
  ALib.simplbeq.
  fold set in * .
  fold upd in * .
  fold get in * .
  substH IHm' with (IHm' H).
  rewrite IHm'.
  trivial.
  unfold get, upd, set.
  induction m as [|(a',b') m' IHm'].
  trivial.
  destruct (eq_dec a a') as [Heq | Hneq].
  subst a'.
  unfold beq in * .
  ALib.simplbeq.
  intro Hf; discriminate Hf.
  unfold beq in * .
  ALib.simplbeq.
  fold set in * .
  fold upd in * .
  fold get in * .
  intros H.
  rewrite (IHm' H).
  trivial.
Qed.

Lemma upd_sem : forall m a b,
    match get_dec m a with 
      | left _ => upd m a b = Some (set m a b)
      | right _ => upd m a b = None
    end.
Proof.
  intros.
  destruct (get_dec m a) as [[b' Hi] | Hni].
  assert (H := proj1 (upd_sem_0 m a b' b)).
  tauto.
  assert (H := proj2 (upd_sem_0 m a b b)).
  tauto.
Qed.

Lemma emp_sem : 
  forall a : A, get emp_map a = None.
Proof.
  intros; unfold get, emp_map; trivial.
Qed.

Lemma sig_sem : forall (a a' : A) (b : B),
  get (sig_map a' b) a = 
  match (beq a a') with
    | true => Some b
    | false => None
  end.
Proof.
  unfold get. 
  unfold sig_map. intros a a' b.
    destruct (eq_dec a a'); trivial.
Qed.

(************************)
(** merge               *)
(************************)

Fixpoint merge (m1 m2 : map) {struct m1} : map := 
  match m1 with 
    | nil => m2
    | (a, b) :: m1' => (a, b) :: (merge m1' m2)
  end.

Lemma merge_sem : forall (m1 m2 : map) (a : A),
  get (merge m1 m2) a 
  = match get m1 a, get m2 a with
      | Some b1, Some b2 => Some b1
      | Some b1, None => Some b1
      | None, Some b2 => Some b2
      | None, None => None
    end.
Proof.
  intros m1 m2 a. unfold merge.
  induction m1 as [| (a1, b1) m1' IHm1'].
  unfold get at 2.
  destruct (get m2 a); trivial.
  unfold merge in *; fold merge in *.
  unfold get; fold get.
  destruct (eq_dec a a1) as [Heq | Hneq].
  subst a1.
  unfold beq; ALib.simplbeq.
  destruct (get m2 a).
  trivial.
  trivial.
  unfold beq; ALib.simplbeq.
  assumption.
Qed.

End ListMap.


Unset Implicit Arguments.
