(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* Separation Logic                                            *)
(*                                                             *)
(***************************************************************)

(* $Id: seplogic.v,v 1.1.1.1 2006/12/07 16:43:22 ssg05 Exp $ *)

Require Import Coq.Lists.List.
Require Import machine.

Require Import extac.

Ltac hypreplace H y :=
  let H' := fresh in
    (assert (H' := y); try (clear H; rename H' into H)).

Set Implicit Arguments.
Unset Strict Implicit.

Notation hpred := (heap -> Prop).

Definition hpred_true (h:heap) (A:hpred) := A h.

Notation "h ||- A" := (hpred_true h A) (at level 80) : seplogic_scope.

Open Local Scope seplogic_scope.

Definition sconj (A B : hpred) : hpred :=
  fun h => exists ha, exists hb,
    Heap.join ha hb h /\
    ha ||- A /\ hb ||- B.

Notation " A ** B " := (sconj A B) 
  (at level 60, right associativity) : seplogic_scope.

(* alternative conjuction *)
Definition aconj (A B:hpred) : hpred :=
  fun h => A h /\ B h.

Notation "a &&& b " := (aconj a b) (at level 60, right associativity)
  : seplogic_scope.

Definition unit : hpred := 
  fun h : heap =>
    h = Heap.emp_map.

Definition top : hpred := fun h : heap => True.

Definition lex (A:Set) (f:A -> hpred) : hpred :=
  fun h => exists x, f x h.

Notation "'lexists' x , A" := (lex (fun x => A))
  (at level 200, x ident) : seplogic_scope.


Definition bang (A:Prop) : hpred :=
  fun h => unit h /\ A.

Notation "! A" := (bang A) (at level 55) : seplogic_scope.

Definition cont l v : hpred :=
  fun h : heap =>
    h = Heap.sig_map l v.

(* Definition cont (l:nat) (x:nat) : hpred := *)
(*   fun h => heap_eqv h (singleton_heap l x). *)

Infix "|->" := cont (at level 59, right associativity)
   : seplogic_scope.

(* Fixpoint itr (p:nat -> hpred) (l : list nat) {struct l} : hpred := *)
(*   match l with *)
(*     | nil => unit *)
(*     | cons x l' => (p x) ** (itr p l') *)
(*   end. *)

(* Definition fas (p:nat -> hpred) (s: set) : hpred := *)
(*   itr p (sd_of s). *)

(* Notation "'foralls' x 'in' s , A" := (fas (fun x => A) s)  *)
(*   (at level 200, x ident). *)

(* l |-> x1, ..., xn *)

Fixpoint contL (l:nat) (xx:list nat) {struct xx} : hpred :=
  match xx with
    | nil => unit
    | x :: nil => l |-> x
    | x :: xx' => l |-> x ** contL (S l) xx'
  end.

Infix "|->," := contL (at level 59, right associativity)
   : seplogic_scope.

Definition pairCont x v0 v1 := x |->, (v0::v1::nil).

Infix "|->>" := pairCont (at level 59, right associativity)
   : seplogic_scope.

(* l |-> _ *)
Notation "l |->_" := (lexists xyzyzy, l |-> xyzyzy) 
  (at level 59)  : seplogic_scope.
(* l |-> _, _ *)
(* Notation " label |->__" :=  *)
(*   (sconj (lexists x, cont label x) (lexists x, cont (label+1) x)) (at level 59). *)
(* (* l |-> _, _, _ *) *)
(* Notation "l |->___" := (((l) |->_) ** (S l |->__)) (at level 59). *)
(* (* l |-> _, _, _, _ *) *)
(* Notation "l |->____" := (((l) |->_) ** (S l |->___)) (at level 59). *)

Notation "A ==> B" :=
  (forall h, h ||- A -> h ||- B) (at level 13) 
   : seplogic_scope.

(* ************************************************)
(* sconj                                          *)

Lemma sconj_intro : forall ha hb h A B,
  ha ||- A -> hb ||- B -> Heap.join ha hb h ->
    h ||- A ** B.
Proof.
  intros.
  exists ha; exists hb; auto.
Qed.

    (* elim is unfold *)

Lemma sconj_sym : forall h A B,
  h ||- A ** B -> h ||- B ** A. 
Proof. 
  intros h A B [ha [hb [J [Ha Hb]]]].
  apply (sconj_intro Hb Ha (Heap.join_comm J)).
Qed.

Lemma sconj_mono_left : forall h A B,
  h ||- (A ** B) -> 
    forall A', (A ==> A') -> 
      h ||- (A' ** B).
Proof.
  destruct 1 as [ha [hb [J [HA HB]]]].
  exists ha; exists hb.
  intuition.
Qed.

Lemma sconj_mono_right : forall h A B,
  h ||- (A ** B) -> 
    forall B', (B ==> B') -> 
      h ||- (A ** B').
Proof.
  destruct 1 as [ha [hb [J [HA HB]]]].
  exists ha; exists hb.
  intuition.
Qed.

Lemma sconj_mono : forall h A A' B B',
  h ||- A ** B -> 
    (A ==> A') -> (B ==> B') -> 
    h ||- A' ** B'.
Proof.
  intros.
  assert (h ||- (A' ** B)).
  eapply sconj_mono_left; [eauto | assumption].
  eapply sconj_mono_right; [eauto | assumption].
Qed.

Lemma sconj_assoc_left : forall A B C h,
  h ||- (A ** B) ** C -> h ||- A ** (B ** C).
Proof.
  intros.
  destruct H as [hab [hc [Jabc [HAB HC]]]].
  destruct HAB as [ha [hb [Jab [HA HB]]]].
  destruct (Heap.join_assoc_l Jab Jabc) as [hbc [Jabc' Jbc]].
  eapply sconj_intro; eauto.
  eapply sconj_intro; eauto.
Qed.

Lemma sconj_assoc_right : forall A B C h,
  h ||- A ** (B ** C) -> h ||- (A ** B) ** C.
Proof.
  intros.
  apply sconj_sym.
  apply sconj_assoc_left.
  exact (sconj_sym (sconj_assoc_left (sconj_sym H))).
Qed.

(* ************************************************)
(* unit                                           *)

Lemma unit_intro : Heap.emp_map ||- unit.
Proof.
  unfold hpred_true.
  unfold unit. 
  trivial.
Qed.

Lemma unit_intro_left : forall h P,
  h ||- P -> h ||- unit ** P.
Proof.
  intros.
  eapply sconj_intro.
  2: apply H.
  apply unit_intro.
  intro addr. 
  destruct (Heap.get h addr); simpl; trivial.
Qed.
Implicit Arguments unit_intro_left [h P].

Lemma unit_intro_right : forall h P,
  h ||- P -> h ||- P ** unit.
Proof.
  intros.
  apply (sconj_sym (unit_intro_left H)).
Qed.

Lemma unit_elim : forall h,
  h ||- unit -> h = Heap.emp_map.
Proof.
  intros.
  unfold hpred_true, unit in H.
  trivial.
Qed.

Lemma unit_elim_left : forall h P,
  h ||- unit ** P -> h ||- P.
Proof.
  intros.
  destruct H as [h1 [h2 [HJ [H1 H2]]]].
  unfold hpred_true, unit in H1.
  rewrite H1 in HJ.
  rewrite (Heap.join_emp_l HJ) in H2.
  trivial.
Qed.

Lemma unit_elim_right : forall h P,
  h ||- P ** unit -> h ||- P.
Proof.
  intros.
  apply (unit_elim_left (sconj_sym H)).
Qed.

(* ************************************************)
(* top                                            *)

Lemma top_intro : forall h, h ||- top.
Proof.
  unfold hpred_true.
  unfold top. 
  trivial.
Qed.

Lemma top_intro_left : forall h P,
  h ||- P -> h ||- top ** P.
Proof.
  intros.
  assert (Heap.join Heap.emp_map h h).
  intro addr. 
  destruct (Heap.get h addr); simpl; trivial.
  apply (sconj_intro (top_intro Heap.emp_map) H H0).
Qed.

Lemma top_intro_right : forall h P,
  h ||- P -> h ||- P ** top.
Proof.
  intros.
  apply (sconj_sym (top_intro_left H)).
Qed.

(* ************************************************)
(* bang                                           *)

Lemma bang_intro : forall h (A:Prop),
  h ||- unit -> A ->
    h ||- ! A.
Proof.
  unfold bang.
  intros h A Hp Aok.
  split; assumption; assumption.
Qed.

Lemma bang_elim1 : forall h A,
  h ||- ! A -> A.
Proof.
  unfold bang.
  unfold hpred_true.
  intuition.
Qed.

Lemma bang_elim2 : forall h A,
    h ||- ! A -> h ||- unit.
Proof.
  unfold bang; unfold hpred_true.
  intuition.
Qed.

Lemma bang_elim : forall h A,
  h ||- ! A ->
    h ||- unit /\ A.
Proof.
  intros h A Hp.
  split.
  exact (bang_elim2 Hp).
  exact (bang_elim1 Hp).
Qed.

Lemma bang_elim_left : forall h A B,
    h ||- ! A ** B ->
      A /\ h ||- B.
Proof.
  intros h A B H.
  assert (A /\ h ||- unit ** B).
  destruct H as [ha [hb [J [HA HB]]]].
  constructor.
  apply (bang_elim1 HA).
  assert (HA' := bang_elim2 HA).
  eapply sconj_intro; eauto.
  intuition.
  exact (unit_elim_left H2).
Qed.

Lemma bang_elim_left1 : forall h A B,
  h ||- ! A ** B ->
    A.
Proof.
  intros h A B H.
  destruct (bang_elim_left H).
  assumption.
Qed.

Lemma bang_elim_left2 : forall h A B,
  h ||- ! A ** B ->
    h ||- B.
Proof.
  intros h A B H.
  destruct (bang_elim_left H).
  assumption.
Qed.

Lemma bang_elim_right : forall h A B,
    h ||- B ** ! A ->
      A /\ h ||- B.
Proof.
  intros h A B H.
  substH H with (sconj_sym H).
  apply bang_elim_left; auto.
Qed.

Lemma bang_intro_left : forall h (A:Prop) B,
  A -> h ||- B ->
    h ||- ! A ** B.
Proof.
  intros h A B HA HB.
  unfold bang.
  assert (Q := unit_intro_left HB).
  eapply sconj_mono; unfold hpred_true; eauto.
  assumption.
Qed.

Lemma bang_intro_right : forall h (A:Prop) B,
  A -> h ||- B ->
    h ||- B ** ! A.
Proof.
  intros.
  apply sconj_sym.
  apply bang_intro_left; auto.
Qed.

(* ************************************************)
(* cont                                           *)

Lemma cont_elim : forall h l w,
  h ||- l |-> w -> h = (Heap.sig_map l w).
Proof.
  unfold hpred_true; unfold cont.
  tauto.
Qed.

Lemma cont_get_eq : forall h l w,
  h ||- l |-> w -> 
    Heap.get h l = Some w.
Proof.
  intros h l w H.
  unfold hpred_true, cont in H.
  rewrite H.
  rewrite (Heap.sig_sem l l w).
  destruct (Heap.eq_dec l l).
  Heap.simplbeq.
  trivial.
  Heap.simplbeq.
  trivial.
Qed.

Lemma sconjDisj : forall x y z h,
  h ||- x |-> y ** x |-> z ->
    False.
Proof.
  intros.
  destruct H as [h1 [h2 [J [H1 H2]]]].
  unfold Heap.join in J.
  substH J with (J x).
  rewrite (cont_get_eq H1) in J.
  rewrite (cont_get_eq H2) in J.
  trivial.
Qed.

Lemma cont_get_eq1 : forall h l w A,
  h ||- l |-> w ** A -> 
    Heap.get h l = Some w.
Proof.
  intros h l w A Q.
  destruct Q as [ha [hb [J [HA HB]]]].
  unfold Heap.join in J.
  substH J with (J l).
  substH HA with (cont_get_eq HA).
  rewrite HA in J.
  destruct (Heap.get hb l);
    destruct (Heap.get ha l);
      destruct (Heap.get h l);
        intuition.
Qed.

(* Lemma cont_set_eq1 : forall h x y A, *)
(*   h ||- x |-> y ** A -> forall z, *)
(*     heap_set' h x z = Some (heap_set h x z). *)
(* Proof. *)
(*   intros. *)
(*   unfold heap_set'. *)
(*   rewrite (cont_get_eq1 H). *)
(*   reflexivity. *)
(* Qed. *)

Lemma cont_update : forall a v1 v2 h,
  h ||- a |-> v1 ->
    Heap.set h a v2 ||- a |-> v2.
Proof.
  intros a v1 v2 h Hh.
  rewrite Hh.
  rewrite (Heap.set_sig_a_sig a v1 v2).
  unfold cont, hpred_true.
  trivial.
Qed.

Lemma cont_update1 : forall h a v1 A,
  h ||- (a |-> v1 ** A) -> forall v2,
    Heap.set h a v2 ||- (a |-> v2 ** A).
Proof.
  intros h a v1 A [h1 [h2 [J [H1 H2]]]] v2.
  exists (Heap.set h1 a v2).
  exists h2.
  split.
  intro x.
  assert (HJ :=J a).
  rewrite (cont_get_eq H1) in HJ.
  destruct (nat_eq_dec a x) as [Heq | Hneq].
  subst x.
  Heap.simplgs.
  destruct (Heap.get h2 a); trivial.
  Heap.simplgs.
  substH HJ with (J x).
  destruct (Heap.get h1 x);
    destruct (Heap.get h2 x);
      try tauto ; try auto.
  split_r.
  eapply cont_update; eauto.
Qed.

Lemma join_sing_eqv : forall h h1 h2 h3 h4 a v,
  h1 ||- a |-> v -> Heap.join h1 h2 h ->
    h3 ||- a |-> v -> Heap.join h3 h4 h 
      -> h2 = h4.
Proof.
  intros h h1 h2 h3 h4 a v Hh1 Hh Hh3 Hh'.
  unfold hpred_true, cont in Hh1, Hh3.
  rewrite <- Hh1 in Hh3.
  rewrite Hh3 in Hh'.
  apply (Heap.join_id_eq_r Hh Hh').
Qed.

Lemma cont_cont_eq : forall h x v1 v2,
  h ||- x |-> v1 ->
    h ||- x |-> v2 ->
      v1 = v2.
Proof.
  intros h x v1 v2 H1 H2.
  unfold cont, hpred_true in *.
  rewrite H1 in H2.
  injection H2.
  trivial.
Qed.

Lemma cont_cont_eq1 : forall h x v1 v2 A B,
  h ||- x |-> v1 ** A ->
    h ||- x |-> v2 ** B ->
      v1 = v2.
Proof.
  intros h x v1 v2 A B H1 H2.
  assert (Some v1 = Some v2).
  rewrite <- (cont_get_eq1 H1).
  rewrite <- (cont_get_eq1 H2).
  reflexivity.
  injection H; auto.
Qed.

Lemma cont_cont_false : forall h a x y A,
  h ||-  a |-> x ** a |-> y ** A -> False.
Proof.
  intros.
  destruct H as [h1 [h2 [J [Hp1 Hp2]]]].
  hypreplace J (J a).
  rewrite (cont_get_eq Hp1) in J.
  rewrite (cont_get_eq1 Hp2) in J.
  trivial.
Qed.

Lemma lex_intro : forall h (A:Set) (f:A->hpred) x,
  h ||- f x ->
    h ||- lex f.
Proof.
  intros.
  exists x.
  assumption.
Qed.

Lemma lex_intro1 : forall h (A:Set) (f:A->hpred),
  (exists x, h ||- f x) ->
  h ||- lex f.
Proof.
  intros h A f [x Hp].
  eapply lex_intro; eauto.
Qed.

Lemma lex_elim : forall h (A:Set) (f:A->hpred),
  h ||- lex f -> 
    exists x, h ||- f x.
Proof.
  unfold lex; auto.
Qed.

Lemma lex_elim_left  : forall (T:Set) h (A:T -> hpred) B,
  h ||- (lex A) ** B -> h ||- lexists x, (A x) ** B.
Proof.
  intros T h A B H.
  destruct H as [h1 [h2 [J [HA HB]]]].
  destruct (lex_elim HA) as [x HA']; clear HA.
  eapply lex_intro; eapply sconj_intro; eauto.
Qed.

Lemma lex_elim_right : forall (T:Set) h A (B:T -> hpred),
  h ||- A ** (lex B) -> h ||- lexists x, A ** (B x).
Proof.
  intros T h A B H.
  destruct (lex_elim_left (sconj_sym H)) as [x H'].
  exists x.
  apply sconj_sym; assumption.
Qed.

Lemma lex_intro_left : forall (T:Set) h A (B:T -> hpred),
  h ||- (lexists x, (B x) ** A) -> h ||- (lex B) ** A.
Proof.
  intros T h A B Hp.
  destruct (lex_elim Hp) as [x Hp'].
  eapply sconj_mono_left; eauto.
  intros; eapply lex_intro; eauto.
Qed.

Lemma lex_intro_right : forall (T:Set) h A (B:T -> hpred),
  h ||- (lexists x, A ** (B x)) -> h ||- A ** (lex B).
Proof.
  intros T h A B Hp.
  destruct (lex_elim Hp) as [x Hp'].
  eapply sconj_mono_right; eauto.
  intros; eapply lex_intro; eauto.
Qed.


Ltac sepelim Hp :=
  match type of Hp with
    | _ ||- unit => hypreplace Hp (unit_elim Hp); subst h
    | _ ||- bang _ => 
      let A := fresh "A" in let Hp' := fresh "Hp" in
	(destruct (bang_elim Hp) as  [Hp' A]; 
	  hypreplace Hp' (unit_elim Hp'); subst h; clear Hp')
    | _ ||- lex _ => 
      let x := fresh "x" in let Hp' := fresh "Hp'" in
	(destruct (lex_elim Hp) as [x Hp']; 
	  clear Hp; rename Hp' into Hp; sepelim Hp)

    | _ ||- sconj (bang _) _ => 
      let A := fresh "A" in let Hp' := fresh "Hp" in
	(destruct (bang_elim_left Hp) as  [A Hp']; 
	  clear Hp; rename Hp' into Hp; sepelim Hp)
    | _ ||- sconj _ (bang _) => 
      let A := fresh "A" in let Hp' := fresh "Hp" in
	(destruct (bang_elim_right Hp) as  [A Hp']; 
	  clear Hp; rename Hp' into Hp; sepelim Hp)

    | _ ||- sconj unit _ => hypreplace Hp (unit_elim_left Hp); sepelim Hp
    | _ ||- sconj _ unit => hypreplace Hp (unit_elim_right Hp); sepelim Hp
    | _ ||- sconj (sconj _ _) _ => hypreplace Hp (sconj_assoc_left Hp); sepelim Hp

    | _ ||- sconj _ _ =>  
      let h1 := fresh "h" in let h2 := fresh "h" in let J := fresh "J" in 
	let Hp1 := fresh "Hp" in let Hp2 := fresh "Hp" in
	  (destruct Hp as [h1 [h2 [J [Hp1 Hp2]]]];
	    sepelim Hp1; sepelim Hp2;
	    assert (Hp := sconj_intro Hp1 Hp2 J);
	      clear h1 h2 J Hp1 Hp2)
    | _ => idtac
  end.

(* old --- to be deleted 
Ltac assocl H :=
 match goal with
   | [ H : _ ||- contL _ _ ** _ |- _ ] => idtac
   | [ H : _ ||- (_ ** _) ** _ |- _ ] => hypreplace H (sconj_assoc_left H);
     assocl H
   | _ => idtac
 end.

Ltac assocr :=
 match goal with
   | [ |- _ ||- contL _ _ ** _ ] => idtac
   | [ |- _ ||- (_ ** _) ** _ ] => apply sconj_assoc_right; assocr
   | _ => idtac
 end. *)

Ltac assocl H :=
  repeat (hypreplace H (sconj_assoc_left H)).

Ltac assocr :=
  repeat (apply sconj_assoc_right).

Ltac matchhp ig ih sg sh :=
  match goal with
    | |- _ ||- top => apply top_intro (* true *)
    | [Hp : ?h ||- ?A |- ?h ||- ?A ] => exact Hp (* assumption *)
      
    | |- _ ||- unit => try (apply unit_intro)
    | |- _ ||- bang _ => apply bang_intro; auto

    | |- _ ||- sconj _ unit => apply unit_intro_right; 
      matchhp (pred ig) sh (pred sg) sh
    | |- _ ||- sconj unit _ => apply unit_intro_left;
      matchhp (pred ig) sh (pred sg) sh
    | |- _ ||- sconj (bang _) _ => apply bang_intro_left; 
      [auto | matchhp (pred ig) sh (pred sg) sh]
    | |- _ ||- sconj _ (bang _) => apply bang_intro_right; 
      [auto | matchhp (pred ig) sh (pred sg) sh]
      
(*     | [NI : ~ set_in ?x ?s |- *)
(*       ?h ||- fas ?P (set_add ?x ?s)] => apply fas_intro1;  *)
(*     [assocr; matchhp (S ig) sh (S sg) sh | assumption] *)

(*     | [NI : ~ set_in ?x ?s |- *)
(*       ?h ||- sconj (fas ?P (set_add ?x ?s)) _] => apply fas_intro;  *)
(* 	[assocr; matchhp (S ig) sh (S sg) sh | assumption] *)
      
    | |- _ ||- sconj _ (bang _) => apply bang_intro_right; 
      [auto | matchhp (pred ig) sh (pred sg) sh]

    | [Hp : ?h ||- sconj ?A ?B |- ?h ||- sconj ?A ?C] => (* match left *)
      let h1 := fresh "h" in
	(apply (sconj_mono_right Hp);
	  (clear h Hp; intros h Hp) || (clear Hp; intros h1 Hp);
	    assocl Hp; assocr;
            matchhp (pred ig) (pred sh) (pred sg) (pred sh))
    | [Hp : ?h ||- sconj ?A ?B |- ?h ||- sconj ?C ?B] => (* match right *)
      let h' := fresh "h" in
	(apply (sconj_mono_left Hp);
	  (clear h Hp; intros h Hp) || (clear Hp; intros h' Hp);
	    assocl Hp; assocr;
            matchhp (pred ig) (pred sh) (pred sg) (pred sh))
	
    | [Hp : ?h ||- sconj ?A ?B |- ?h ||- sconj ?C ?D] => (* shift *)
      match eval compute in ig with
	| O => (* no more shift on goal *)
	  match eval compute in ih with
	    | O => idtac (* finished *)
	    | _ => (* shift hyp, continue with matching *)
              hypreplace Hp (sconj_sym Hp); assocl Hp;
	      matchhp sg (pred ih) sg sh
	  end
	| _ => (* shift goal *)
	  apply sconj_sym; assocr;  matchhp (pred ig) ih sg sh
      end
    | _ => idtac
  end.

Ltac simplsep sizeg sizeh :=
 match goal with
 | [Hp : ?h ||- _ |- ?h ||- _ ] =>
   assocl Hp; assocr; matchhp sizeg sizeh sizeg sizeh
 end.

Ltac lexintro size :=
  match goal with
    | |- _ ||- lex _ => apply lex_intro1
    | |- _ ||- sconj (lex _) _ => apply lex_intro_left; apply lex_intro1
    | |- _ ||- sconj _ (lex _) => apply lex_intro_right; apply lex_intro1
    | |- _ ||- sconj _ _ => 
      match eval compute in size with
	| O => idtac
	| S _ => assocr; apply sconj_sym; assocr; lexintro (pred size)
      end
  end.

Ltac hvinject' i1 i2 s1 s2 :=
  match goal with
    | [Hp1 : ?h ||- (?addr |-> ?v1),
      Hp2 : ?h ||- (?addr |-> ?v2) |- _] =>
    let Eq := fresh "Eq" in
      assert (Eq := cont_cont_eq Hp1 Hp2);
	clear Hp1 Hp2; try (clear h)
    | [Hp1 : ?h ||- sconj (?addr |-> ?v1) _ ,
      Hp2 : ?h ||- sconj (?addr |-> ?v2) _ |- _] =>
    let Eq := fresh "Eq" in 
      let h11 := fresh "h" in let h12 := fresh "h" in
	let h21 := fresh "h" in let h22 := fresh "h" in
	  let J1 := fresh "J" in let J2 := fresh "J" in
	    let Hp11 := fresh "Hp" in let Hp12 := fresh "Hp" in
	      let Hp21 := fresh "Hp" in let Hp22 := fresh "Hp" in
		(assert (Eq := cont_cont_eq1 Hp1 Hp2);
		  destruct Hp1 as [h11 [h12 [J1 [Hp11 Hp12]]]];
		    destruct Hp2 as [h21 [h22 [J2 [Hp21 Hp22]]]];
		      rewrite Eq in Hp11;
 			try (rewrite <- (join_sing_eqv Hp11 J1 Hp21 J2) in Hp22;
			  clear h11 h21 J1 J2 Hp11 Hp21; try (clear h);
                            assocl Hp21; assocl Hp22;
			    hvinject' (pred i1) (pred i2) (pred s1) (pred s2)))
    | [ Hp1 : ?h ||- sconj _ _,
	Hp2 : ?h ||- sconj _ _ |- _] =>
    match eval compute in i1 with
      | O => (* no more shift on Hp1 *)
	match eval compute in i2 with
	  | O => clear Hp1 Hp2; try (clear h)(* finished *)
	  | _ => (* shift hyp, continue with matching *)
            hypreplace Hp2 (sconj_sym Hp2); assocl Hp2;
	    hvinject' s1 (pred i2) s1 s2
	end
      | _ => (* shift on Hp2 *)
        hypreplace Hp1 (sconj_sym Hp1); assocl Hp1;
	hvinject' (pred i1) i2 s1 s2
    end
  end.

Ltac hvinject s1 s2 :=
  match goal with
    | [ Hp1 : ?h ||- _, Hp2 : ?h ||- _ |- _] =>
      generalize Hp1; generalize Hp2;
	assocl Hp1; assocl Hp2;
	hvinject' s1 s2 s1 s2; intros
  end.

Ltac sconjlift n :=
  match eval compute in n with
    | O => assocr
    | _ => assocr; apply sconj_sym; sconjlift (pred n)
  end.

Ltac sconjlifth n H :=
  match eval compute in n with
    | O => assocl H
    | _ => assocl H; hypreplace H (sconj_sym H); sconjlifth (pred n) H
  end.

(* **********************************************************************)

(* move from vcgen.v  *)

Notation "s |||- A" := (S_H s ||- A) (at level 80) : seplogic_scope.

Ltac simpllw x n :=
  match goal with
    | [ |- context [(Heap.get ?h ?addr)]] =>
      let Hp := fresh "Hp" in
        (assert (Hp : Heap.get h addr = Some x);
          [ apply cont_get_eq1 with (A := top);
            simplsep 2 n | rewrite Hp; clear Hp; simplst])
    | [ |- context [(Heap.get ?h ?addr)]] =>
      let Hp := fresh "Hp" in
        (assert (Hp : Heap.get h addr = Some x);
          [ apply cont_get_eq1 with (A := top);
            simplsep 2 n | rewrite Hp; clear Hp; simplst])
    | [ |- context [(?h ?addr)]] =>
      let Hp := fresh "Hp" in
        (assert (Hp : Heap.get h addr = Some x);
          [ apply cont_get_eq1 with (A := top);
            simplsep 2 n | rewrite Hp; clear Hp; simplst])
  end.

(* Ltac simplif := *)
(*   match goal with *)
(*     | [ |- context [(if beq_nat ?x ?y then _ else _)]] => *)
(*       let R := fresh "R" in *)
(*         (destruct (eq_nat_dec x y) as [R | R]; [rewrite R | idtac]; simplbeq) *)
(*     | [ |- context [(if blt_nat ?x ?y then _ else _)]] => *)
(*       let R := fresh "R" in *)
(*         (destruct (lt_nat_dec x y) as [R | R];  *)
(*           [rewrite (blt_true R) | rewrite (blt_false R)]) *)
(*   end. *)

Ltac simplsw' sg v:=
  match goal with
    | |- (Heap.set ?h ?addr ?val) ||- ?addr |-> ?val => 
      apply cont_update with (v1 := v)
    | |- (Heap.set ?h ?addr ?val) ||- sconj (?addr |-> ?val) ?A  => (* apply update *)
      apply cont_update1 with (A := A) (v1 := v)
        
    | |- (Heap.set ?h ?addr ?val) ||- sconj _ _  => (* apply update *)
      match eval compute in sg with
        | O => idtac
        | S _ => assocr; apply sconj_sym; assocr; simplsw' (pred sg) v
      end
  end.

Ltac simplsw sg sh v :=
  assocr; simplsw' sg v; simplsep sg sh.

(* Ltac recoverhget := *)
(*   match goal with *)
(*     | [ |- context [(hp_of ?s ?addr)]] => *)
(*       replace (hp_of s addr) with (hget s addr); *)
(*         [recoverhget | reflexivity] *)
(*     | [ |- context [(heap_set ?h ?a ?v ?addr)]] => *)
(*       replace (heap_set h a v addr) with (heap_get (heap_set h a v) addr); *)
(*         [recoverhget | reflexivity] *)
(*     | _ => idtac *)
(*   end. *)

Ltac lwomega :=
  match goal with
    | [H : _ ||- _ |- _] =>
      clear H; lwomega
    | _ => abstract omega
  end.
