(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* Typed Assembly Language                                     *)
(*                                                             *)
(* tal-0, a monomorphic version                                *)
(***************************************************************)

(* $Id: tal0.v,v 1.1.1.1 2006/12/07 16:43:20 ssg05 Exp $ *)

Require Import ListB_MAP.
Require Import machine.
Require Import extac.

Open Local Scope tm_scope.

Inductive flag : Set :=
  | init : flag
  | junk : flag.

Inductive valTy : Set :=
  | valty_var : nat -> valTy
  | valty_int : valTy
  | valty_tup : tupTy -> valTy
  | valty_code : rfTy -> valTy
with rfTy : Set := 
  | rfty_nil : rfTy
  | rfty_cons : (prod reg valTy) -> rfTy -> rfTy
with tupTy : Set :=
  | tupty_nil : tupTy
  | tupty_cons : (prod flag valTy) -> tupTy 
                  -> tupTy.

Implicit Types Y Gamma : rfTy.
Implicit Types t tau : valTy.
Implicit Types o tup : tupTy.

Scheme valTy_ind2 :=
  Induction for valTy Sort Prop
 with rfTy_ind2 :=
  Induction for rfTy Sort Prop
 with tupTy_ind2 :=
  Induction for tupTy Sort Prop.

(* **************************************************)
(* cell type                                        *)

Definition cellTy := prod flag valTy.

(* **************************************************)
(* code type                                        *)

Definition codeTy := rfTy.


(* **************************************************)
(* Notations                                        *)

Notation " ^ n " := (valty_var n) 
  (at level 40) : tal_scope.
Open Local Scope tal_scope.

Notation " 'int' " := (valty_int)
  (at level 0) : tal_scope.

Notation " []- Y " := (valty_code Y)
  (at level 40) : tal_scope.

Notation "[{ r , .. , r' }]" := 
  (rfty_cons r .. (rfty_cons r' rfty_nil) ..) 
  (at level 0) : tal_scope.

Notation "[{}]" := (rfty_nil)  
  (at level 0) : tal_scope.

Notation "[< o , .. , o' >]" := 
  (valty_tup (tupty_cons o .. (tupty_cons o' tupty_nil) ..)) 
  (at level 0) : tal_scope.

(* **************************************************)
(* regfile type                                     *)

Module RfTy.

Fixpoint sel (m : rfTy) (r : reg) {struct m} : option valTy :=
  match beq_reg r zero with
    | true => Some valty_int
    | false =>
      match m with 
        | rfty_nil => None
        | rfty_cons (r', b') m' =>
          match beq_reg r r' with
            | true => Some b'
            | false => sel m' r
          end
      end
  end.

Fixpoint upd (m : rfTy) (r : reg) (b : valTy) {struct m} : rfTy :=
  match beq_reg r zero with
    | true => m
    | false =>
      match m return rfTy with
        | rfty_nil => rfty_cons (r, b) rfty_nil
        | rfty_cons (r', b') m' =>
          match beq_reg r r' with
            | true => rfty_cons (r, b) m'
            | false => rfty_cons (r', b') (upd m' r b)
          end
      end
  end.

Definition indom (m : rfTy) (r : reg) := 
  exists b : valTy, sel m r = Some b.

Definition lookup (m : rfTy) (r : reg) (t : valTy) := 
  sel m r = Some t.

Lemma lookup_nil :
  forall r t,
    beq_reg r zero = false
    ->lookup rfty_nil r t 
    -> False.
Proof.
  intros r t Hr H.
  unfold lookup in H.
  unfold sel in H.
  rewrite Hr in H.
  discriminate.
Qed.

Lemma indom_nil :
  forall r,
    beq_reg r zero = false
    -> indom rfty_nil r -> False.
Proof.
  unfold indom; simpl; intros r Hr [b H].
  rewrite Hr in H; discriminate.
Qed.

Lemma sel_upd_req :
  forall Y r r' t,
    beq_reg zero r = false
    -> beq_reg r r' = true
    -> sel (upd Y r' t) r = Some t.
Proof.
  intros Y r r' t H H1.
  induction Y as [|(r0, t0) Y IHY]; unfold upd, sel;
    simplbreg; trivial.
  decbeqreg r' r0; trivial.
Qed.

Lemma sel_upd_rneq :
  forall Y r r' t,
    beq_reg r r' = false
    -> sel (upd Y r' t) r = sel Y r.
Proof.
  intros Y r r' t H.
  induction Y as [|(rx, tx) Y IHY]; simpl.
    decbeqreg r r0; simpl; simplbreg; trivial.
    decbeqreg r' r0; simpl; simplbreg; trivial.
    decbeqreg r r0; simpl; simplbreg; trivial.
  decbeqreg r' rx; simpl; simplbreg; trivial.
  decbeqreg r rx; simpl; simplbreg; trivial;
    (decbeqreg r' zero; simpl; simplbreg; trivial).
  decbeqreg r' rx; simpl; simplbreg; trivial.
Qed.

End RfTy.


(* **************************************************)
(* tuple type                                       *)

Module TupTy.

Fixpoint mem (tup : tupTy) (ct : cellTy) {struct tup} : Prop := 
  match tup with
    | tupty_nil => False
    | tupty_cons ct' tup' => ct = ct' \/ mem tup' ct
  end.

Fixpoint lookup (tup : tupTy) (n:nat) (ct:cellTy) 
  {struct n} : Prop :=
  match n, tup with
    | O, tupty_cons x l' => x = ct
    | O, tupty_nil => False
    | S m, tupty_nil => False
    | S m, tupty_cons x l' => lookup l' m ct
  end.

Fixpoint get (tup : tupTy) (n : nat) 
  {struct n} : option cellTy :=
  match n, tup with
    | O, tupty_cons ct l' => Some ct
    | O, tupty_nil => None
    | S m, tupty_nil => None
    | S m, tupty_cons x l' => get l' m
  end.
  
Fixpoint upd (tup : tupTy) (n : nat) (ct : cellTy) 
  {struct n}: tupTy :=
  match n, tup with 
    | O, tupty_cons ct' l' => tupty_cons ct l'
    | S m, tupty_cons ct' l' => tupty_cons ct' (upd l' m ct)
    | _, _ => tupty_nil
  end.

Definition indom (tup : tupTy) (n : nat) := 
  exists ct : cellTy, lookup tup n ct.

Lemma lookup_nil : 
  forall n ct,
    lookup tupty_nil n ct -> False.
Proof.
  intros n ct.
  intros Hf.
  induction n; simpl; intros; auto.
Qed.

Lemma lookup_get : forall n l ct,
  lookup l n ct -> get l n = Some ct.
Proof.
  induction n; destruct l; 
    simpl; intros;
      tauto || auto.
  subst; trivial.
Qed.

Lemma get_lookup : forall n l ct,
  get l n = Some ct -> lookup l n ct.
Proof.
  induction n; destruct l; 
    simpl; intros;
      discriminate || auto.
  injection H; intro; trivial.
Qed.

Lemma lookup_upd_indom :
  forall n l n' (ct ct' : cellTy),
    lookup (upd l n ct) n' ct' -> indom l n'.
Proof.
  unfold indom;
  induction n; destruct l; destruct n'; 
    simpl; intros;
    tauto || auto.
  exists p; trivial.
  exists ct'; trivial.
  exists p; trivial.
  eapply IHn; eauto.
Qed.

Lemma upd_a_lookup_a : 
  forall n l (ct ct' : cellTy),
    lookup (upd l n ct') n ct -> ct = ct'.
Proof.
  induction n; destruct l; simpl; intros;
    try tauto.
  subst ct; trivial.
  eapply IHn; eauto.
Qed.

Lemma upd_a_lookup_a' : 
  forall addr addr' l (ct ct' : cellTy),
    addr <> addr'
    -> lookup (upd l addr ct') addr' ct 
       = lookup l addr' ct.
Proof.
  induction addr; destruct l; simpl in * ;
    try tauto.
  induction addr'; intros.
  destruct (H (refl_equal _)).
  simpl.
  trivial.
  intros.
  induction addr'; intros.
  simpl. trivial.
  simpl.
  apply IHaddr.
  intro.
  apply H.
  rewrite H0.
  trivial.
Qed.

Lemma lookup_unique :
  forall addr l  (ct ct' : cellTy),
    lookup l addr ct 
    -> lookup l addr ct'
    -> ct = ct'.
Proof.
  induction addr; destruct l; simpl; intros .
  destruct H.
  subst p; trivial.
  destruct H.
  apply (IHaddr l ct ct' H H0).
Qed.

Lemma lookup_upd : 
  forall addr l (ct ct' : cellTy),
    lookup l addr ct
    -> lookup (upd l addr ct') addr ct'.
Proof.
  induction addr; destruct l; simpl; intros; trivial.
  eapply IHaddr; eauto.
Qed.

Lemma lookup_upd_neq :
  forall addr l addr' (ct ct' : cellTy),
    addr <> addr'
    -> lookup l addr ct
    -> lookup (upd l addr' ct') addr ct.
Proof.
  induction addr; destruct l; simpl; intros; 
    tauto || trivial.
  subst p.
  destruct addr'.
  tauto.
  simpl in * .
  trivial.
  destruct addr'.
  simpl.
  assumption.
  simpl.
  apply IHaddr.
  intro Hf.
  apply H.
  subst addr; trivial.
  trivial.
Qed.

Lemma indom_indom_upd :
  forall n (tup : tupTy) (n': nat) (ct : cellTy),
    indom tup n -> indom (upd tup n' ct) n.
Proof.
  unfold indom.
  intros n tup n' ct [ct' H].
  destruct (nat_eq_dec n n') as [Heq | Hneq].
  subst n'.
  exists ct.
  eapply lookup_upd; eauto.
  exists ct'.
  assert (H1 : n' <> n).
  intro H0; apply (Hneq (sym_eq H0)).
  apply lookup_upd_neq; trivial.
Qed.

Lemma indom_upd_indom :
  forall n (tup : tupTy) (n': nat) (ct : cellTy),
    indom (upd tup n' ct) n -> indom tup n.
Proof.
  intros n tup n' ct [ct' H].
  eapply lookup_upd_indom; eauto.
Qed.

End TupTy.

(* ************************************************ *)
(* Psi : CHSpec                                     *)
(* Phi : DHSPec                                     *)


Module CdSpec <: ONE_SET.
  Definition B := codeTy.
End CdSpec.

Module CHSpecDef := ListMap CdSpec.

Module CHSpec := MapEx CHSpecDef.
(* psi : chspec *)
Definition chspec := CHSpec.map.

Implicit Types Psi psi: chspec.

Module DtSpec <: ONE_SET.
  Definition B := cellTy.
End DtSpec.

Module DHSpecDef := ListMap DtSpec.
Module DHSpec := MapEx DHSpecDef.
(* psi : chspec *)
Definition dhspec := DHSpec.map.

Implicit Types Phi phi: dhspec.

(* **************************************************)
(* cell subtyping                                   *)

Inductive ceTy_sub : forall (ct ct' : cellTy), Prop :=
  | ceTy_sub_refl : forall (ct : cellTy),
                       ceTy_sub ct ct
  | ceTy_sub_init : forall (t : valTy), 
                       ceTy_sub (init, t) (junk, t).

Lemma ceTy_sub_trans : forall ct ct' ct'',
  ceTy_sub ct ct'
  -> ceTy_sub ct' ct''
  -> ceTy_sub ct ct''.
Proof.
  intros ct ct' ct'' H12 H23.
  destruct ct; destruct ct'; destruct ct''.
  destruct f1; inversion_clear H12;  inversion_clear H23;
    try apply ceTy_sub_refl ; try apply ceTy_sub_init.
Qed.

Lemma ceTy_sub_init_ct : forall t ct ct',
  ceTy_sub ct' ct
  -> ceTy_sub (init, t) ct
  -> ceTy_sub (init, t) ct'.
Proof.
  intros; destruct ct'; destruct ct.
  inversion_clear H; inversion_clear H0;
    try apply ceTy_sub_init;
      try apply ceTy_sub_refl.
Qed.

(* **************************************************)
(* regfile subtyping                                *)

Inductive rfTy_sub : forall (Y Y' : rfTy), Prop :=
  | rfty_sub :
      forall (Y Y' : rfTy),
        (forall r : reg,
             RfTy.indom Y' r
             -> RfTy.sel Y' r = RfTy.sel Y r)
        -> rfTy_sub Y Y'.

Lemma rfty_sub_refl : forall Y, rfTy_sub Y Y.
Proof.
  intros.
  apply rfty_sub; trivial.
Qed.

Lemma rfty_sub_trans : forall Y Y' Y'',
  rfTy_sub Y Y'
  -> rfTy_sub Y' Y''
  -> rfTy_sub Y Y''.
Proof.
  intros Y Y' Y'' H12 H23.
  inversion_clear H12.
  inversion_clear H23.
  apply rfty_sub; trivial.
  intros r Hindom.
  inversion Hindom.
  rewrite (H0 r Hindom) in H1.
  assert (H3 : RfTy.indom Y' r).
  unfold RfTy.indom.
  exists x.
  apply H1.
  rewrite (H0 r Hindom).
  apply (H r H3).
Qed.

Lemma rfty_lookup_sub_lookup : forall Y Y' r w,
  RfTy.lookup Y' r w -> rfTy_sub Y Y' 
  -> RfTy.lookup Y r w.
Proof.
   intros Y Y' r w HY' Hsub.
   inversion_clear Hsub.
   unfold RfTy.indom, RfTy.lookup in * .
   assert (H0 : exists b, RfTy.sel Y' r = Some b).
   exists w; trivial.
   rewrite (H r H0) in HY'.
   assumption.
Qed.

(* **************************************************)
(* code subtyping                                   *)

(*  |- a1 <= a2 *)

Inductive codeTy_sub : forall (a a' : codeTy), Prop :=
  | codety_sub : 
        forall  (Y Y' : rfTy),
          rfTy_sub Y Y'
          -> codeTy_sub Y Y'.


(* **************************************************)
(* value typing                                     *)

Inductive valok : chspec -> dhspec -> word -> valTy -> Prop :=
  | valok_lab : 
              forall (psi : chspec) (Phi : dhspec) 
                (f : word) (Y : rfTy),
                CHSpec.lookup psi f Y
                -> valok psi Phi f (valty_code Y)
  | valok_int : forall (psi : chspec) (Phi : dhspec) (w : word),
              valok psi Phi w valty_int
  | valok_tup : forall (psi : chspec) (Phi : dhspec) 
                    (l : label) (tup : tupTy),
              (forall i, 
                TupTy.indom tup i
                -> exists ct, exists ct',   
                  TupTy.lookup tup i ct
                  /\ DHSpec.lookup Phi (l+i) ct'
                  /\ ceTy_sub ct' ct)
              -> valok psi Phi l (valty_tup tup).
              
Inductive regok : rfTy -> reg -> valTy -> Prop :=
  | regok_reg : forall (Y : rfTy) (r : reg) (t : valTy),
              RfTy.sel Y r = Some t
              -> regok Y r t.

Inductive ceok : chspec -> dhspec -> word -> cellTy -> Prop :=
  | ceok_junk : forall psi Phi v (t : valTy),
                  ceok psi Phi v (junk, t)
  | ceok_init : forall psi Phi v t,
                  valok psi Phi v t
                   -> ceok psi Phi v (init, t).


(* **************************************************)
(* instruction typing                               *)

Inductive wfis : forall (psi : chspec)
  (Y :rfTy) (is : instrseq), Prop :=

  | wfis_j : 
          forall psi Gamma Gamma'' f,
            CHSpec.lookup psi f Gamma''
            -> codeTy_sub Gamma Gamma''
            -> wfis psi Gamma (ione (j f))

  | wfis_jal : 
          forall psi Y Y' Y'' f fret,
            CHSpec.lookup psi f Y''
            -> CHSpec.lookup psi fret Y'
            -> codeTy_sub (RfTy.upd Y ra (valty_code Y')) Y''
            -> wfis psi Y (ione (jal f fret))

  | wfis_jr : 
          forall psi Gamma Gamma'' r,
            regok Gamma r (valty_code Gamma'')
            -> codeTy_sub Gamma Gamma''
            -> wfis psi Gamma (ione (jr r))

  | wfis_addu : 
          forall psi Gamma rd rs rt I',
            beq_reg rd zero = false
            -> regok Gamma rs valty_int
            -> regok Gamma rt valty_int
            -> wfis psi (RfTy.upd Gamma rd valty_int) I'
            -> wfis psi Gamma (iseq (addu rd rs rt) I')

  | wfis_addiu : 
          forall psi Gamma rd rs w I',
            beq_reg rd zero = false
            -> regok Gamma rs valty_int
            -> wfis psi (RfTy.upd Gamma rd valty_int) I'
            -> wfis psi Gamma (iseq (addiu rd rs w) I')

  | wfis_subu : 
          forall psi Gamma rd rs rt I',
            beq_reg rd zero = false
            -> regok Gamma rs valty_int
            -> regok Gamma rt valty_int
            -> wfis psi (RfTy.upd Gamma rd valty_int) I'
            -> wfis psi Gamma (iseq (subu rd rs rt) I')

  | wfis_subiu : 
          forall psi Gamma rd rs w I',
            beq_reg rd zero = false
            -> regok Gamma rs valty_int
            -> wfis psi (RfTy.upd Gamma rd valty_int) I'
            -> wfis psi Gamma (iseq (subiu rd rs w) I')

  | wfis_move : 
          forall psi Gamma t rd rs I',
            beq_reg rd zero = false
            -> regok Gamma rs t
            -> wfis psi (RfTy.upd Gamma rd t) I'
            -> wfis psi Gamma (iseq (move rd rs) I')


  | wfis_li : 
          forall psi Gamma Gamma'' rd w I',
            beq_reg rd zero = false
            -> CHSpec.lookup psi w Gamma''
            -> wfis psi (RfTy.upd Gamma rd (valty_code Gamma'')) I'
            -> wfis psi Gamma (iseq (li rd w) I')

  | wfis_lw : 
          forall psi Gamma o t rt w rs I',
            beq_reg rt zero = false
            -> regok Gamma rs (valty_tup o)
            -> TupTy.lookup o w (init, t)
            -> wfis psi (RfTy.upd Gamma rt t) I'
            -> wfis psi Gamma (iseq (lw rt w rs) I')

  | wfis_sw :
          forall psi Gamma o ct t rt w rs I',
            beq_reg rs zero = false
            -> regok Gamma rs (valty_tup o)
            -> TupTy.lookup o w ct 
            -> ceTy_sub (init, t) ct
            -> regok Gamma rt t
            -> wfis psi 
                (RfTy.upd Gamma rs 
                (valty_tup (TupTy.upd o w (init, t)))) I'
            -> wfis psi Gamma (iseq (sw rt w rs) I')

  | wfis_blt :
          forall psi Gamma Gamma'' rs rt f I',
            CHSpec.lookup psi f Gamma''
            -> regok Gamma rs valty_int
            -> regok Gamma rt valty_int
            -> codeTy_sub Gamma Gamma''
            -> wfis psi Gamma I'
            -> wfis psi Gamma (iseq (blt rs rt f) I')

  | wfis_beq :
          forall psi Gamma Gamma'' rs rt f I',
            CHSpec.lookup psi f Gamma''
            -> regok Gamma rs valty_int
            -> regok Gamma rt valty_int
            -> codeTy_sub Gamma Gamma''
            -> wfis psi Gamma I'
            -> wfis psi Gamma (iseq (beq rs rt f) I')

  | wfis_bgtz :
          forall psi Gamma Gamma'' rs f I',
            CHSpec.lookup psi f Gamma''
            -> regok Gamma rs valty_int
            -> codeTy_sub Gamma Gamma''
            -> wfis psi Gamma I'
            -> wfis psi Gamma (iseq (bgtz rs f) I').


(* **************************************************)
(* codeheap typing                                  *)

Definition wfch (psi psi' : chspec) (C : codeheap) : Prop :=
  forall (f : label) (Y : rfTy),
    CHSpec.lookup psi' f Y
    -> (exists I : _ , 
        codeheap.getis C f = Some I
        /\ wfis psi Y I).

(* **************************************************)
(* regfile typing                                   *)

Definition wfrf psi phi Y R : Prop :=
  forall r t, 
    RfTy.sel Y r = Some t
    -> valok psi phi (gR R r) t.

Definition wfdh psi phi H : Prop :=
  forall l ct,
    DHSpec.lookup phi l ct
    -> exists v,
        Heap.lookup H l v
        /\ ceok psi phi v ct.

Definition wfst psi Y (s : state) : Prop :=
  let (hp, rf, pc) := s in
    exists phi : dhspec,
      wfdh psi phi hp /\ wfrf psi phi Y rf.

(* **************************************************)
(* program typing                                   *)

Definition wfmach (psi : chspec) (Y : rfTy) (P : machine) 
  : Prop :=
  let (C, S) := P in
    let (H, R, pc) := S in
      wfch psi psi C
      /\ (exists phi : dhspec,
        wfdh psi phi H /\ wfrf psi phi Y R)
      /\ (exists I : instrseq, 
        (codeheap.getis C pc = Some I) 
        /\ wfis psi Y I).

Set Implicit Arguments.
Unset Strict Implicit.

Lemma reg_typing : 
  forall psi phi Y R r t,
    wfrf psi phi Y R
    -> regok Y r t
    -> valok psi phi (Regfile.get R r) t.
Proof.
  intros psi phi Y R r t HR Hr.
  destruct Hr as [Y r t H].
  unfold wfrf in HR.
  auto.
Qed.

Lemma regfile_update :
  forall psi phi Y R r w t,
    beq_reg r zero = false
    -> wfrf psi phi Y R
    -> valok psi phi w t
    -> wfrf psi phi (RfTy.upd Y r t) (Regfile.set R r w).
Proof.
  intros psi phi Y R r w t Hzero HR Hw.
  unfold wfrf in * .
  intros r0 t0.
  assert (HR' := HR r0 t0).
  decbeqreg r0 r; simpl; simplbreg; trivial.
    rewrite (RfTy.sel_upd_req Y r r t 
      Hzero (beq_refl r)).
    intros Ht; injection Ht; intro; subst t0.
    Regfile.simplgs.
    trivial.
  Regfile.simplgs.
    intro Ht.
    rewrite (RfTy.sel_upd_rneq Y r0 r t Hb) in Ht.
    apply (HR' Ht).
Qed.
Implicit Arguments regfile_update
  [psi phi Y R r w t].

(* Lemma wfrf_weaken_Y :  *)
Lemma regfile_weakening :
  forall psi phi Y Y' R,
    wfrf psi phi Y R
    -> rfTy_sub Y Y'
    -> wfrf psi phi Y' R.
Proof.
  intros psi phi Y Y' R HR Hsub.
  unfold wfrf in * .
  intros r t Hr.
  inversion_clear Hsub as [Yx Y'x Hin].
  unfold RfTy.indom in Hin.
  apply HR.
  apply sym_eq.
  assert (Hin' := Hin r).
  rewrite Hr in Hin'.
  apply Hin'.
  exists t.
  trivial.
Qed.
Implicit Arguments regfile_weakening 
   [psi phi Y Y' R].

Lemma heapval_exists:
  forall psi phi H o ct addr offset,
    wfdh psi phi H
    -> TupTy.lookup o offset ct
    -> valok psi phi addr (valty_tup o)
    -> exists v, 
      Heap.lookup H (addr+offset) v.
Proof.
  intros psi phi H o ct0 addr offset.
  intros HH Ht Ho.
  inversion_clear Ho.
  assert (Hin : TupTy.indom o offset).
  unfold TupTy.indom.
  exists ct0; trivial.
  destruct (H0 offset Hin) as [ct [ct' [Hct [ Hct' Hsub]]]].
  assert (He := TupTy.lookup_unique _ _ _ _ Hct Ht).
  subst ct0.
  unfold wfdh in HH.
  clear Hin.
  destruct (HH (addr+offset) ct' Hct') as
    [v [Hin Hceok]].
  exists v.
  trivial.
Qed.
Implicit Arguments heapval_exists
    [psi phi H o ct addr offset].

(* Lemma heap_val_typing :  *)
Lemma heapval_typing :
  forall psi phi H o t addr offset,
    wfdh psi phi H
    -> TupTy.lookup o offset (init, t)
    -> valok psi phi addr (valty_tup o)
    -> exists v, Heap.lookup H (addr+offset) v
      /\ valok psi phi v t.
Proof.
  intros psi phi H o t addr offset.
  intros HH Ht Ho.
  inversion_clear Ho.
  assert (Hin : TupTy.indom o offset).
  unfold TupTy.indom.
  exists (init, t); trivial.
  destruct (H0 offset Hin) as [ct [ct' [Hct [ Hct' Hsub]]]].
  assert (He := TupTy.lookup_unique _ _ _ _ Hct Ht).
  subst ct.
  unfold wfdh in HH.
  clear Hin.
  destruct (HH (addr+offset) ct' Hct') as
    [v [Hin Hceok]].
  exists v.
  split_l.
  inversion Hsub.
  subst ct'; subst ct.
  inversion_clear Hceok.
  trivial.
Qed.
Implicit Arguments heapval_typing 
    [ psi phi H o t addr offset].

Lemma heap_typing_reg_tying_ex_t :
  forall psi phi Y H R r o offset ct,
    wfdh psi phi H
    -> wfrf psi phi Y R
    -> regok Y r (valty_tup o)
    -> TupTy.lookup o offset ct
    -> exists ct',
      DHSpec.lookup phi (Regfile.get R r + offset) ct'
      /\ ceTy_sub ct' ct.
Proof.
  intros psi phi Y H R r o offset ct.
  intros HH HR Hr Hct.
  inversion_clear Hr.
  unfold wfrf in HR.
  substH HR with (HR r (valty_tup o) H0).
  inversion_clear HR.
  assert (Hin : TupTy.indom o offset).
  unfold TupTy.indom.
  exists ct; trivial.
  substH H1 with (H1 offset Hin).
  destruct H1 as [ctx [ctx' [Hctx [Hctx' Hsub]]]].
  exists ctx'.
  split_l.
  rewrite (TupTy.lookup_unique _ _ _ _ Hct Hctx).
  trivial.
Qed.
Implicit Arguments heap_typing_reg_tying_ex_t
    [psi phi Y H R r o offset ct].

Lemma heap_extend_valok : 
  forall psi phi addr ct t v,
    DHSpec.get phi addr = None
    -> valok psi phi v t
    -> valok psi (DHSpec.set phi addr ct) v t.
Proof.
  intros psi phi addr ct t v.
  intros Hnin Hv.
  inversion_clear Hv.
  apply valok_lab; trivial.
  apply valok_int; trivial.
  apply valok_tup.
  intros i Hindom.
  substH H with (H i Hindom).
  destruct H as [ctx [ctx' [Hctx [Hctx' Hsub]]]].
  destruct (nat_eq_dec addr (v + i)) as [Heq | Hneq].
(* case addr = v + i *)
  subst addr.
  rewrite Hctx' in Hnin.
  discriminate Hnin.
(* case addr <> v + i *)
  exists ctx. exists ctx'.
  split_l.
  split.
  apply DHSpec.get_lookup.
  rewrite (DHSpec.set_a_get_a' phi 
    ct Hneq).
  apply (DHSpec.lookup_get Hctx').
  assumption.
Qed.
Implicit Arguments heap_extend_valok
  [psi phi addr ct t v].

Lemma heap_update_valok : 
  forall psi phi phi' addr ct ct' t v,
    DHSpec.lookup phi addr ct
    -> ceTy_sub ct' ct
    -> DHSpec.canupd phi addr ct' phi'
    -> valok psi phi v t
    -> valok psi phi' v t.
Proof.
  intros psi phi phi' addr ct ct' t v.
  intros Hct Hsub Hphi' Hvalok.
  inversion_clear Hvalok.
  apply valok_lab; trivial.
  apply valok_int; trivial.
  apply valok_tup.
  intros i Hindom.
  substH H with (H i Hindom).
  destruct H as [ctx [ctx' [Hctx [Hctx' Hsubx]]]].
  destruct (nat_eq_dec addr (v + i)) as [Heq | Hneq].
(* case addr = v + i *)
  subst addr.
  exists ctx. exists ct'.
  split_l.
  split.
  apply (DHSpec.upd_lookup Hphi').
  assert (Hg := DHSpec.upd_a_get_a Hphi').
  substH Hct with (DHSpec.get_lookup
    (m:=phi) Hct).
  assert (Heq := DHSpec.lookup_unique Hctx' Hct).
  subst ctx'.
  apply (ceTy_sub_trans ct' ct ctx Hsub Hsubx).
(* case addr <> v + i *)
  assert (Hg := DHSpec.upd_a_get_a' Hneq Hphi').
  exists ctx. exists ctx'.
  split_l.
  split.
  apply DHSpec.get_lookup.
  rewrite Hg.
  apply (DHSpec.lookup_get Hctx').
  assumption.
Qed.
Implicit Arguments heap_update_valok
  [psi phi phi' addr ct ct' t v].

Lemma heap_extend_ceok :
  forall psi phi addr ct ct' v,
    DHSpec.get phi addr = None
    -> ceok psi phi v ct
    -> ceok psi (DHSpec.set phi addr ct') v ct.
Proof.
  intros psi phi addr ct ct' v.
  intros Hnin Hceok.
  inversion_clear Hceok.
  apply ceok_junk; trivial.
  apply ceok_init; trivial.
  destruct ct'.
  eapply heap_extend_valok; eauto.
Qed.
Implicit Arguments heap_extend_ceok
  [psi phi addr ct ct' v].
  
Lemma heap_update_ceok :
  forall psi phi phi' addr ct ct' ct'' v,
    DHSpec.lookup phi addr ct
    -> ceTy_sub ct' ct
    -> DHSpec.canupd phi addr ct' phi'
    -> ceok psi phi v ct''
    -> ceok psi phi' v ct''.
Proof.
  intros psi phi phi' addr ct ct' ct0 v.
  intros Hct Hsub Hphi' Hceok.
  inversion_clear Hceok.
  apply ceok_junk; trivial.
  apply ceok_init; trivial.
  eapply heap_update_valok; eauto.
Qed.
Implicit Arguments heap_update_ceok
  [psi phi phi' addr ct ct' ct'' v].
  
Lemma heap_extend_wfdh :
  forall psi phi H ct addr w,
    wfdh psi phi H
    -> DHSpec.get phi addr = None
    -> ceok psi (DHSpec.set phi addr ct) w ct
    -> wfdh psi (DHSpec.set phi addr ct)
            (Heap.set H addr w).
Proof.
  intros psi phi H ct addr w.
  intros HH Hnin Hw.
  unfold wfdh in * .
  intros l ct0 Hct0.
  destruct (nat_eq_dec addr l) as [Heq | Hneq].
(* case l = addr *)
  subst addr.
  exists w.
  substH Hct0 with (DHSpec.lookup_get Hct0).
  rewrite (DHSpec.set_a_get_a phi l ct) in Hct0.
  inj_hyp Hct0.
  subst ct0.
  split_r.
  apply Heap.get_lookup.
  apply Heap.set_a_get_a.
(* case l <> addr *)
  substH Hct0 with (DHSpec.lookup_get Hct0).
  rewrite (DHSpec.set_a_get_a' phi ct Hneq) in Hct0.
  substH Hct0 with (DHSpec.get_lookup Hct0).
  destruct (HH l ct0 Hct0) as [v [Hv Hceok]].
  exists v.
  split.
  apply Heap.get_lookup.
  rewrite <- (Heap.lookup_get Hv).
  apply (Heap.set_a_get_a' H w Hneq).
  apply (heap_extend_ceok (ct':=ct) Hnin Hceok).
Qed.
Implicit Arguments heap_extend_wfdh
    [psi phi H ct addr w].

Lemma heap_update_wfdh :
  forall psi phi phi' H ct ct' addr w,
    wfdh psi phi H
    -> ceTy_sub ct' ct
    -> DHSpec.lookup phi addr ct
    -> DHSpec.canupd phi addr ct' phi'
    -> ceok psi phi' w ct'
    -> wfdh psi phi' (Heap.set H addr w).
Proof.
  intros psi phi phi' H ct ct' addr w.
  intros HH Hsub Hct Hphi' Hw.
  unfold wfdh in * .
  intros l ct0 Hct0.
  destruct (nat_eq_dec addr l) as [Heq | Hneq].
(* case l = addr *)
  subst l.
  assert (Hg := DHSpec.get_lookup (m := phi) Hct).
  substH Hphi' with (DHSpec.upd_lookup Hphi').
  assert (Hcteq := DHSpec.lookup_unique Hct0 Hphi').
  subst ct0.
  exists w.
  split_r.
  apply Heap.get_lookup.
  apply Heap.set_a_get_a.
(* case l <> addr *)
  assert (H0 := DHSpec.upd_a_get_a' Hneq Hphi').
  substH Hct0 with (DHSpec.lookup_get (m := phi') Hct0).
  rewrite Hct0 in H0; clear Hct0.
  substH HH with (HH l ct0 
    (DHSpec.get_lookup (m:=phi) (sym_eq H0))).
  destruct HH as [v [Hv Hceok]].
  exists v.
  split.
  apply Heap.get_lookup.
  rewrite <- (Heap.lookup_get Hv).
  apply (Heap.set_a_get_a' H w Hneq).
  apply (heap_update_ceok Hct Hsub Hphi' Hceok).
Qed.
Implicit Arguments heap_update_wfdh
    [psi phi phi' H ct ct' addr w].

Lemma heap_extend_wfrf :
  forall psi phi Y R ct addr,
    wfrf psi phi Y R
    -> DHSpec.get phi addr = None
    -> wfrf psi (DHSpec.set phi addr ct) Y R.
Proof.
  intros psi phi Y R ct addr.
  intros HR Hnin.
  unfold wfrf in * .
  intros r t Hr.
  apply heap_extend_valok; trivial.
  apply HR; trivial.
Qed.
Implicit Arguments heap_extend_wfrf 
    [psi phi Y R ct addr].

Lemma heap_update_wfrf :
  forall psi phi phi' Y R ct ct' addr,
    wfrf psi phi Y R
    -> DHSpec.lookup phi addr ct
    -> DHSpec.canupd phi addr ct' phi'
    -> ceTy_sub ct' ct
    -> wfrf psi phi' Y R.
Proof.
  intros psi phi phi' Y R ct ct' addr.
  intros HR Hct Hphi' Hsub.
  unfold wfrf in * .
  intros r t Hr.
  apply heap_update_valok with (phi:=phi) (addr:=addr) 
    (ct := ct) (ct' := ct'); trivial.
  apply HR; trivial.
Qed.
Implicit Arguments heap_update_wfrf 
    [psi phi phi' Y R ct ct' addr].

(* Lemma heap_update_valok_strong :  *)
Lemma heap_update_tuple_valok :
  forall psi phi phi' addr offset o ct t,
    valok psi phi addr (valty_tup o)
    -> TupTy.lookup o offset ct
    -> ceTy_sub (init, t) ct
    -> DHSpec.canupd phi (addr + offset) (init, t) phi'
    -> valok psi phi' addr 
           (valty_tup (TupTy.upd o offset (init, t))).
Proof.
  intros psi phi phi' addr offset tup ct t.
  intros Hvalok Htup Hsub Hphi'.
  inversion_clear Hvalok.
  apply valok_tup.
  intros i Hindom. 
  assert (Hindom' : TupTy.indom tup i).
  eapply TupTy.indom_upd_indom; eauto.
  substH H with (H i Hindom').
  destruct H as [ctx [ctx' [Hctx [Hctx' Hsubx]]]].
  destruct (nat_eq_dec i offset) as [Heq | Hneq].
  subst i.
  exists (init, t). exists (init, t).
  split_l by (apply (TupTy.lookup_upd _ _ _ (init, t) Hctx)).
  split_l by (apply (DHSpec.canupd_lookup Hphi')).
  apply ceTy_sub_refl.
  exists ctx. exists ctx'.
  split_l by 
    (apply (TupTy.lookup_upd_neq _ _ _ _ (init, t) Hneq Hctx)).
  split_r.
  assert (Hneq' : addr + i <> addr + offset).
  intro Hf; apply Hneq.
  eapply Arith.Plus.plus_reg_l; eauto.
  apply (DHSpec.canupd_lookup_neq Hneq' Hphi' Hctx').
Qed.
Implicit Arguments heap_update_tuple_valok 
   [psi phi phi' addr offset o ct t].

Lemma heap_update_wfrf_reg :
  forall psi phi phi' Y R o ct ct' t r offset,
    beq_reg r zero = false
    -> wfrf psi phi Y R
    -> regok Y r (valty_tup o)
    -> TupTy.lookup o offset ct
    -> DHSpec.lookup phi 
         (Regfile.get R r + offset) ct'
    -> DHSpec.canupd phi 
         (Regfile.get R r + offset) (init, t) phi'
    -> ceTy_sub ct' ct
    -> ceTy_sub (init, t) ct'
    -> wfrf psi phi' (RfTy.upd Y r 
         (valty_tup (TupTy.upd o offset (init, t)))) R.
Proof.
  intros psi phi phi' Y R o ct ct' t r offset.
  intros Hnr0 HR Hr Hct Hctin Hphi' Hsub Hsub'.
  unfold wfrf in * .
  intros r0 t0 Hr0.
  decbeqreg r0 r.
(* case : r = r0 *)
  rewrite (RfTy.sel_upd_req Y r r 
    (valty_tup (TupTy.upd o offset (init, t))))
    in Hr0; [idtac|solvebreg|solvebreg].
  inj_hyp Hr0.
  subst t0.
  assert (H : valok psi phi (Regfile.get R r) 
      (valty_tup o)).
  inversion_clear Hr.
  apply (HR r (valty_tup o) H).
  assert (Hsub'' : ceTy_sub (init, t) ct).
  eapply ceTy_sub_trans; eauto.
  apply (heap_update_tuple_valok H Hct Hsub'' Hphi').
(* case : r <> r0 *)
  rewrite (RfTy.sel_upd_rneq Y r0 r 
    (valty_tup (TupTy.upd o offset (init, t))) Hb)
    in Hr0.
  apply (heap_update_valok Hctin Hsub' Hphi' 
    (HR r0 t0 Hr0)).
Qed.
Implicit Arguments heap_update_wfrf_reg
    [psi phi phi' Y R o ct ct' t r offset].

(* *********************************************************** *)

Theorem Soundness :
  forall psi (Y : rfTy) P,
    wfmach psi Y P ->
    (exists P' : _, (P ~~> P')
      /\ (exists Y' : _,  wfmach psi Y' P')).
Proof.
  intros psi Y P HP.
  unfold wfmach in HP.
  destruct P as [C [H R pc]].
  destruct HP as [HC [[phi [HH HR]] [I [Hpc HI]]]].
  destruct I.
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rd rs | rd w
      | rt i rs | rt i rs
      | rs rt f | rs rt f | rs f
      | f | f fret | r];
    inversion_clear HI as 
      [ p0 g0 Y'' f0 Hf Hsub
        | p0 g0 Y' Y'' f0 fret0 Hf Hfret Hsub
        | p0 g0 Y'' r0 Hr Hsub
        | p0 g0 r0 r1 r2 i0 Hrdnr0 Hrs Hrt HI'
        | p0 g0 r0 r1 w0 i0 Hrdnr0 Hrs HI'
        | p0 g0 r0 r1 r2 i0 Hrdnr0 Hrs Hrt HI'
        | p0 g0 r0 r1 w0 i0 Hrdnr0 Hrs HI'
        | p0 g0 t r0 r2 i0 Hrdnr0 Hrt HI'
        | p0 g0 Y'' r0 w0 i0 Hrdnr0 Hf HI'
        | p0 g0 o t r0 w0 r1 i0 Hrtnr0 Hrs Ht HI'
        | p0 g0 o ct t r0 w0 r1 i0 Hrsnr0 Hrs Hct Hsub Hrt HI'
        | p0 g0 Y'' r0 r1 f0 i0 Hf Hrs Hrt Hsub HI'
        | p0 g0 Y'' r0 r1 f0 i0 Hf Hrs Hrt Hsub HI'
        | p0 g0 Y'' r0 f0 i0 Hf Hrs Hsub HI'
      ].
  (* case j f *)
  exists (C, <.H, R, f.>).
  inversion_clear Hpc.
  unfold step. 
  rewrite H0.
  simpl.
  split_l.
  inversion_clear Hsub.
  exists Y''.
  split_l.
  split.
  exists phi.
  split_l.
  eapply regfile_weakening; eauto.
  apply (HC f Y'' Hf).
  (* case jal f fret *)
  exists (C, <.H, set R ra fret, f.>).
  inversion_clear Hpc.
  unfold step.
  rewrite H0.
  simpl.
  split_l.
  inversion_clear Hsub.
  exists Y''.
  split_l.
  split.
  exists phi.
  split_l.
  eapply regfile_weakening; eauto.
  apply regfile_update; trivial.
  constructor; trivial.
  apply (HC f Y'' Hf).
  (* case wfis_jr r *)
  exists (C, <.H, R, Regfile.get R r.>).
  inversion_clear Hpc.
  unfold step.
  rewrite H0.
  simpl.
  split_l.
  exists Y''.
  split_l.
  split.
  exists phi.
  split_l.
  eapply regfile_weakening; eauto.
  inversion_clear Hsub; trivial.
  unfold wfch in HC.
  inversion_clear Hr.
  unfold wfrf in HR.
  assert (H3 := HR r (valty_code Y'') H1).
  inversion_clear H3.
  apply (HC (gR R r) Y'' H2).
  (* case iseq *)
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rd rs | rd w
      | rt i rs | rt i rs
      | rs rt f | rs rt f | rs f
      | f | f fret | r];
    inversion_clear HI as 
      [ p0 g0 Y'' f0 Hf Hsub
        | p0 g0 Y' Y'' f0 fret0 Hf Hfret Hsub
        | p0 g0 Y'' r0 Hr Hsub
        | p0 g0 r0 r1 r2 i0 Hrdnr0 Hrs Hrt HI'
        | p0 g0 r0 r1 w0 i0 Hrdnr0 Hrs HI'
        | p0 g0 r0 r1 r2 i0 Hrdnr0 Hrs Hrt HI'
        | p0 g0 r0 r1 w0 i0 Hrdnr0 Hrs HI'
        | p0 g0 t r0 r2 i0 Hrdnr0 Hrt HI'
        | p0 g0 Y'' r0 w0 i0 Hrdnr0 Hf HI'
        | p0 g0 o t r0 w0 r1 i0 Hrtnr0 Hrs Ht HI'
        | p0 g0 o ct t r0 w0 r1 i0 Hrsnr0 Hrs Hct Hsub Hrt HI'
        | p0 g0 Y'' r0 r1 f0 i0 Hf Hrs Hrt Hsub HI'
        | p0 g0 Y'' r0 r1 f0 i0 Hf Hrs Hrt Hsub HI'
        | p0 g0 Y'' r0 f0 i0 Hf Hrs Hsub HI'
      ].
  (* case wfis_addu *)
  exists (C, <.H, sR R rd (gR R rs + gR R rt), (pc+1) .>).
  inversion_clear Hpc.
  unfold step.
  rewrite H0.
  simpl.
  split_l.
  exists (RfTy.upd Y rd valty_int).
  split_l.
  split.
  exists phi.
  split_l.
  eapply regfile_update; eauto; apply valok_int.
  exists I;  split_lr.
  (* case wfis_addiu *)
  exists (C, <.H, sR R rd (gR R rs + w), (pc+1) .>).
  inversion_clear Hpc as [ | | | Hx0 Hx1 Hx2 Hx3 Hi HCI].
  unfold step.
  rew_clear Hi.
  simpl.
  split_l.
  exists (RfTy.upd Y rd valty_int).
  split_l.
  split.
  exists phi.
  split_l.
  eapply regfile_update; eauto; 
    apply valok_int.
  exists I;  split_lr.
  (* case wfis_subu *)
  exists (C, <.H, sR R rd (gR R rs - gR R rt), (pc+1) .>).
  inversion_clear Hpc.
  unfold step.
  rewrite H0.
  simpl.
  split_l.
  exists (RfTy.upd Y rd valty_int).
  split_l.
  split.
  exists phi.
  split_l.
  eapply regfile_update; eauto; apply valok_int.
  exists I;  split_lr.
  (* case wfis_subiu *)
  exists (C, <.H, sR R rd (gR R rs - w), (pc+1) .>).
  inversion_clear Hpc as [ | | | Hx0 Hx1 Hx2 Hx3 Hi HCI].
  unfold step.
  rew_clear Hi.
  simpl.
  split_l.
  exists (RfTy.upd Y rd valty_int).
  split_l.
  split.
  exists phi.
  split_l.
  eapply regfile_update; eauto; apply valok_int.
  exists I;  split_lr.
  (* case wfis_move *)
  exists (C, <.H, sR R rd (gR R rs), (pc+1) .>).
  inversion_clear Hpc as [ | | | Hx0 Hx1 Hx2 Hx3 Hi HCI].
  unfold step.
  rew_clear Hi.
  simpl.
  split_l.
  exists (RfTy.upd Y rd t).
  split_l.
  split.
  exists phi.
  split_l.
  simplbreg.
  Regfile.simplgs.
  simpl.
  inversion_clear Hrt.
  eapply regfile_update; eauto.
  exists I;  split_lr.
  (* case wfis_li *)
  exists (C, <.H, sR R rd w, (pc+1) .>).
  inversion_clear Hpc as [ | | | Hx0 Hx1 Hx2 Hx3 Hi HCI].
  unfold step.
  rew_clear Hi.
  simpl.
  split_l.
  exists (RfTy.upd Y rd (valty_code Y'')).
  split_l.
  split.
  exists phi.
  split_l.
  simplbreg.
  Regfile.simplgs.
  simpl.
  eapply regfile_update; eauto; 
    apply valok_lab; trivial.
  exists I; split_lr.
  (* case wfis_lw *)
  destruct (heapval_typing HH Ht (reg_typing HR Hrs))
    as [v [Hin Hv]].
  exists (C, <.H, sR R rt v, (pc+1) .>).
  inversion_clear Hpc as [ | | | Hx0 Hx1 Hx2 Hx3 Hi HCI].
  unfold step.
  rew_clear Hi.
  simpl.
  rewrite Hin.
  split_l.
  exists (RfTy.upd Y rt t).
  split_l.
  split.
  exists phi.
  split_l.
  apply regfile_update; auto.
  exists I; split_lr.
  (* case wfis_sw *)
  set (H' := sH H (gR R rs + i) (gR R rt)).
  exists (C, <.H', R, (pc+1) .>).
  inversion_clear Hpc as [ | | | Hx0 Hx1 Hx2 Hx3 Hi HCI].
  unfold step.
  rew_clear Hi.
  simpl.
  assert (Hval :=reg_typing HR Hrs).
  destruct (heapval_exists HH Hct Hval)
    as [v Hvin].
  rewrite Hvin.
  split_l.
  exists (RfTy.upd Y rs (valty_tup 
    (TupTy.upd o i (init, t)))).
  destruct (heap_typing_reg_tying_ex_t
    HH HR Hrs Hct) as [ct' [Hct'in Hsub']].
  set (addr := (Regfile.get R rs + i)) in * .
  set (phi' := DHSpec.set phi addr (init, t)).
  split_l.
  split.
    exists phi'.
    assert (Hx := DHSpec.lookup_canupd (m:=phi) (a:=addr) 
      (b:= ct') (init, t) Hct'in).
    assert (Hsubx := ceTy_sub_init_ct t ct ct' Hsub' Hsub).
    split.
    refine (heap_update_wfdh HH Hsubx Hct'in Hx _).
    apply ceok_init.
    eapply reg_typing; eauto.
    eapply heap_update_wfrf; eauto.
    eapply heap_update_wfrf_reg; eauto.
  exists I.
  split_lr.
  (* case wfis_blt *)
  inversion_clear Hpc as [ | | | Hx0 Hx1 Hx2 Hx3 Hi HCI].
  unfold step, nextS.
  rew_clear Hi.
  decbltnat (gR R rs) (gR R rt).
  exists (C, <.H, R, f.>).
  split_l.
  inversion_clear Hsub.
  exists Y''.
  split_l.
  split.
    exists phi.
    split_l.
    eapply regfile_weakening; eauto.
  apply (HC f Y'' Hf).

  exists (C, <.H, R, pc+1.>).
  split_l.
  exists Y.
  split_l.
  split.
    exists phi.
    split_lr.
  exists I; split_lr.
  (* case wfis_beq *)
  inversion_clear Hpc as [ | | | Hx0 Hx1 Hx2 Hx3 Hi HCI].
  unfold step, nextS.
  rew_clear Hi.
  decbeqnat (gR R rs) (gR R rt).
  exists (C, <.H, R, f.>).
  split_l.
  inversion_clear Hsub.
  exists Y''.
  split_l.
  split.
    exists phi.
    split_l.
    eapply regfile_weakening; eauto.
  apply (HC f Y'' Hf).

  exists (C, <.H, R, pc+1.>).
  split_l.
  exists Y.
  split_l.
  split.
    exists phi.
    split_lr.
  exists I; split_lr.
  (* case wfis_bgtz *)
  inversion_clear Hpc as [ | | | Hx0 Hx1 Hx2 Hx3 Hi HCI].
  unfold step, nextS.
  rew_clear Hi.
  decbltnat O (gR R rs).
  exists (C, <.H, R, f.>).
  split_l.
  inversion_clear Hsub.
  exists Y''.
  split_l.
  split.
    exists phi.
    split_l.
    eapply regfile_weakening; eauto.
  apply (HC f Y'' Hf). 
  
  exists (C, <.H, R, pc+1.>).
  split_l.
  exists Y.
  split_l.
  split.
    exists phi.
    split_lr.
  exists I; split_lr.
Qed.


Unset Implicit Arguments.
