(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* Typed Assembly Language                                     *)
(*                                                             *)
(* tal-c3                                                      *)
(***************************************************************)

(* $Id: Ty.v,v 1.1.1.1 2006/12/07 16:43:21 ssg05 Exp $ *)

Require Export machine.
Require Import ListB_MAP.

Require Import extac.

Require Export TyEnv.

Notation tenv := TyEnv.env.

(* **************************************************)
(* flag type                                        *)

Inductive flag : Set :=
  | init : flag
  | junk : flag.

(* **************************************************)
(* value type                                       *)

Inductive valTy : Set :=
  | valTy_var : nat -> valTy
  | valTy_int : valTy
  | valTy_tup : tupTy -> valTy
  | valTy_code : tenv -> rfTy -> valTy
with rfTy : Set := 
  | rfTy_nil : rfTy
  | rfTy_cons : (prod reg valTy) -> rfTy -> rfTy
with tupTy : Set :=
  | tupTy_nil : tupTy
  | tupTy_cons : (prod flag valTy) -> tupTy 
                  -> tupTy.

Implicit Types Y Gamma : rfTy.
Implicit Types t tau : valTy.
Implicit Types o tup : tupTy.

Scheme valTy_ind4 :=
  Induction for valTy Sort Prop
 with rfTy_ind4 :=
  Induction for rfTy Sort Prop
 with tupTy_ind4 :=
  Induction for tupTy Sort Prop.

Section correct_valTy_ind2.

  Variables 
    (P : valTy -> Prop) 
    (P0 : tenv -> rfTy -> Prop)
    (P1 : tupTy -> Prop).
  Hypotheses
    (ftvar : forall n : nat, P (valTy_var n))
    (ftint : P valTy_int)
    (fttup : forall o : tupTy, P1 o -> P (valTy_tup o))
    (ftcode : forall (e:tenv) (Y:rfTy), 
                P0 e Y -> P (valTy_code e Y))
    (frftO : forall (e:tenv), P0 e rfTy_nil)
    (frftS : forall (e:tenv) (r:reg)(t:valTy)(Y:rfTy),
                P t -> P0 e Y -> P0 e (rfTy_cons (r, t) Y))
    (ftuptO : P1 tupTy_nil)
    (ftuptS : forall (f:flag)(t:valTy)(o:tupTy),
                P t -> P1 o -> P1 (tupTy_cons (f, t) o)).

  Fixpoint valTy_ind2 (t : valTy) : P t :=
    match t as t0 return (P t0) with
      | valTy_var n => ftvar n
      | valTy_int => ftint
      | valTy_tup o => fttup o (tupTy_ind2 o)
      | valTy_code e Y => ftcode e Y (rfTy_ind2 e Y)
    end
    with rfTy_ind2 (e : tenv) (Y : rfTy) {struct Y}: P0 e Y :=
    match Y as Y0 return (P0 e Y0) with
      | rfTy_nil => frftO e
      | rfTy_cons (r, t) Y' => 
          frftS e r t Y' (valTy_ind2 t) (rfTy_ind2 e Y')
    end
    with tupTy_ind2 (o : tupTy) : P1 o :=
    match o as o0 return (P1 o0) with
      | tupTy_nil => ftuptO
      | tupTy_cons (f, t) o' => 
          ftuptS f t o' (valTy_ind2 t) (tupTy_ind2 o')
    end.

End correct_valTy_ind2.


Section correct_valTy_ind3.

  Variables 
    (P : valTy -> Prop) 
    (P0 : rfTy -> Prop)
    (P1 : tupTy -> Prop).
  Hypotheses
    (ftvar : forall n : nat, P (valTy_var n))
    (ftint : P valTy_int)
    (fttup : forall o : tupTy, P1 o -> P (valTy_tup o))
    (ftcode : forall (e:tenv) (Y:rfTy), 
                P0 Y -> P (valTy_code e Y))
    (frftO : P0 rfTy_nil)
    (frftS : forall (r:reg)(t:valTy)(Y:rfTy),
                P t -> P0 Y -> P0 (rfTy_cons (r, t) Y))
    (ftuptO : P1 tupTy_nil)
    (ftuptS : forall (f:flag)(t:valTy)(o:tupTy),
                P t -> P1 o -> P1 (tupTy_cons (f, t) o)).

  Fixpoint valTy_ind3 (t : valTy) : P t :=
    match t as t0 return (P t0) with
      | valTy_var n => ftvar n
      | valTy_int => ftint
      | valTy_tup o => fttup o (tupTy_ind3 o)
      | valTy_code e Y => ftcode e Y (rfTy_ind3 Y)
    end
    with rfTy_ind3 (Y : rfTy) : P0 Y :=
    match Y as Y0 return (P0 Y0) with
      | rfTy_nil => frftO
      | rfTy_cons (r, t) Y' => 
          frftS r t Y' (valTy_ind3 t) (rfTy_ind3 Y')
    end
    with tupTy_ind3 (o : tupTy) : P1 o :=
    match o as o0 return (P1 o0) with
      | tupTy_nil => ftuptO
      | tupTy_cons (f, t) o' => 
          ftuptS f t o' (valTy_ind3 t) (tupTy_ind3 o')
    end.

End correct_valTy_ind3.

(***************************************************************)

(* **************************************************)
(* cell type                                        *)

Definition cellTy := prod flag valTy.

(* **************************************************)
(* code type                                        *)

Definition codeTy := prod tenv rfTy.

Notation " ^ n " := (valTy_var n) 
  (at level 40) : tal_scope.
Open Local Scope tal_scope.

Notation " 'int' " := (valTy_int)
  (at level 0) : tal_scope.

Notation " [ e ]- Y " := (valTy_code e Y)
  (at level 40) : tal_scope.

Notation "[{ r , .. , r' }]" := 
  (rfTy_cons r .. (rfTy_cons r' rfTy_nil) ..) 
  (at level 0) : tal_scope.

Notation "[{}]" := (rfTy_nil)  
  (at level 0) : tal_scope.

Notation "[< o , .. , o' >]" := 
  (valTy_tup (tupTy_cons o .. (tupTy_cons o' tupTy_nil) ..)) 
  (at level 0) : tal_scope.

(***************************************************************)

Fixpoint beq_valTy (t t' : valTy) {struct t} : bool :=
  match t, t' with
    | valTy_var n0, valTy_var n1 => beq_nat n0 n1
    | valTy_int, valTy_int => true
    | valTy_tup o, valTy_tup o' => beq_tupTy o o'
    | valTy_code e Y, valTy_code e' Y' =>
        if TyEnv.beq_env e e' 
          then
            beq_rfTy Y Y' 
          else false
    | _, _ => false
  end
with beq_tupTy (o o' : tupTy) {struct o} : bool :=
  match o, o' with
    | tupTy_nil, tupTy_nil => true
    | tupTy_cons (init, t) o, tupTy_cons (init, t') o' =>
        andb (beq_valTy t t') (beq_tupTy o o')
    | tupTy_cons (junk, t) o, tupTy_cons (junk, t') o' =>
        andb (beq_valTy t t') (beq_tupTy o o')
    | _, _ => false
  end
with beq_rfTy (Y Y' : rfTy) {struct Y} : bool :=
  match Y, Y' with
    | rfTy_nil, rfTy_nil => true
    | rfTy_cons (r, t) Y, rfTy_cons (r', t') Y' =>
        if beq_reg r r' 
          then 
            andb (beq_valTy t t') (beq_rfTy Y Y')
          else
            false
    | _, _ => false
  end.

Lemma beq_valTy_dec :
  forall t t',
    beq_valTy t t' = true \/ beq_valTy t t' = false.
Proof.
  intros t.
  elim t using valTy_ind3 with
    (P0 := fun Y =>
      forall Y',
        beq_rfTy Y Y' = true 
        \/ beq_rfTy Y Y' = false)
    (P1 := fun o =>
      forall o',
        beq_tupTy o o' = true
        \/ beq_tupTy o o' = false);
    simpl;
    clear t.
  (* t = var n0 *)
  intros n0.
  intros t; destruct t; simpl.
  decbeqnat n0 n.
  left; trivial.
  right; trivial.
  right; trivial.
  right; trivial.
  right; trivial.
  (* t = int *)
  intros t'; destruct t'; simpl.
  right; trivial.
  left; trivial.
  right; trivial.
  right; trivial.
  (* t = tup o *)
  intros o IHo.
  intros t'; destruct t'; simpl.
  right; trivial.
  right; trivial.
  apply IHo.
  right; trivial.
  (* t = code e0 Y *)
  intros e0 Y IHY.
  intros t'; destruct t'; simpl.
  right; trivial.
  right; trivial.
  right; trivial.
  unfold TyEnv.beq_env.
  decbeqnat e0 e.
  apply IHY.
  right; trivial.
  (* Y = e0 nil *)
  intros Y'; destruct Y'; simpl.
  left; trivial.
  right; trivial.
  (* Y = e0 (r, t)::Y *)
  intros r t Y IHt IHY.
  intros o'; destruct o'; simpl.
  right; trivial.
  destruct p.
  decbeqreg r r0.
  destruct (IHt v); destruct (IHY o').
  rewrite H; rewrite H0; left; trivial.
  rewrite H; rewrite H0; right; trivial.
  rewrite H; rewrite H0; right; trivial.
  rewrite H; rewrite H0; right; trivial.
  right; trivial.
  (* o = nil *)
  intros o'; destruct o'; simpl.
  left; trivial.
  right; trivial.
  (* o = (f, t) :: o *)
  intros f t o IHt IHo.
  intros o'; destruct o'; try destruct p;
    try destruct f0; destruct f; simpl.
  right; trivial.
  right; trivial.
  destruct (IHt v); destruct (IHo o').
  rewrite H; rewrite H0; left; trivial.
  rewrite H; rewrite H0; right; trivial.
  rewrite H; rewrite H0; right; trivial.
  rewrite H; rewrite H0; right; trivial.
  right; trivial.
  right; trivial.
  destruct (IHt v); destruct (IHo o').
  rewrite H; rewrite H0; left; trivial.
  rewrite H; rewrite H0; right; trivial.
  rewrite H; rewrite H0; right; trivial.
  rewrite H; rewrite H0; right; trivial.
Qed.

Lemma beq_valTy_refl : forall t,
  beq_valTy t t = true.
Proof.
  intro t.
  elim t using valTy_ind3 with
    (P0 := fun Y => 
      beq_rfTy Y Y = true)
    (P1 := fun o =>
      beq_tupTy o o = true);
    clear t;
    simpl; trivial; intros; simplbnat; auto.
  unfold TyEnv.beq_env.
  rewrite H; simplbnat; trivial.
  rewrite H; rewrite H0; simplbreg; trivial.
  rewrite H; rewrite H0; destruct f; simpl; trivial.
Qed.

Lemma beq_valTy_sym : forall t t',
  beq_valTy t t' = beq_valTy t' t.
Proof.
  intro t.
  elim t using valTy_ind3 with
    (P0 := fun Y => 
      forall Y', beq_rfTy Y Y' = beq_rfTy Y' Y)
    (P1 := fun o =>
      forall o', beq_tupTy o o' = beq_tupTy o' o);
    clear t;
    simpl; trivial; intros; simplbnat; auto.
  destruct t'; simpl; trivial.
  apply bnat.beq_sym; trivial.
  destruct t'; simpl; trivial.
  destruct t'; simpl; trivial.
  destruct t'; simpl; trivial.
  unfold TyEnv.beq_env.
  rewrite bnat.beq_sym2.
  rewrite H; trivial.
  destruct Y'; try destruct p; simpl; trivial.
  destruct Y'; try destruct p; simpl; trivial.
  rewrite Reg.beq_sym.
  rewrite H.
  rewrite H0; trivial.
  destruct o'; try destruct p; 
    try destruct f; simpl; trivial.
  destruct o'; try destruct p; 
    try destruct f; try destruct f0; simpl; trivial.
  rewrite H; rewrite H0; trivial.
  rewrite H; rewrite H0; trivial.
Qed.

Lemma beq_valTy_eq :
  forall t t',
    beq_valTy t t' = true
    -> t = t'.
Proof.
  intros t.
  elim t using valTy_ind3 with
    (P0 := fun Y =>
      forall Y',
        beq_rfTy Y Y' = true
        -> Y = Y')
    (P1 := fun o =>
      forall o',
        beq_tupTy o o' = true
        -> o = o');
    simpl;
    clear t.
  (* t = var n0 *)
  intros n0.
  intros t' Hb; destruct t';
    try discriminate.
  substbnat n0; trivial.
  (* t = int *)
  intros t' Hb; destruct t';
    try discriminate.
  trivial.
  (* t = tup o *)
  intros o IHo.
  intros t' Hb; destruct t';
    try discriminate.
  rewrite (IHo t Hb); trivial.
  (* t = code e0 Y *)
  intros e0 Y IHY.
  intros t' Hb; destruct t';
    try discriminate.
  unfold beq_env in Hb.
  decbeqnat e0 e.
  substbnat e.
  rewrite (IHY r Hb); trivial.
  (* Y = e0 nil *)
  intros t' Hb; destruct t';
    try discriminate.
  trivial.
  (* Y = e0 (r, t)::Y *)
  intros r t Y IHt IHY.
  intros t' Hb; destruct t'; try destruct p;
    try discriminate.
  decbeqreg r r0.
  destruct (andb_prop _ _ Hb).
  rewrite (IHt v H).
  rewrite (IHY t' H0).
  trivial.
  (* o = nil *)
  intros t' Hb; destruct t';
    try discriminate.
  trivial.
  (* o = (f, t) :: o *)
  intros f t o IHt IHo.
  intros t' Hb; destruct t'; try destruct p;
    try destruct f; try destruct f0;
    try discriminate.
  destruct (andb_prop _ _ Hb).
  rewrite (IHt v H).
  rewrite (IHo t' H0).
  trivial.
  destruct (andb_prop _ _ Hb).
  rewrite (IHt v H).
  rewrite (IHo t' H0).
  trivial.
Qed.

Lemma beq_rfTy_eq :
  forall Y Y',
    beq_rfTy Y Y' = true
    -> Y = Y'.
Proof.
  induction Y; destruct Y'; 
    try destruct p; try destruct p0;
    simpl; intros; trivial.
  discriminate.
  discriminate.
  decbeqreg r r0.
  destruct (andb_prop _ _ H).
  rewrite (IHY Y' H1). 
  rewrite (beq_valTy_eq v v0 H0).
  trivial.
Qed.

Lemma beq_tupTy_eq :
  forall o o',
    beq_tupTy o o' = true
    -> o = o'.
Proof.
  induction o; destruct o'; 
    try destruct p; 
    try destruct p0; 
    try destruct f0;
    try destruct f;
    simpl; intros; trivial;
    try discriminate.
  destruct (andb_prop _ _ H).
  rewrite (IHo o' H1). 
  rewrite (beq_valTy_eq v v0 H0).
  trivial.
  destruct (andb_prop _ _ H).
  rewrite (IHo o' H1). 
  rewrite (beq_valTy_eq v v0 H0).
  trivial.
Qed.

Fixpoint valTy_boute2 (t : valTy) (e e' : tenv) {struct t} : bool :=
  match t with
    | valTy_var n0 => negb (TyEnv.bine2 e e' n0)
    | valTy_int => true
    | valTy_tup o => tupTy_boute2 o e e'
    | valTy_code e0 Y => rfTy_boute2 Y e (TyEnv.app e' e0)
  end
with rfTy_boute2 (Y : rfTy) (e e' : tenv) {struct Y} : bool :=
  match Y with
    | rfTy_nil => true 
    | rfTy_cons (r, t) Y => 
        andb (valTy_boute2 t e e') (rfTy_boute2 Y e e')
  end
with tupTy_boute2 (o : tupTy) (e e' : tenv) {struct o} : bool :=
  match o with
    | tupTy_nil => true 
    | tupTy_cons (f, t) o =>
        andb (valTy_boute2 t e e') (tupTy_boute2 o e e')
  end.

Fixpoint valTy_bine (t : valTy) (e : tenv) {struct t} : bool :=
  match t with
    | valTy_var n0 => TyEnv.bine e n0
    | valTy_int => true
    | valTy_tup o => tupTy_bine o e
    | valTy_code e0 Y => rfTy_bine Y (TyEnv.app e e0)
  end
with rfTy_bine (Y : rfTy) (e : tenv) {struct Y} : bool :=
  match Y with
    | rfTy_nil => true 
    | rfTy_cons (r, t) Y => 
        andb (valTy_bine t e) (rfTy_bine Y e)
  end
with tupTy_bine (o : tupTy) (e : tenv) {struct o} : bool :=
  match o with
    | tupTy_nil => true 
    | tupTy_cons (f, t) o =>
        andb (valTy_bine t e) (tupTy_bine o e)
  end.

Lemma valTy_boute2_dec :
  forall t e e',
  valTy_boute2 t e e' = true 
  \/ valTy_boute2 t e e' = false.
Proof.
  intros t.
  elim t using valTy_ind3 with
    (P0 := fun Y =>
      forall e e',
        rfTy_boute2 Y e e' = true 
        \/ rfTy_boute2 Y e e' = false)
    (P1 := fun o =>
      forall e e',
        tupTy_boute2 o e e' = true
        \/ tupTy_boute2 o e e' = false);
    simpl;
    clear t.
  (* t = var n0 *)
  intros n0.
  intros e e'.
  unfold TyEnv.bine2.
  decbltnat n0 (TyEnv.len e + TyEnv.len e');
    decbltnat  n0 (TyEnv.len e'); simpl.
  left; trivial.
  right; trivial.
  left; trivial.
  left; trivial.
  (* t = int *)
  intros e e'.
  left; trivial.
  (* t = tup o *)
  intros o IHo.
  intros e e'.
  apply (IHo e e').
  (* t = code e0 Y *)
  intros e0 Y IHY.
  intros e e'.
  apply IHY.
  (* Y = e0 nil *)
  intros e e'.
  left; trivial.
  (* Y = e0 (r, t)::Y *)
  intros r t Y IHt IHY.
  intros e e'.
  destruct (IHt e e'); destruct (IHY e e'); 
    rewrite H; rewrite H0; simpl.
  left; trivial.
  right; trivial.
  right; trivial.
  right; trivial.
  (* o = nil *)
  intros e e'.
  left; trivial.
  (* o = (f, t) :: o *)
  intros f t o IHt IHo.
  intros e e'.
  destruct (IHt e e'); destruct (IHo e e');
    rewrite H; rewrite H0; simpl.
  left; trivial.
  right; trivial.
  right; trivial.
  right; trivial.
Qed.

Lemma rfTy_boute2_dec :
  forall Y e e',
  rfTy_boute2 Y e e' = true 
  \/ rfTy_boute2 Y e e' = false.
Proof.
  induction Y; try destruct p; simpl; intros.
  left; trivial.
  destruct (IHY e e') as [HY | HY]; rewrite HY;
    destruct (valTy_boute2_dec v e e') as [Hv | Hv];
    rewrite Hv.
  left; trivial.
  right; trivial.
  right; trivial.
  right; trivial.
Qed.

Lemma tupTy_boute2_dec :
  forall o e e',
  tupTy_boute2 o e e' = true 
  \/ tupTy_boute2 o e e' = false.
Proof.
  induction o; try destruct p; 
    try destruct f; simpl; intros.
  left; trivial.
  destruct (IHo e e') as [Ho | Ho]; rewrite Ho;
    destruct (valTy_boute2_dec v e e') as [Hv | Hv];
    rewrite Hv.
  left; trivial.
  right; trivial.
  right; trivial.
  right; trivial.
  destruct (IHo e e') as [Ho | Ho]; rewrite Ho;
    destruct (valTy_boute2_dec v e e') as [Hv | Hv];
    rewrite Hv.
  left; trivial.
  right; trivial.
  right; trivial.
  right; trivial.
Qed.

(* **************************************************)
(* regfile type                                     *)

Module RfTy.

Fixpoint sel (m : rfTy) (r : reg) {struct m} : option valTy :=
  match m with 
    | rfTy_nil => None
    | rfTy_cons (r', b') m' =>
      match beq_reg r r' with
        | true => Some b'
        | false => sel m' r
      end
  end.

Fixpoint upd (m : rfTy) (r : reg) (b : valTy) {struct m} : rfTy :=
  match m return rfTy with
    | rfTy_nil => rfTy_cons (r, b) rfTy_nil
    | rfTy_cons (r', b') m' =>
        match beq_reg r r' with
          | true => rfTy_cons (r, b) m'
          | false => rfTy_cons (r', b') (upd m' r b)
        end
  end.

Fixpoint sort (Y Y' : rfTy) {struct Y'} : option rfTy :=
  match Y' with
    | rfTy_nil => Some rfTy_nil
    | rfTy_cons (r', t') Y' =>
        match RfTy.sel Y r' with
          | None => None
          | Some t => 
            match sort Y Y' with
              | None => None
              | Some Y => Some (rfTy_cons (r', t) Y)
            end
        end
  end.

Lemma sel_dec : 
  forall Y r,
    (exists t, sel Y r = Some t)
    \/ sel Y r = None.
Proof.
  induction Y; try destruct p; simpl; intros; trivial.
  right; trivial.
  decbeqreg r0 r.
  left; exists v; trivial.
  apply IHY.
Qed.

Lemma sort_dec : 
  forall Y Y',
    (exists Y'', sort Y Y' = Some Y'')
    \/ sort Y Y' = None.
Proof.
  intros Y Y'; generalize Y; clear Y.
  induction Y'; try destruct p; simpl; intros; trivial.
  left; exists rfTy_nil; trivial.
  destruct (sel_dec Y r) as [[t Hs] | Hs]; 
    rewrite Hs; try(right; trivial; fail).
  destruct (sort Y Y').
  left; exists (rfTy_cons (r, t) r0); trivial.
  right; trivial.
Qed.

Definition indom (m : rfTy) (r : reg) := 
  exists b : valTy, sel m r = Some b.

Definition lookup (m : rfTy) (r : reg) (t : valTy) := 
  sel m r = Some t.

Lemma lookup_nil :
  forall r t,
    lookup rfTy_nil r t -> False.
Proof.
  intros r t H.
  unfold lookup in H.
  unfold sel in H.
  discriminate.
Qed.

Lemma indom_nil :
  forall r,
    indom rfTy_nil r -> False.
Proof.
  unfold indom; simpl; intros r [b H]; discriminate.
Qed.

Lemma sel_upd_req :
  forall Y r r' t,
    beq_reg r r' = true
    -> sel (upd Y r' t) r = Some t.
Proof.
  intros Y r r' t H.
  induction Y as [|(r0, t0) Y IHY].
  simpl.
    rewrite H.
    trivial.
  simpl.
    decbeqreg r' r0; simpl; simplbreg; trivial.
Qed.

Lemma sel_upd_rneq :
  forall Y r r' t,
    beq_reg r r' = false
    -> sel (upd Y r' t) r = sel Y r.
Proof.
  intros Y r r' t H.
  induction Y as [|(r0, t0) Y IHY].
  simpl.
    rewrite H.
    trivial.
  simpl.
    decbeqreg r' r0; simpl; simplbreg; trivial.
    decbeqreg r r0; simpl; simplbreg; trivial.
Qed.

Lemma sort_sel :
  forall Y Y' Y'' r,
    indom Y' r
    -> sort Y Y' = Some Y''
    -> sel Y r = sel Y'' r.
Proof.
  intros Y Y'; generalize Y; clear Y.
  induction Y'; try destruct p; simpl; intros.
  destruct (indom_nil r H).
  destruct (sel_dec Y r) as [[t Hs] | Hs];
    rewrite Hs in H0; try discriminate.
  destruct (sort_dec Y Y') as [[Yx HY] | HY];
    rewrite HY in H0; try discriminate.
  inj_hyp H0; subst Y''.
  simpl.
  unfold indom in H.
  simpl in H.
  decbeqreg r0 r; trivial.
  apply IHY'; auto.
Qed.

Lemma sort_sel_2 :
  forall Y Y' Y'' r,
    indom Y'' r
    -> sort Y Y' = Some Y''
    -> sel Y r = sel Y'' r.
Proof.
  intros Y Y'; generalize Y; clear Y.
  induction Y'; try destruct p; simpl; intros.
  inj_hyp H0; subst Y''.
  destruct (indom_nil r H).
  destruct (sel_dec Y r) as [[t Hs] | Hs];
    rewrite Hs in H0; try discriminate.
  destruct (sort_dec Y Y') as [[Yx HY] | HY];
    rewrite HY in H0; try discriminate.
  inj_hyp H0; subst Y''.
  simpl.
  unfold indom in H.
  simpl in H.
  decbeqreg r0 r; trivial.
  apply IHY'; auto.
Qed.

End RfTy.

(* **************************************************)
(* tuple type                                       *)

Module TupTy.

Fixpoint mem (tup : tupTy) (ct : cellTy) {struct tup} : Prop := 
  match tup with
    | tupTy_nil => False
    | tupTy_cons ct' tup' => ct = ct' \/ mem tup' ct
  end.

Fixpoint lookup (tup : tupTy) (n:nat) (ct:cellTy) 
  {struct n} : Prop :=
  match n, tup with
    | O, tupTy_cons x l' => x = ct
    | O, tupTy_nil => False
    | S m, tupTy_nil => False
    | S m, tupTy_cons x l' => lookup l' m ct
  end.

Fixpoint get (tup : tupTy) (n : nat) 
  {struct n} : option cellTy :=
  match n, tup with
    | O, tupTy_cons ct l' => Some ct
    | O, tupTy_nil => None
    | S m, tupTy_nil => None
    | S m, tupTy_cons x l' => get l' m
  end.
  
Fixpoint upd (tup : tupTy) (n : nat) (ct : cellTy) 
  {struct n}: tupTy :=
  match n, tup with 
    | O, tupTy_cons ct' l' => tupTy_cons ct l'
    | S m, tupTy_cons ct' l' => tupTy_cons ct' (upd l' m ct)
    | _, _ => tupTy_nil
  end.

Definition indom (tup : tupTy) (n : nat) := 
  exists ct : cellTy, lookup tup n ct.

Lemma lookup_nil : 
  forall n ct,
    lookup tupTy_nil n ct -> False.
Proof.
  intros n ct.
  intros Hf.
  induction n; simpl; intros; auto.
Qed.

Lemma lookup_get : forall n l ct,
  lookup l n ct -> get l n = Some ct.
Proof.
  induction n; destruct l; 
    simpl; intros;
      tauto || auto.
  subst; trivial.
Qed.

Lemma get_lookup : forall n l ct,
  get l n = Some ct -> lookup l n ct.
Proof.
  induction n; destruct l; 
    simpl; intros;
      discriminate || auto.
  injection H; intro; trivial.
Qed.

Lemma lookup_upd_indom :
  forall n l n' (ct ct' : cellTy),
    lookup (upd l n ct) n' ct' -> indom l n'.
Proof.
  unfold indom;
  induction n; destruct l; destruct n'; 
    simpl; intros;
    tauto || auto.
  exists p; trivial.
  exists ct'; trivial.
  exists p; trivial.
  eapply IHn; eauto.
Qed.

Lemma upd_a_lookup_a : 
  forall n l (ct ct' : cellTy),
    lookup (upd l n ct') n ct -> ct = ct'.
Proof.
  induction n; destruct l; simpl; intros;
    try tauto.
  subst ct; trivial.
  eapply IHn; eauto.
Qed.

Lemma upd_a_lookup_a' : 
  forall addr addr' l (ct ct' : cellTy),
    addr <> addr'
    -> lookup (upd l addr ct') addr' ct 
       = lookup l addr' ct.
Proof.
  induction addr; destruct l; simpl in * ;
    try tauto.
  induction addr'; intros.
  destruct (H (refl_equal _)).
  simpl.
  trivial.
  intros.
  induction addr'; intros.
  simpl. trivial.
  simpl.
  apply IHaddr.
  intro.
  apply H.
  rewrite H0.
  trivial.
Qed.

Lemma lookup_unique :
  forall addr l  (ct ct' : cellTy),
    lookup l addr ct 
    -> lookup l addr ct'
    -> ct = ct'.
Proof.
  induction addr; destruct l; simpl; intros .
  destruct H.
  subst p; trivial.
  destruct H.
  apply (IHaddr l ct ct' H H0).
Qed.

Lemma lookup_upd : 
  forall addr l (ct ct' : cellTy),
    lookup l addr ct
    -> lookup (upd l addr ct') addr ct'.
Proof.
  induction addr; destruct l; simpl; intros; trivial.
  eapply IHaddr; eauto.
Qed.

Lemma lookup_upd_neq :
  forall addr l addr' (ct ct' : cellTy),
    addr <> addr'
    -> lookup l addr ct
    -> lookup (upd l addr' ct') addr ct.
Proof.
  induction addr; destruct l; simpl; intros; 
    tauto || trivial.
  subst p.
  destruct addr'.
  tauto.
  simpl in * .
  trivial.
  destruct addr'.
  simpl.
  assumption.
  simpl.
  apply IHaddr.
  intro Hf.
  apply H.
  subst addr; trivial.
  trivial.
Qed.

Lemma indom_indom_upd :
  forall n (tup : tupTy) (n': nat) (ct : cellTy),
    indom tup n -> indom (upd tup n' ct) n.
Proof.
  unfold indom.
  intros n tup n' ct [ct' H].
  destruct (nat_eq_dec n n') as [Heq | Hneq].
  subst n'.
  exists ct.
  eapply lookup_upd; eauto.
  exists ct'.
  assert (H1 : n' <> n).
  intro H0; apply (Hneq (sym_eq H0)).
  apply lookup_upd_neq; trivial.
Qed.

Lemma indom_upd_indom :
  forall n (tup : tupTy) (n': nat) (ct : cellTy),
    indom (upd tup n' ct) n -> indom tup n.
Proof.
  intros n tup n' ct [ct' H].
  eapply lookup_upd_indom; eauto.
Qed.

End TupTy.

(* ************************************************ *)
(* Psi : CHSpec                                     *)
(* Phi : DHSPec                                     *)

Module CdSpec <: ONE_SET.
  Definition B := codeTy.
End CdSpec.
Module CHSpecDef := ListMap CdSpec.
Module CHSpec := MapEx CHSpecDef.
(* psi : chspec *)
Definition chspec := CHSpec.map.

Implicit Types Psi psi: chspec.

Module DtSpec <: ONE_SET.
  Definition B := cellTy.
End DtSpec.
Module DHSpecDef := ListMap DtSpec.
Module DHSpec := MapEx DHSpecDef.
(* psi : chspec *)
Definition dhspec := DHSpec.map.

Implicit Types Phi phi: dhspec.
