(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* Typed Assembly Language                                     *)
(*                                                             *)
(* tal-c3                                                      *)
(***************************************************************)

Require Import bnat.
Require Import natpm.
Require Import extac.

Notation Local A := nat.

Definition env : Set := nat.

Definition app (e e' : env) : env :=
  e + e'.

Definition tail (e : env) : env :=
  e - 1.

Definition len (e : env) : nat :=
  e.

Definition head (e : env) : nat :=
  e - 1.

Definition ine (e : env) (a : A) : Prop :=
  blt_nat a (len e) = true.

Definition emp : env := O.

Definition sigl (a : A) : env := 1.

Definition beq_env (e e' : env) : bool :=
  beq_nat e e'.

Definition bine (e : env) (a : A) : bool :=
  blt_nat a (len e).

Definition bine2 (e e' : env) (a : A) : bool :=
  andb (blt_nat a (len e + len e'))
  (negb (blt_nat a (len e'))).

Lemma ine_blt :
  forall e n,
    ine e n -> blt_nat n (len e) = true.
Proof.
  intros; auto.
Qed.

Lemma blt_ine :
  forall e n,
    blt_nat n (len e) = true -> ine e n.
Proof.
  intros; auto.
Qed.

Lemma bine_ine :
  forall e n,
    blt_nat n (len e) = true -> ine e n.
Proof.
  intros; auto.
Qed.

Lemma ine_weaken_l : forall e e' a,
  ine e a -> ine (app e' e) a.
Proof.
  unfold app, ine, len.
  intros; solvebnat.
Qed.

Lemma ine_weaken_r : forall e e' a,
  ine e a -> ine (app e e') a.
Proof.
  unfold app, ine, len.
  intros; solvebnat.
Qed.

Lemma app_assoc_l :
  forall e e' e'',
    app (app e e') e'' = app e (app e' e'').
Proof.
  unfold app, ine, len.
  intros.
  auto with arith.
Qed.

Lemma len_app :
  forall e e',
    len (app e e') = len e + len e'.
Proof.
  unfold app, ine, len.
  intros.
  auto with arith.
Qed.

Lemma app_emp_r :
  forall e,
    TyEnv.app e emp = e.
Proof.
  intros e.
  unfold app.
  auto with arith.
Qed.

Lemma app_emp_l :
  forall e,
    TyEnv.app emp e = e.
Proof.
  intros e.
  unfold app.
  auto with arith.
Qed.

Lemma len_emp : len emp = O.
Proof.
  unfold len, emp; auto with arith.
Qed.

Lemma len_sigl : forall a, len (sigl a)= 1.
Proof.
  unfold len, sigl; auto with arith.
Qed.

Lemma env_ind : forall e,
  e = emp \/ exists e', exists a, e = app (sigl a) e'.
Proof.
  induction e.
  left; trivial.
  destruct IHe.
  right.
  exists emp.
  exists 1.
  subst e; simpl; trivial.
  right.
  destruct H as [e' [a  He]].
  subst e.
  exists (S e').
  exists 1.
  simpl; trivial.
Qed.

Ltac simplenv :=
  match goal with
    | H : context [len emp] |- _ =>
      rewrite (len_emp) in H; simplenv
    | |- context [len emp] =>
      rewrite (len_emp); simplenv

    | H : context [len (sigl ?a)] |- _ =>
      rewrite (len_sigl a) in H; simplenv
    | |- context [len (sigl ?a)] =>
      rewrite (len_sigl a); simplenv

    | H : context [len (app ?e ?e')] |- _ =>
      rewrite (len_app e e') in H; simplenv
    | |- context [len (app ?e ?e')] =>
      rewrite (len_app e e'); simplenv

    | H : context [app ?e emp] |- _ =>
      rewrite (app_emp_r e) in H; simplenv
    | |- context [app ?e emp] =>
      rewrite (app_emp_r e); simplenv
    | H : context [app emp ?e] |- _ =>
      rewrite (app_emp_l e) in H; simplenv
    | |- context [app emp ?e] =>
      rewrite (app_emp_l e); simplenv

    | H : context [app emp ?e] |- _ =>
      rewrite (app_emp_l e) in H; simplenv
    | |- context [app emp ?e] =>
      rewrite (app_emp_l e); simplenv

    | H : context [app (app ?e ?e') ?e''] |- _ =>
      rewrite (app_assoc_l e e' e'') in H; simplenv
    | |- context [app (app ?e ?e') ?e''] =>
      rewrite (app_assoc_l e e' e''); simplenv

    | _ => idtac
  end.

Ltac inv_ine H :=
  match type of H with
    | ine ?e ?n =>
      let Hblt := fresh "Hblt" in
      (generalize (ine_blt e n H); intro Hblt)
    | _ => fail
  end.

Ltac substenv e :=
  match goal with 
    | H : beq_env ?e' e = true |- _ =>
      unfold beq_env in H;
      substbnat e
    | H : beq_env e ?e' = true |- _ =>
      unfold beq_env in H;
      substbnat e
  end.
