
Axiom extensionality:
  forall (A B: Set) (f g : A -> B),
  (forall x, f x = g x) -> f = g.

(*Axiom extensional_eq : forall (A B : Set) (f g: A -> B), (forall x : A, f x = g x) -> f = g. *)

Section Map.

Set Implicit Arguments.

Variables A B : Set.

Variable beq_A : forall (x y : A), {x = y} + {x <> y}.


(* partial map *)
Definition Map := A -> option B.

(* empty map *)
Definition emp : Map :=
  fun a : A => None.


(* map extension *)
Definition put : Map -> A -> B -> Map :=
  fun M a b =>
    fun x => if (beq_A x a) then (Some b) else M x.


Definition remove : Map -> A -> Map :=
  fun M a =>
    fun x => if (beq_A x a) then None else M x.


Definition maps_to : Map -> A -> B -> Prop :=
  fun M a b => M a = Some b.


Definition in_dom : A -> Map -> Prop :=
  fun a M => exists b : B, M a = Some b.


Definition not_in_dom : A -> Map -> Prop :=
  fun a M => M a = None.


(* map that has only one element in the domain *)
Definition sgltonM : Map -> A -> B -> Prop :=
  fun M a b => 
    maps_to M a b /\ (forall (x : A), x <> a -> not_in_dom x M).


Definition sub_dom : Map -> Map -> Prop :=
  fun M M' => forall (x : A), in_dom x M -> in_dom x M'.


Definition eqDom : Map -> Map -> Prop :=
  fun M M' => sub_dom M M' /\ sub_dom M' M.


Definition merge : Map -> Map -> Map :=
  fun M1 M2 =>
    fun x =>
      match M1 x with
        None => M2 x
      | _ => M1 x
      end.


Definition disjoint : Map -> Map -> Prop :=
  fun M1 M2 =>
    forall (x : A), not_in_dom x M1 \/ not_in_dom x M2.



Definition disjUnion : Map -> Map -> Map -> Prop :=
  fun M' M'' M => disjoint M' M'' /\ (M = merge M' M'').


Definition subseteq : Map -> Map -> Prop :=
  fun M1 M2 =>
    forall (x : A), in_dom x M1 -> M1 x = M2 x.


Definition compat : Map -> Map -> Prop :=
  fun M1 M2 =>
    forall (x : A) (y : B), maps_to M1 x y -> not_in_dom x M2 \/ maps_to M2 x y.


(* extensional equality about Maps *)
Definition extensional_eqMap := extensionality A (option B).



(* ****************************** *)
(* properties about map extension *)
(* ****************************** *)

Lemma put_get_eq :
  forall (M : Map) (x: A) (y : B), (put M x y) x = Some y.
Proof.
intros.
unfold put.
case (beq_A x x).
auto.
intro.
assert (x=x). auto. contradiction.
Qed.


Lemma pub_indepedent:
  forall (M : Map) (x x' : A) (y : B), x <> x' -> (put M x y) x' = M x'.
Proof.
intros M x x' y x_neq.
unfold put.
case (beq_A x' x).
intro. assert (x = x'). symmetry. auto. contradiction.
auto.
Qed.


Lemma remove_shrinks_dom:
  forall (M : Map) (x : A),
    forall (y : A), not_in_dom y M -> not_in_dom y (remove M x).
Proof.
  intros.
  unfold not_in_dom.
  unfold remove.
  induction (beq_A y x).
  auto.
  unfold not_in_dom in H. auto.
Qed.


(******************************************)
(* properties about in_dom and not_in_dom *)
(******************************************)

Lemma in_notin_exclusive_left :
  forall (x : A) (M : Map), in_dom x M -> ~ (not_in_dom x M).
Proof.
  intros.
  intro.
  unfold in_dom in H. unfold not_in_dom in H0.
  inversion_clear H.
  rewrite H0 in H1. inversion_clear H1.
Qed.


Lemma in_notin_exclusive_right :
  forall (x : A) (M : Map), not_in_dom x M -> ~ (in_dom x M).
Proof.
  intros.
  intro.
  unfold in_dom in H0. unfold not_in_dom in H.
  inversion_clear H0.
  rewrite H in H1. inversion_clear H1.
Qed.


Lemma not_in_notin:
  forall (x : A) (M : Map), ~ in_dom x M -> not_in_dom x M.
Proof.
  intros.
  unfold not_in_dom.
  unfold in_dom in H.
  induction (M x).
  assert (exists b : B, Some a = Some b). exists a; auto.
  contradiction.
  auto.
Qed.


Lemma not_notin_in:
  forall (x : A) (M : Map), ~ not_in_dom x M -> in_dom x M.
Proof.
  intros.
  unfold in_dom. unfold not_in_dom in H.
  induction (M x). exists a. auto. elim H. auto.
Qed.


Lemma in_or_notin_dom:
  forall x M, in_dom x M \/ not_in_dom x M.
Proof.
  intros x M.
  unfold in_dom.
  unfold not_in_dom.
  elim (M x).
  left. exists a. auto.
  right. auto.
Qed.


(* ******************************** *)
(* Disjoint and in_dom / not_in_dom *)
(* ******************************** *)
Lemma disj_exclusive_left :
  forall (x : A) (M M' : Map), 
    disjoint M M' -> in_dom x M -> not_in_dom x M'.
Proof.
  intros.
  unfold disjoint in H.
  assert (not_in_dom x M \/ not_in_dom x M'). auto.
  inversion_clear H1.
  assert (~ in_dom x M). apply in_notin_exclusive_right; auto. contradiction.
  auto.
Qed.


(* properties about disjoint *)
Lemma disjoint_commut: forall M1 M2, disjoint M1 M2 -> disjoint M2 M1.
Proof.
  intros M1 M2.
  unfold disjoint.
  intros.
  assert (not_in_dom x M1 \/ not_in_dom x M2); auto.
  inversion_clear H0.
  right; auto.
  left; auto.
Qed.


Lemma extend_presv_disj:
  forall (M1 M2 : Map) (x : A) (y : B),
    disjoint M1 M2 -> not_in_dom x M2 -> disjoint (put M1 x y) M2.
Proof.
  intros.
  unfold disjoint.
  unfold not_in_dom in H0.
  unfold disjoint in H.
  intro.
  assert (not_in_dom x0 M1 \/ not_in_dom x0 M2).
  apply H.
  inversion_clear H1.
  unfold put.
  unfold not_in_dom.
  induction (beq_A x0 x).
  right. rewrite a. auto.
  left. unfold not_in_dom in H2. auto.
  right. auto.
Qed.

Lemma update_presv_disj:
  forall (M1 M2 : Map) (x : A) (y : B),
    disjoint M1 M2 -> in_dom x M1 -> disjoint (put M1 x y) M2.
Proof.
  intros.
  assert (not_in_dom x M2).
  apply disj_exclusive_left with M1; auto.
  apply extend_presv_disj; auto.
Qed.

Lemma remove_presv_disj:
  forall (M1 M2 : Map) (x : A),
    disjoint M1 M2 -> disjoint (remove M1 x) M2.
Proof.
  intros.
  unfold disjoint.
  unfold disjoint in H.
  intro.
  assert (not_in_dom x0 M1 \/ not_in_dom x0 M2).
  apply H.
  inversion_clear H0.
  left.
  apply remove_shrinks_dom; auto.
  right. auto.
Qed.

(* ********************* *)
(* Propeties about merge *)
(* ********************* *)
Lemma merge_assoc: 
  forall M1 M2 M3, merge (merge M1 M2) M3 = merge M1 (merge M2 M3).
Proof.
intros M1 M2 M3.
apply extensional_eqMap.
intro.
unfold merge.
elim (M1 x).
auto.
elim (M2 x); auto.
Qed.


Lemma put_merge_merge_put:
  forall (M1 M2 : Map) (x : A) (y : B),
    put (merge M1 M2) x y = merge (put M1 x y) M2.
Proof.
  intros M1 M2 x y.
  apply extensional_eqMap.
  intro.
  unfold put. unfold merge.
  induction (beq_A x0 x). auto. induction (M1 x0); auto.
Qed.

Lemma disj_remove_merge_merge_remove :
  forall (M1 M2 : Map) (x : A),
    disjoint M1 M2 -> in_dom x M1 -> remove (merge M1 M2) x = merge (remove M1 x) M2.
Proof.
  intros.
  assert (not_in_dom x M2).
  apply disj_exclusive_left with M1; auto.
  apply extensional_eqMap.
  intro.
  unfold remove. unfold merge.
  induction (beq_A x0 x).
  rewrite a. unfold not_in_dom in H1. auto. 
  auto.
Qed.


(* ******************************** *)
(*  Merge and in_dom / not_in_dom   *)
(* ******************************** *)
Lemma notin_merge_distr_left : 
  forall (M1 M2 : Map) (x : A), 
    not_in_dom x (merge M1 M2) -> not_in_dom x M1.
Proof.
intros M1 M2 x.
unfold not_in_dom.
unfold merge.
elim (M1 x); auto.
Qed.


Lemma notin_merge_distr_right : 
  forall (M1 M2 : Map) (x : A), 
    not_in_dom x (merge M1 M2) -> not_in_dom x M2.
Proof.
intros M1 M2 x.
unfold not_in_dom.
unfold merge.
elim (M1 x). 
intros.
inversion_clear H.
auto.
Qed.


Lemma merge_presv_in_dom:
  forall (M1 M2: Map) (x: A), in_dom x M1 -> in_dom x (merge M1 M2).
Proof.
  intros M1 M2 x.
  unfold in_dom.
  unfold merge.
  intro.
  inversion_clear H.
  exists x0.
  rewrite H0.
  auto.
Qed.


Lemma merge_presv_in_dom_right:
  forall (M1 M2: Map) (x: A), in_dom x M2 -> in_dom x (merge M1 M2).
Proof.
  intros.
  unfold in_dom. unfold in_dom in H.
  unfold merge.
  assert ((exists y, M1 x = Some y) \/ (M1 x = None)).
  induction (M1 x).
  left. exists a. auto. right. auto.
  inversion_clear H0.
  inversion_clear H1.
  exists x0. rewrite H0. auto.
  inversion_clear H.
  exists x0. rewrite H1. auto.
Qed.

(* *************************************** *)
(*              merge and subseteq         *)
(* *************************************** *)
Lemma merge_extends_map:
  forall (M1 M2 : Map), subseteq M1 (merge M1 M2).
Proof.
  intros.
  unfold subseteq.
  unfold merge.
  intros.
  unfold in_dom in H.
  inversion H.
  rewrite H0.
  auto.
Qed.

(* *************************************** *)
(*  Properties about disjoint and  merge   *)
(* *************************************** *)
Lemma disj_distr_left: 
  forall M1 M2 M' : Map, disjoint (merge M1 M2) M' -> disjoint M1 M'.
Proof.
intros M1 M2 M'.
unfold disjoint.
intros.
assert (not_in_dom x (merge M1 M2) \/ not_in_dom x M'); auto.
inversion_clear H0.
left.
apply notin_merge_distr_left with M2; auto.
right. auto.
Qed.


Lemma disj_distr_right: 
  forall M1 M2 M' : Map, disjoint (merge M1 M2) M' -> disjoint M2 M'.
Proof.
intros M1 M2 M'.
unfold disjoint.
intros.
assert (not_in_dom x (merge M1 M2) \/ not_in_dom x M'); auto.
inversion_clear H0.
left.
apply notin_merge_distr_right with M1; auto.
right. auto.
Qed.



Lemma disj_merge_commut : forall M1 M2 : Map, disjoint M1 M2 -> (merge M1 M2) = (merge M2 M1).
Proof.
intros M1 M2 disj.
apply extensional_eqMap.
intro.
unfold merge.
unfold disjoint in disj.
unfold not_in_dom in disj.
assert (M1 x = None \/ M2 x = None); auto.
inversion H.
rewrite H0.
elim (M2 x); auto.
rewrite H0.
elim (M1 x); auto.
Qed.


Lemma disj_merge_assoc:
  forall M1 M2 M3, disjoint M1 M2 -> disjoint M1 M3 -> disjoint M1 (merge M2 M3).
Proof.
  intros M1 M2 M3. 
  unfold disjoint.
  intros disj1 disj2.
  intro.
  assert (not_in_dom x M1 \/ not_in_dom x M2); auto.
  assert (not_in_dom x M1 \/ not_in_dom x M3); auto.
  inversion_clear H.
  left; auto.
  inversion_clear H0.
  left; auto.
  right.
  unfold not_in_dom in * |- * .
  unfold merge.
  rewrite H1.
  auto.
Qed.
  

Lemma disj_merge_assoc_right:
  forall M1 M2 M3, disjoint M1 M2 -> disjoint (merge M1 M2) M3 -> disjoint M1 (merge M2 M3).
Proof.
  intros M1 M2 M3 disj1 disj2.
  assert (disjoint M1 M3).
  apply disj_distr_left with M2.
  auto.
  apply disj_merge_assoc; auto.
Qed.


Lemma disj_merge_assoc_left:
  forall M1 M2 M3, disjoint M1 M2 -> merge (merge M1 M2) M3 = merge M2 (merge M1 M3).
Proof.
intros M1 M2 M3 disj.
assert (merge M1 M2 = merge M2 M1).
apply disj_merge_commut. auto.
rewrite H.
apply merge_assoc.
Qed.

Lemma merge_ABC_CAB:
  forall (M_1 M_2 M_3 : Map), 
    disjoint M_1 M_2 -> disjoint M_3 M_1 -> disjoint M_3 M_2 ->
    (merge (merge M_1 M_2) M_3) = (merge (merge M_3 M_1) M_2).
Proof.
  intros.
  assert (merge M_1 M_2 = merge M_2 M_1).
  apply disj_merge_commut; auto.
  rewrite H2.
  assert (merge M_3 M_1 = merge M_1 M_3).
  apply disj_merge_commut; auto.
  rewrite H3.
  assert ((merge (merge M_1 M_3) M_2) = (merge M_2 (merge M_1 M_3))).
  apply disj_merge_commut.
  apply disjoint_commut.
  apply disj_merge_assoc.
  apply disjoint_commut; auto.
  apply disjoint_commut; auto.
  rewrite H4.
  apply (merge_assoc M_2 M_1 M_3).
Qed.


(* Properties about disjUnion and merge *)

Lemma disjUnion_commut : 
  forall M1 M2 M, disjUnion M1 M2 M -> disjUnion M2 M1 M.
Proof.
  intros M1 M2 M.
  unfold disjUnion.
  intro. inversion_clear H.
  split. 
  apply disjoint_commut.
  auto.
  assert (merge M1 M2 = merge M2 M1).
  apply disj_merge_commut. auto. rewrite <- H; auto.
Qed.

Lemma disjUnion_in_left:
  forall (M1 M2 M : Map) (x : A) (y : B),
    disjUnion M1 M2 M -> in_dom x M1 -> maps_to M x y -> maps_to M1 x y.
Proof.
  intros.
  unfold disjUnion in H.
  inversion_clear H.
  assert (not_in_dom x M2).
  apply disj_exclusive_left with M1; auto.
  rewrite H3 in H1. unfold not_in_dom in H.
  unfold maps_to. unfold maps_to in H1.
  unfold merge in H1. rewrite H in H1.
  induction (M1 x). auto. inversion_clear H1.
Qed.


Lemma disjUnion_assoc:
  forall M1 M2 M3 M M', disjUnion M1 M2 M' -> disjUnion M' M3 M -> disjUnion M1 (merge M2 M3) M.
Proof.
  intros M1 M2 M3 M M'.
  unfold disjUnion.
  intros du1 du2.
  inversion_clear du1. inversion_clear du2.
  split.
  rewrite H0 in H1.
  apply disj_merge_assoc_right; auto.
  rewrite H0 in H2.
  rewrite H2.
  apply merge_assoc.
Qed.


Lemma disjUnion_assoc_right:
  forall M1 M2 M3 M, disjoint M2 M3 -> disjUnion M1 (merge M2 M3) M -> disjUnion (merge M1 M2) M3 M.
Proof.
  intros.
  assert (merge M1 M2 = merge M2 M1).
  apply disj_merge_commut; auto.
  unfold disjUnion in H0. inversion_clear H0.
  apply disjoint_commut.
  apply disj_distr_left with M3; auto.
  apply disjoint_commut; auto.
  rewrite H1.
  apply disjUnion_commut.
  apply disjUnion_assoc with (merge M3 M2); auto.
  unfold disjUnion.
  split.
  apply disjoint_commut; auto. auto.
  assert (merge M2 M3 = merge M3 M2).
  apply disj_merge_commut; auto.
  rewrite <- H2. apply disjUnion_commut; auto.
Qed.

Lemma disjUnion_shuffle_left:
  forall M1 M2 M3 M M', disjUnion M1 M2 M' -> disjUnion M' M3 M -> disjUnion M1 M3 (merge M1 M3).
Proof.
  intros.
  unfold disjUnion in H. unfold disjUnion in H0.
  inversion_clear H. inversion_clear H0.
  unfold disjUnion. split; auto.
  rewrite H2 in H.
  apply disj_distr_left with M2; auto.
Qed.

Lemma disjUnion_shuffle_right:
  forall M1 M2 M3 M M', disjUnion M1 M2 M' -> disjUnion M' M3 M -> disjUnion M2 M3 (merge M2 M3).
Proof.
  intros.
  unfold disjUnion in H. unfold disjUnion in H0. unfold disjUnion. split; auto.
  inversion_clear H. inversion_clear H0.
  rewrite H2 in H.
  apply disj_distr_right with M1; auto.
Qed.


Lemma disj_merge_unique : 
  forall M1 M2 M3 M4, disjoint M1 M2 -> disjoint M3 M4 -> merge M1 M2 = merge M3 M4 -> M1 = M3 -> M2 = M4.
Proof.
  intros M1 M2 M3 M4 disj1 disj2.
  intro eqMerge. intro eq13.
  rewrite eq13 in eqMerge.
  apply extensional_eqMap.
  intro.
  assert (merge M3 M2 x = merge M3 M4 x).
  rewrite <- eqMerge. auto.
  unfold merge in H.
  assert (in_dom x M3 \/ not_in_dom x M3).
  apply in_or_notin_dom.
  inversion_clear H0.
  rewrite eq13 in disj1.
  unfold disjoint in disj1.
  unfold disjoint in disj2.
  assert (not_in_dom x M3 \/ not_in_dom x M4); auto.
  assert (not_in_dom x M3 \/ not_in_dom x M2); auto.
  inversion_clear H0.
  unfold not_in_dom in H3.
  unfold in_dom in H1.
  inversion_clear H1.
  rewrite H3 in H0. inversion_clear H0.
  inversion_clear H2.
  unfold not_in_dom in H0.
  unfold in_dom in H1.
  inversion_clear H1.  
  rewrite H0 in H2. inversion_clear H2.
  unfold not_in_dom in H3.
  unfold not_in_dom in H0.
  rewrite <- H3 in H0. auto.
  unfold not_in_dom in H1.
  rewrite H1 in H. auto.
Qed.

Lemma disjUnion_unique :
  forall M1 M2 M3 M4, disjoint M3 M4 -> disjUnion M1 M2 (merge M3 M4) -> M1 = M3 -> M2 = M4.
Proof.
  intros M1 M2 M3 M4 disj1 disj2.
  unfold disjUnion in disj2.
  inversion_clear disj2.
  intro eq13.
  apply disj_merge_unique with M1 M3; auto.
Qed.


(* properties about subDom and eqDom *)
Lemma eqDom_refl: forall M, eqDom M M.
Proof.
  intro.
  unfold eqDom.
  unfold sub_dom. auto.
Qed.

Lemma eqDom_sym: forall M1 M2, eqDom M1 M2 -> eqDom M2 M1.
Proof.
  intros M1 M2.
  unfold eqDom.
  tauto.
Qed.

Lemma subDom_trans: forall M1 M2 M3, sub_dom M1 M2 -> sub_dom M2 M3 -> sub_dom M1 M3.
Proof.
  intros M1 M2 M3.
  unfold sub_dom.
  intros. auto.
Qed.

Lemma eqDom_trans : forall M1 M2 M3, eqDom M1 M2 -> eqDom M2 M3 -> eqDom M1 M3.
Proof.
  intros M1 M2 M3.
  unfold eqDom.
  intros. inversion_clear H. inversion_clear H0.
  split. apply subDom_trans with M2; auto. apply subDom_trans with M2; auto.
Qed.

Lemma subDom_update: forall (M : Map) (x : A) (y : B), sub_dom M (put M x y).
Proof.
  intros M x y.
  unfold sub_dom.
  intros x0 INDOM.
  unfold in_dom.
  unfold in_dom in INDOM.
  inversion_clear INDOM.
  unfold put.
  induction (beq_A x0 x). exists y. auto.
  rewrite H. exists x1. auto.
Qed.

Lemma subDom_inDom_update :
  forall (M : Map) (x : A) (y : B), in_dom x M -> sub_dom (put M x y) M.
Proof.
  intros.
  unfold sub_dom.
  unfold in_dom.
  intros.
  inversion_clear H0.
  unfold put in H1.
  induction (beq_A x0 x). unfold in_dom in H. 
  inversion_clear H. rewrite a. exists x2. auto. 
  exists x1. auto.
Qed.

Lemma eqDom_indom_update_eq : 
  forall (M : Map) (x : A) (y : B), in_dom x M -> eqDom M (put M x y).
Proof.
  intros M x y indom.
  unfold eqDom.
  split.
  apply subDom_update.
  apply subDom_inDom_update; auto.
Qed.


(* eqDom and merge / disjUnion *)
Lemma eqDom_merge: 
  forall M1 M2 M3 M4, merge M1 M2 = merge M3 M4 -> eqDom M1 M3 -> M1 = M3.
Proof.
  intros M1 M2 M3 M4 MG ED.
  apply extensional_eqMap.
  intro.
  unfold eqDom in ED.
  unfold sub_dom in ED.
  inversion_clear ED.
  assert (in_dom x M1 \/ not_in_dom x M1).
  apply in_or_notin_dom.
  inversion_clear H1.
  assert (in_dom x M3). apply H. auto.
  unfold in_dom in H2. inversion_clear H2.
  unfold in_dom in H1. inversion_clear H1.
  assert (merge M1 M2 x = merge M3 M4 x). rewrite MG. auto.
  unfold merge in H1.
  rewrite H3 in H1. rewrite H2 in H1.
  rewrite H2. rewrite H3. auto.
  assert (in_dom x M3 \/ not_in_dom x M3).
  apply in_or_notin_dom.
  inversion_clear H1.
  assert (in_dom x M1). apply H0. auto.
  unfold in_dom in H1. unfold not_in_dom in H2.
  rewrite H2 in H1. inversion_clear H1. inversion_clear H4.
  unfold not_in_dom in H2. unfold not_in_dom in H3.
  rewrite <- H3 in H2. auto.
Qed.


Lemma eqDom_disjUnion_left:  
  forall (M1 M2 M3 M4 M : Map), 
    disjUnion M1 M2 M -> disjUnion M3 M4 M -> eqDom M1 M3 -> M1 = M3.
Proof.
  intros M1 M2 M3 M4 M. unfold disjUnion.
  intros DU1 DU2 ED.
  inversion_clear DU1. inversion_clear DU2.
  rewrite H0 in H2.
  apply eqDom_merge with M2 M4; auto.
Qed.

Lemma eqDom_disjUnion_right:  
  forall M1 M2 M3 M4 M, 
    disjUnion M1 M2 M -> disjUnion M3 M4 M -> eqDom M1 M3 -> M2 = M4.
Proof.
  intros M1 M2 M3 M4 M DU1 DU2 ED.
  assert (M1 = M3). apply eqDom_disjUnion_left with M2 M4 M; auto.
  unfold disjUnion in DU1. unfold disjUnion in DU2.
  inversion_clear DU1. inversion_clear DU2.
  rewrite H1 in H3.
  apply disj_merge_unique with M1 M3; auto.
Qed.

Lemma eqDom_notin_presv:
  forall (M1 M2 : Map) (x : A), eqDom M1 M2 -> not_in_dom x M1 -> not_in_dom x M2.
Proof.
  intros M1 M2 x ED.
  unfold not_in_dom.
  unfold eqDom in ED.
  unfold sub_dom in ED. 
  unfold in_dom in ED.
  inversion_clear ED.
  intro.
  assert ((exists b : B, M1 x = Some b) -> exists b : B, M2 x = Some b). auto.
  assert ((exists b : B, M2 x = Some b) -> exists b : B, M1 x = Some b). auto.
  induction (M2 x).
  assert (exists b : B, M1 x = Some b). apply H3. exists a. auto.
  inversion_clear H4. rewrite H1 in H5. inversion_clear H5. 
  auto.
Qed.

Lemma eqDom_disj_presv:
  forall (M1 M2 M : Map), eqDom M1 M2 -> disjoint M1 M -> disjoint M2 M.
Proof.
  intros M1 M2 M ED.
  unfold disjoint.
  intros.
  assert (not_in_dom x M1 \/ not_in_dom x M). auto.
  inversion_clear H0. left.
  unfold eqDom in ED.
  apply eqDom_notin_presv with M1; auto.
  right. auto.
Qed.



Lemma disjUnion_presv_in_dom:
  forall (M1 M2 M : Map) (x : A), in_dom x M1 -> disjUnion M1 M2 M -> in_dom x M.
Proof.
  intros.
  unfold disjUnion in H0.
  inversion_clear H0.
  rewrite H2.
  apply merge_presv_in_dom.
  auto.
Qed.

Lemma disjUnion_presv_eqDom:
  forall (M1 M2 M3 M4 M M' : Map), 
    disjUnion M1 M2 M -> disjUnion M3 M4 M' -> eqDom M1 M3 -> eqDom M2 M4 -> eqDom M M'.
Proof.
  unfold eqDom.
  unfold sub_dom.
  unfold disjUnion.
  intros.
  inversion_clear H1.
  inversion_clear H2.
  inversion_clear H.
  inversion_clear H0.
  unfold in_dom.
  split.
  intros.
  inversion_clear H0.
  rewrite H6 in H8.
  unfold merge in H8.
  assert ((exists b : B, M1 x = Some b) \/ (M1 x = None)).
  induction (M1 x).
  left. exists a. auto. right. auto.
  inversion_clear H0.
  assert (in_dom x M3). apply H3. unfold in_dom. auto.
  unfold in_dom in H0. inversion_clear H0. rewrite H7. unfold merge.
  exists x1. rewrite H10. auto.
  assert (not_in_dom x M3).
  apply not_in_notin. intro. assert (in_dom x M1). apply H4; auto.
  unfold in_dom in H10. inversion_clear H10. rewrite H9 in H11. inversion_clear H11.
  rewrite H9 in H8. assert (in_dom x M4). apply H1. unfold in_dom. exists x0. auto.
  unfold in_dom in H10. inversion_clear H10. exists x1.
  rewrite H7. unfold merge. unfold not_in_dom in H0. rewrite H0. auto.
  
  intros.
  inversion_clear H0.
  rewrite H7 in H8.
  unfold merge in H8.
  assert ((exists b : B, M3 x = Some b) \/ (M3 x = None)).
  induction (M3 x).
  left. exists a. auto. right. auto.
  inversion_clear H0.
  assert (in_dom x M1). apply H4. unfold in_dom. auto.
  unfold in_dom in H0. inversion_clear H0. rewrite H6. unfold merge.
  exists x1. rewrite H10. auto.
  assert (not_in_dom x M1).
  apply not_in_notin. intro. assert (in_dom x M3). apply H3; auto.
  unfold in_dom in H10. inversion_clear H10. rewrite H9 in H11. inversion_clear H11.
  rewrite H9 in H8. assert (in_dom x M2). apply H5. unfold in_dom. exists x0. auto.
  unfold in_dom in H10. inversion_clear H10. exists x1.
  rewrite H6. unfold merge. unfold not_in_dom in H0. rewrite H0. auto.
Qed.


Lemma remove_left_presv_disjUnion:
  forall (M1 M2 M : Map) (x : A),
    disjUnion M1 M2 M -> in_dom x M1 -> disjUnion (remove M1 x) M2 (remove M x).
Proof.
  intros M1 M2 M x.
  unfold disjUnion. intros.
  inversion_clear H.
  split.
  apply remove_presv_disj; auto.
  rewrite H2.
  apply disj_remove_merge_merge_remove; auto.
Qed.

Lemma extend_presv_disjUnion:
  forall (M1 M2 M : Map) (x : A) (y : B),
    disjUnion M1 M2 M -> not_in_dom x M2 -> disjUnion (put M1 x y) M2 (put M x y).
Proof.
  intros M1 M2 M x y.
  unfold disjUnion.
  intros.
  inversion_clear H.
  split.
  apply extend_presv_disj; auto.
  rewrite H2.
  apply put_merge_merge_put.
Qed.

Lemma update_presv_disjUnion:
  forall (M1 M2 M : Map) (x : A) (y : B),
    disjUnion M1 M2 M -> in_dom x M1 -> disjUnion (put M1 x y) M2 (put M x y).
Proof.
  intros.
  assert (not_in_dom x M2).
  apply disj_exclusive_left with M1; auto.
  unfold disjUnion in H. inversion_clear H. auto.
  apply extend_presv_disjUnion; auto.
Qed.


(* disjunion and subseteq *)
Lemma disjUnion_subseteq_left:
  forall M_1 M_2 M, disjUnion M_1 M_2 M -> subseteq M_1 M.
Proof.
  intros.
  unfold disjUnion in H.
  inversion_clear H.
  rewrite H1.
  apply merge_extends_map.
Qed.

Lemma disjUnion_subseteq_right:
  forall M_1 M_2 M, disjUnion M_1 M_2 M -> subseteq M_2 M.
Proof.
  intros.
  apply disjUnion_subseteq_left with M_1.
  apply disjUnion_commut; auto.
Qed.

End Map.

